-- dos32lib.e
-- dos version of Win32Lib, v.5c
-- David Cuny
--  dcuny@lanset.com
-- Modified by Matt Arriola
--  euman2376@yahoo.com

-- List of changes is in CHANGES.TXT

constant VERSION = "0.5c pre-Alpha"
without warning
--with trace

-- to do:
-- there is no need to redraw the whole desktop when another window gets
-- focus - I need to set up an getFocus and loseFocus event to handle it
-- instead of refreshing the entire desktop. at a minimum, that should
-- stop the awful flashing when a message box pops up.
-- isVisible takes parent's visiblilty into account, for my purposes.
-- perhaps I need to split it into another routine - isVisible gets the
-- attribute, isViewable says it can be drawn.

-- should probably have a "close modal" option or something for dialogs.
--   example: openDialog() and closeDialog()
-- move drag and double click into event loop
-- mouse needs to be hidden during graphics.

--$ Win32Lib Emulation
-- Dos32Lib is an emulation of the Win32Lib library. These routines are
-- behave more or less the same as their Win32Lib counterparts.


without warning
include machine.e
include graphics.e
include get.e
include mighty.e
include font.e
include image.e
include wildcard.e
--include keyread.e
--include scan.e
constant
    ZOOM_WHEN_GROWING       = 0,
    ZOOM_WHEN_MINIMIZING    = 1,
    MENU_ON_TOP             = 0,
    blinks                  = 3
sequence usedcells

tick_rate(200)
procedure delay (atom tm)
    atom t
    t=time()
    while time() < t+tm do
    end while
end procedure
-----------------------------------------------------------------------------
-- General Routines
-- The following are general support routines.


-----------------------------------------------------------------------------
function iif( integer flag, object o1, object o2 )
    -- Return a value based on a flag.
    -- This is resembles the
    -- \code iif()
    -- instruction from C, but evaluates both all parameters. As
    -- a result, it cannot be used as a "real"
    -- \code iff
    -- statement.

    if flag then
	return o1
    else
	return o2
    end if
end function

-----------------------------------------------------------------------------
function fix( atom a )
    -- convert an atom to an integer
    integer int

    -- get integer portion
    int = floor( a )

    -- round up or down?
    if a - int >= 0.5 then
	int = int + 1
    end if

    return int

end function


-----------------------------------------------------------------------------
function abs( atom a )
    -- return absolute value of a number

    if a > 0 then
	return a
    else
	return -a
    end if

end function



-----------------------------------------------------------------------------
function remove( object item, sequence list )
    -- Remove all instance of item from the list.

    integer at

    while 1 do

	-- look for item
	at = find( item, list )

	-- not more instances?
	if at = 0 then
	    exit
	end if

	-- remove instance
	list = list[1..at-1] & list[at+1..length(list)]

    end while

    return list

end function


-----------------------------------------------------------------------------
function min( integer i1, integer i2 )
    if i1 < i2 then
	return i1
    else
	return i2
    end if
end function


-----------------------------------------------------------------------------
function max( integer i1, integer i2 )
    if i1 > i2 then
	return i1
    else
	return i2
    end if
end function


-----------------------------------------------------------------------------
global function or_all(sequence s)
-- or together all elements of a sequence
    atom result

    result = 0
    for i = 1 to length(s) do
	result = or_bits(result, s[i])
    end for
    return result
end function

-----------------------------------------------------------------------------
global function lo_word( atom a )
    -- return low word
    return and_bits( a, #FFFF )
end function

-----------------------------------------------------------------------------
global function hi_word( atom a )
    -- return high word
    return floor(a/power( 2, 16))
end function




integer abortHook
-----------------------------------------------------------------------------
procedure abortErr( sequence error, sequence args )
    -- print error and abort.
    -- while loading, the routine uses a simple text message
    -- for run time, it pops up a dialog box.
    call_proc( abortHook, {error, args} )
end procedure


-----------------------------------------------------------------------------
procedure abortErr1( sequence error, sequence args )

    -- Simple abortErr behavior.
    -- This returns the video display to normal, prints a message,
    -- and aborts.

    integer result

    -- restore screen
    result = graphics_mode(-1)

    -- clear screen
    bk_color( BLACK )
    text_color( WHITE )
    clear_screen()

    -- print message
    printf( 1, error & "\n", args )

    -- abort
    abort(0)

end procedure

abortHook = routine_id("abortErr1")



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Widget Class Attributes


-- control classes
-- Win32Lib
-- not all controls are supported yet
global constant
    Window          = 1,
    PushButton      = 2,
    DefPushButton   = 3,
    CheckBox        = 4,
    TriCheckBox     = 5,        -- not implemented
    Radio           = 6,
    Group           = 7,        -- not implemented
    List            = 8,        -- not implemented
    SortedList      = 9,        -- not implemented
    Combo           = 10,       -- not implemented
    SortedCombo     = 11,       -- not implemented
    SimpleCombo     = 12,       -- not implemented
    DropDownList    = 13,       -- not implemented
    EditText        = 14,
    MleText         = 15,       -- not implemented
    LText           = 16,
    RText           = 17,
    CText           = 18,       -- not implemented
    HScroll         = 19,
    VScroll         = 20,
    Icon            = 21,
    Bitmap          = 22,
    Menu            = 23,       -- NEW! 0.13
    MenuItem        = 24,       -- NEW! 0.13
    Pixmap          = 25


constant
    ControlCount    = MenuItem

-- these controls can't receive events
constant StaticControls = { LText, RText, Icon, Bitmap, Pixmap }


-- class actions
global sequence
    classShow,      -- display a control
    classClick,     -- handle a click event
    classKey        -- handle a key press

    classShow       = repeat( -1, ControlCount )
    classClick      = repeat( -1, ControlCount )
    classKey        = repeat( -1, ControlCount )


constant
    Class       = 1,
    Text        = 2,
    Parent      = 3,
    Size        = 4,
    Flags       = 5,
    State       = 6,        -- active/inactive/checked
    Data        = 7,        -- data
    MenuList    = 8,        -- true if menu exists
    Range       = 9,        -- {min, max, small change, large change}
    PenPos      = 10,       -- pen position
    PenColor    = 11,       -- pen color
    BackColor   = 12,       -- background color of control
    FontAttrib  = 13,       -- font attribute
    FontSize    = 14,       -- font size
    Font        = 15,       -- font
    MousePtr    = 19,       -- mouse pointer associated with control
    Focus       = 20,       -- id of control with focus/position of focus
    TabStops    = 21,       -- controls that can be tabbed to
    Anchor      = 22,       -- offset for list, sle
    Modal       = 23,       -- true if window is modal
    Bmp         = 24,       -- bitmap image associated with control
    PointerAt   = 25,       -- position of pointer/thumb
    Minimized   = 26,
    SavedSize   = 27,
    Grown       = 28,
    Cell        = 29,
    HasFrame    = 30,
    Selected    = 31,
    FirstItem   = 32,
    SavedSize2  = 33,
    TrashedSize = 34,
    Contents    = 35,
    WasGrown    = 36,

    MySize  = WasGrown


-- control states
constant
    IsVisible       = 1,
    IsEnabled       = 2,    -- control is active
    IsChecked       = 4,    -- control is checked
    IsPressed       = 8     -- control is pressed by mouse

-- button states
constant
    ActiveUnchecked     = 1,        -- is "armed" is better than "pressed?"
    ActiveChecked       = 2,
    PressedUnchecked    = 3,
    PressedChecked      = 4,
    InactiveUnchecked   = 5,
    InactiveChecked     = 6,

-- scroll bar value indexes for Range
    ScrollMin           = 1,
    ScrollMax           = 2,
    ScrollSmall         = 3,
    ScrollLarge         = 4

-- mouse action
global constant
    MOUSE_MOVE          = MOVE

sequence my
integer rgbFunc

-- window styles
global constant
    WS_OVERLAPPED   = #0,
    WS_POPUP        = #80000000,
    WS_CHILD        = #40000000,
    WS_MINIMIZE     = #20000000,
    WS_VISIBLE      = #10000000,
    WS_DISABLED     = #8000000,
    WS_CLIPSIBLINGS = #4000000,
    WS_CLIPCHILDREN = #2000000,
    WS_MAXIMIZE     = #1000000,
    WS_CAPTION      = #C00000,      --  WS_BORDER Or WS_DLGFRAME
    WS_BORDER       = #00800000,    -- creates border on window
    WS_DLGFRAME     = #400000,
    WS_HSCROLL      = #00100000,    -- horizontal scroll bar
    WS_VSCROLL      = #00200000,    -- vertical scroll bar
    WS_SYSMENU      = #80000,
    WS_THICKFRAME   = #40000,
    WS_GROUP        = #20000,
    WS_TABSTOP      = #10000,   -- use tab stop
    WS_SIMPLE       = #A0000,

    WS_MINIMIZEBOX = #20000,
    WS_MAXIMIZEBOX = #10000,

    WS_TILED = WS_OVERLAPPED,
    WS_ICONIC = WS_MINIMIZE,
    WS_SIZEBOX = WS_THICKFRAME,
    WS_OVERLAPPEDWINDOW = or_all({  WS_OVERLAPPED,
				    WS_CAPTION,
				    WS_SYSMENU,
				    WS_THICKFRAME,
				    WS_MINIMIZEBOX,
				    WS_MAXIMIZEBOX}),


    WS_TILEDWINDOW = WS_OVERLAPPEDWINDOW,

--
--   Common Window Styles
--  /


    WS_POPUPWINDOW = or_all({WS_POPUP, WS_BORDER, WS_SYSMENU}),

    WS_CHILDWINDOW = (WS_CHILD)



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
--$ Color Scheme

-----------------------------------------------------------------------------
-- set up video mode
global integer res
sequence vAttrib, cmd

-- default resolution
    res = 18

-- read the command line
cmd = command_line()
if length(cmd) = 3 then
    -- convert it
    cmd = value( cmd[3] )
    if cmd[1] = GET_SUCCESS then
	-- use abs() to strip of '-'
	res = abs(cmd[2])
    end if
end if


if graphics_mode(res) then
    abortErr( "Unable to set graphics mode %d", {res} )
end if

-- screen size
global integer
    xPixels,            -- screen x pixels
    yPixels             -- screen y pixels

    xPixels = 0
    yPixels = 0


-- get the attribute
vAttrib = video_config()
xPixels = vAttrib[ VC_XPIXELS ]
yPixels = vAttrib[ VC_YPIXELS ]
usedcells=repeat(0,yPixels/16)


-----------------------------------------------------------------------------
global procedure setPal( integer color, sequence code )
    --$ Color Scheme
    -- Set an rgb color.
    object ignore
    ignore = palette( color, floor( code/4) )
end procedure


-----------------------------------------------------------------------------
-- Win32 System Colors
global constant
    COLOR_SCROLLBAR         = 1,    -- scrollbar
    COLOR_BACKGROUND        = 2,    -- desktop
    COLOR_ACTIVECAPTION     = 3,    -- active window caption
    COLOR_INACTIVECAPTION   = 4,    -- inactive window caption
    COLOR_MENU              = 5,    -- menu background
    COLOR_WINDOW            = 6,    -- window background
    COLOR_WINDOWFRAME       = 7,    -- window frame
    COLOR_MENUTEXT          = 8,    -- menu text
    COLOR_WINDOWTEXT        = 9,    -- window text
    COLOR_CAPTIONTEXT       = 10,   -- in caption, size box and scroll arrow
    COLOR_ACTIVEBORDER      = 11,   -- active border
    COLOR_INACTIVEBORDER    = 12,   -- inactive border
    COLOR_APPWORKSPACE      = 13,   -- mdi background color
    COLOR_HIGHLIGHT         = 14,   -- items selected in control
    COLOR_HIGHLIGHTTEXT     = 15,   -- text of items selected in control
    COLOR_BTNFACE           = 16,   -- NT: face shading on buttons
    COLOR_BTNSHADOW         = 17,   -- edge shading on push buttons
    COLOR_GRAYTEXT          = 18,   -- disabled text
    COLOR_BTNTEXT           = 19,   -- button text
    COLOR_INACTIVECAPTIONTEXT = 20, -- text in inactive caption
    COLOR_BTNHIGHLIGHT      = 21    -- button highlight color

sequence sysColor
    sysColor = repeat( 0, 21 )

    -- window background
    sysColor[COLOR_BACKGROUND]          = BRIGHT_BLUE --WHITE

    -- window title bar
    sysColor[COLOR_ACTIVECAPTION]       = WHITE --BLUE
    sysColor[COLOR_CAPTIONTEXT]         = BLACK --BRIGHT_WHITE
    sysColor[COLOR_INACTIVECAPTION]     = WHITE
    sysColor[COLOR_INACTIVECAPTIONTEXT] = GRAY

    -- the window
    sysColor[COLOR_WINDOW]              = WHITE
    sysColor[COLOR_WINDOWFRAME]         = WHITE
    sysColor[COLOR_WINDOWTEXT]          = BLACK
    sysColor[COLOR_ACTIVEBORDER]        = BLUE
    sysColor[COLOR_INACTIVEBORDER]      = GRAY
    sysColor[COLOR_APPWORKSPACE]        = WHITE

    -- the menu bar
    sysColor[COLOR_MENU]                = WHITE --BRIGHT_BLUE
    sysColor[COLOR_MENUTEXT]            = BLACK

    -- the controls
    sysColor[COLOR_SCROLLBAR]           = BRIGHT_WHITE
    sysColor[COLOR_HIGHLIGHT]           = BLACK
    sysColor[COLOR_HIGHLIGHTTEXT]       = BRIGHT_WHITE
    sysColor[COLOR_BTNFACE]             = WHITE
    sysColor[COLOR_BTNSHADOW]           = GRAY
    sysColor[COLOR_GRAYTEXT]            = GRAY
    sysColor[COLOR_BTNTEXT]             = BLACK
    sysColor[COLOR_BTNHIGHLIGHT]        = BRIGHT_WHITE


-- adjust the colors
setPal( WHITE,         {206, 206, 206} )
setPal( GRAY,          {102, 102, 102} )
--setPal( BLUE,          { 51,  51, 153} )
setPal( BLUE,          { 0,206,206} )
--setPal( BRIGHT_BLUE,          { 99,  99, 206} )
setPal( BRIGHT_BLUE,   {153, 153, 204} )
setPal( BROWN,         {204, 153, 102} )
--setPal( BRIGHT_CYAN,   {156, 206, 206} )
--setPal( CYAN,          {156, 206, 206} )
setPal( MAGENTA,       { 96,   0,  96} )


-----------------------------------------------------------------------------
-- window metrics
integer
    theBorderSize,      -- thickness of border
    theBorderEdge,      -- which border edge is selected
    theGripSize,        -- thickness of grip
    theTitleSize,       -- height of title bar
    theMenuSize,        -- height of menu
    theMousePointer     -- current mouse pointer

-- sizes
    theBorderSize      = 5
    theGripSize        = 15
    theTitleSize       = 18
    theMenuSize        = 18
    theMousePointer = IDC_ARROW


constant
    X1  = 1,
    Y1  = 2,
    X2  = 3,
    Y2  = 4

--    IDC_APPSTARTING = 13, -- Standard arrow and small hourglass
--    IDC_ARROW = 1,        --  Standard arrow
--    IDC_CROSS = 3,        -- Crosshair
--    IDC_IBEAM = 4,        -- Text I-beam
--    IDC_ICON = 5,         -- Empty icon
--    IDC_NO = 6,           -- Slashed circle
--    IDC_SIZE = 7,         -- Four-pointed arrow
--    IDC_SIZEALL = 7,      -- Same as MC_SIZE
--    IDC_SIZENESW = 8,     -- Double-pointed arrow pointing northeast and southwest
--    IDC_SIZENS = 9,       -- Double-pointed arrow pointing north and south
--    IDC_SIZENWSE  = 10,   -- Double-pointed arrow pointing northwest and southeast
--    IDC_SIZEWE = 11,      -- Double-pointed arrow pointing west and east
--    IDC_UPARROW = 12,     -- Vertical arrow
--    IDC_WAIT = 2          -- Hourglass

-- compass directions
constant
    North   = 1,
    South   = 2,
    East    = 4,
    West    = 8



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
--$ Rendering Code
-- These routines enhance the built-in Euphoria routines by adding
-- clipping code.

sequence
    clipWin     -- clipping window
    clipWin = { 0, 0, xPixels, yPixels }


-- bit masks of clip results
constant
    LeftEdge    = 1,
    RightEdge   = 2,
    TopEdge     = 4,
    BottomEdge  = 8



-----------------------------------------------------------------------------
function getClipCode( integer x, integer y )
    -- Support function for clipLine().
    -- This returns bit mask indicating where the point lies,
    -- relative to the clipping window
    integer code

    -- default value
    code = 0

    -- check for intersection with top/bottom
    if y < clipWin[Y1] then
	code = TopEdge

    elsif y > clipWin[Y2] then
	code = BottomEdge

    end if

    -- check for intersection with left/right
    if x < clipWin[X1] then
	code = code+LeftEdge

    elsif x > clipWin[X2] then
	code = code+RightEdge

    end if

    return code

end function


-----------------------------------------------------------------------------
procedure clipLine( integer color, integer x1, integer y1,
			integer x2, integer y2 )
    -- Draws a clipped line.
    -- This routine takes a color and a set of points in the form:
    -- \p {x1, y2, x2, y2}
    -- and draws a clipped line.

    integer
    code1,      -- clip code for {x1,y1}
    code2,      -- clip code for {x2,y2}
    accept,     -- if true, draws the line
    code,       -- current clip code being examined
    x, y        -- intersection for current clip code

    -- get clip codes for end points
    code1 = getClipCode( x1, y1 )
    code2 = getClipCode( x2, y2 )

    -- default is not to accept
    accept = 0

    -- clip one, or both points
    while 1 do

	-- line lies entirely inside of the clipping area?
	if or_bits( code1, code2 ) = 0 then
	    -- set flags
	    accept = 1
	    exit
	end if

	-- line lies entirely outside of bounds?
	if and_bits( code1, code2 ) != 0 then
	    -- trivial reject
	    accept = 0
	    exit
	end if

	-- pick one of the offending points
	if code1 != 0 then
	    code = code1
	else
	    code = code2
	end if

	-- clip against left or right?
	if and_bits( code, LeftEdge ) then
	    -- clip against left edge
	    y = fix( y1+(y2-y1)*(clipWin[X1]-x1) / (x2-x1) )
	    x = clipWin[X1]

	elsif and_bits( code, RightEdge ) then
	    -- clip against bottom edge
	    y = fix( y1+(y2-y1)*(clipWin[X2]-x1) / (x2-x1) )
	    x = clipWin[X2]

	-- clip against top or bottom?
	elsif and_bits( code, TopEdge ) then
	    -- clip against top edge
	    x = fix( x1+(x2-x1)*(clipWin[Y1]-y1) / (y2-y1) )
	    y = clipWin[Y1]

	elsif and_bits( code, BottomEdge ) then
	    -- clip against bottom edge
	    x = fix(x1+(x2-x1)*(clipWin[Y2]-y1) / (y2-y1) )
	    y = clipWin[Y2]

	end if


	-- replace the point in the line
	if code = code1 then
	    -- clipped first point
	    x1 = x
	    y1 = y

	    -- recalculate code
	    code1 = getClipCode( x1, y1 )

	else
	    -- clipped second point
	    x2 = x
	    y2 = y

	    -- recalculate code
	    code2 = getClipCode( x2, y2 )

	end if

    end while

    -- if accepted, draw line
    if accept then
	mouse_pointer( 0 )
	draw_line( color, { {x1,y1}, {x2,y2} } )
	mouse_pointer( 1 )
    end if

end procedure


-----------------------------------------------------------------------------
procedure clipImage( sequence point, sequence image )
    -- Draws a clipped bitmap.
    -- This routine takes a point and a bitmap, clips it to the
    -- current clipping window, and displays it.

    integer x1, y1, x2, y2, diff, start, finish

    -- mangle point; hack
    point = point - 1

    -- get size
    x1 = point[X1]
    y1 = point[Y1]
    x2 = x1 + length(image[1]) - 1
    y2 = y1 + length(image) - 1

    -- off the window?
    if x2 < clipWin[X1]
    or x1 > clipWin[X2]
    or y2 < clipWin[Y1]
    or y1 > clipWin[Y2] then
	return
    end if

    -- perform "cheap" clipping first

    -- off top edge?
    if y1 < clipWin[Y1] then
	-- clip
	diff = clipWin[Y1] - y1
	image = image[diff..length(image)]
	-- new values
	y1 = clipWin[Y1]
	y2 = y2 - diff
    end if

    -- off bottom?
    if y2 > clipWin[Y2] then
	-- clip
	diff = y2 - clipWin[Y2]
	image = image[1..length(image)-diff]
    end if


    -- off right or left edge? these are put into one test
    -- so I go through the loop only once
    if x1 < clipWin[X1]
    or x2 > clipWin[X2] then

	-- clip left?
	if x1 < clipWin[X1] then
	    -- difference
	    start = clipWin[X1] - x1
	    x1 = x1 + start
	else
	    -- no change
	    start = 1
	end if

	-- clip right?
	finish = length( image[1] )
	if x2 > clipWin[X2] then
	    -- difference
	    finish = finish - (x2 - clipWin[X2])
	end if

	-- clip
	for i = 1 to length( image ) do
	    image[i] = image[i][start..finish]
	end for
    end if

    -- hide the mouse
    mouse_pointer( 0 )

    -- display
    display_image( { x1, y1 }, image )

    -- show the mouse
    mouse_pointer( 1 )


end procedure





-----------------------------------------------------------------------------
procedure clipXorLine( integer color, integer x1, integer y1,
					integer x2, integer y2 )
    -- Draw a clipped XOR line.
    -- This <i>only</i> draws horizontal or vertical lines.
    -- Drawing an XOR line twice erases it.

    integer tmp, len
    sequence s

    -- hide the mouse
    mouse_pointer( 0 )

    -- horizontal, or vertical?
    if x1 = x2 then
	-- horizontal
	if y1 > y2 then
	    tmp = y1
	    y2 = y1
	    y1 = tmp
	end if

	for y = y1 to y2 do
	    pixel(xor_bits(get_pixel({x1,y,1}),color),{x1,y})
	end for

    else
	-- vertical
	if x2 < x1 then
	    -- swap
	    tmp = x1
	    x1 = x2
	    x2 = x1
	end if

	-- length
	len = x2-x1+1

	-- line to xor against
	s = repeat( color, len )

	-- get the line, xor it, and draw it
	pixel(xor_bits(get_pixel({x1,y1,len}),s),{x1,y1})

    end if

    -- show the mouse
    mouse_pointer( 1 )


end procedure



-----------------------------------------------------------------------------
procedure clipXorRect( integer color, sequence r )

    -- draw an XOR rectangle.
    -- drawing an XOR rectangle twice erases it.

    integer x1, y1, x2, y2

    -- get points
    x1 = r[X1]
    y1 = r[Y1]
    x2 = r[X2]
    y2 = r[Y2]

    -- draw a rectangle
    -- need to avoid overlapping, or corners will disappear
    clipXorLine( color, x1, y1, x2, y1 )
    clipXorLine( color, x2, y1+1, x2, y2-1 )
    clipXorLine( color, x1, y2, x2, y2 )
    clipXorLine( color, x1, y1+1, x1, y2-1 )

end procedure


-----------------------------------------------------------------------------
procedure clipPixel( object o, integer x1, integer y1 )

    -- need to add clipping code!
    integer x2, p1, p2, diff

    -- single point?
    if integer( o ) then
	-- out of clip window?
	if x1 < clipWin[X1]
	or x1 > clipWin[X2]
	or y1 < clipWin[Y1]
	or y1 > clipWin[Y2] then
	    -- do nothing
	else
	    mouse_pointer( 0 )
	    pixel( o, {x1, y1} )
	    mouse_pointer( 1 )
	end if

    -- seqence of pixels
    else

	-- end position
	x2 = x1+length( o )-1

	-- in clip area?
	if x2 < clipWin[X1]
	or x1 > clipWin[X2]
	or y1 < clipWin[Y1]
	or y1 > clipWin[Y2] then
	    -- do nothing
	else
	    -- clip ends
	    if x1 < clipWin[X1] then
		p1 = clipWin[X1] - x1
	    else
		p1 = 1
	    end if

	    if x2 > clipWin[X2] then
		p2 = length(o) - (x2 - clipWin[X2])
	    else
		p2 = length(o)
	    end if

	    -- display
	    mouse_pointer( 0 )
	    pixel( o[p1..p2], {x1, y1} )
	    mouse_pointer( 1 )

	end if
    end if

end procedure


-----------------------------------------------------------------------------
procedure clipFocusRect( integer fillColor, integer x1, integer y1,
				integer x2, integer y2 )

    -- draw a focus rectangle
    integer long
    sequence pix

    long = x2 - x1
    pix = repeat( fillColor, long )
    for i = 1 to long by 2 do
	pix[i] = BLACK
    end for
    clipPixel( pix, x1, y1 )
    clipPixel( pix, x1, y2 )

    -- vertical lines
    for i = y1 to y2 by 2 do
	clipPixel( BLACK, x1, i )
	clipPixel( BLACK, x2, i )
    end for


end procedure


-----------------------------------------------------------------------------
procedure clipRect( integer color, integer filled, integer x1, integer y1,
				integer x2, integer y2 )

    -- Displays a clipped box, optionally filled.

    -- outside of clip box?
    if x1 > clipWin[X2]
    or x2 < clipWin[X1]
    or y1 > clipWin[Y2]
    or y2 < clipWin[Y1] then
	-- nothing to draw
	return
    end if

    if filled then
	-- clip and use polygon.
	-- faster than putting clipLine in a loop

	if x1 < clipWin[X1] then x1 = clipWin[X1] end if
	if x2 > clipWin[X2] then x2 = clipWin[X2] end if
	if y1 < clipWin[Y1] then y1 = clipWin[Y1] end if
	if y2 > clipWin[Y2] then y2 = clipWin[Y2] end if

	mouse_pointer( 0 )

	-- draw a filled polygon
	polygon( color, 1, { { x1, y1 },
			     { x2, y1 },
			     { x2, y2 },
			     { x1, y2 } } )

	mouse_pointer( 1 )


    else
	-- call clipLine to draw a box
	clipLine( color, x1, y1, x2, y1 )
	clipLine( color, x2, y1, x2, y2 )
	clipLine( color, x2, y2, x1, y2 )
	clipLine( color, x1, y2, x1, y1 )
    end if

end procedure


-----------------------------------------------------------------------------
procedure clipFramedRect( integer color, integer x1, integer y1,
				integer x2, integer y2 )

    -- draw a rectangle with a black border
    clipRect( BLACK, 0, x1, y1, x2, y2 )
    clipRect( color, 1, x1+1, y1+1, x2-1, y2-1 )


end procedure


-----------------------------------------------------------------------------
procedure clipShadowedRect( integer color, integer x1, integer y1,
				integer x2, integer y2 )

    -- draw a rectangle with a black border and shadows
    -- shadow is added past size of rectangle

    clipRect( color, 1, x1, y1, x2, y2 )
    clipRect( BLACK, 0, x1, y1, x2, y2 )
    -- shadow
    clipLine( BLACK, x2+1, y1+1, x2+1, y2+1 )
    clipLine( BLACK, x1+1, y2+1, x2+1, y2+1 )


end procedure

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Routines for Drawing into Bitmaps
-- these need to eventually be folded into screen drawing tools

-----------------------------------------------------------------------------
function bitmapPixel( sequence bmp, integer color, integer x, integer y )

    -- draws a pixel on a bitmap, with clipping
    -- needs to eventually be expanded to handle sequence of pixels

    if y > 0 and y <= length( bmp ) then
	if x > 0 and x <= length( bmp[1] ) then
	    bmp[y][x] = color
	end if
    end if
    return bmp
end function


-----------------------------------------------------------------------------
function bitmapLine( sequence bmp, integer color, integer x1, integer y1,
			    integer x2, integer y2 )

    -- draws a line on a bitmap, with clipping

    -- integer direction
    integer direction, maxX, maxY
    atom deltaX, deltaY, x, y

    -- get bounds
    maxX = length(bmp[1])
    maxY = length(bmp)

    -- draw a line in a bitmap
    if x1 = x2 then
       -- vertical line

	-- swap order?
	if y2 < y1 then
	    -- swap
	    y  = y1
	    y1 = y2
	    y2 = y1
	end if

	-- in range?
	if  y1 <= maxY
	and y2 > 0 then

	    -- clip
	    y1 = iif( y1 < 1, 1, y1 )
	    y2 = iif( y2 > maxY, maxY, y2 )

	    -- draw
	    for i = y1 to y2 do
		bmp[i][x1] = color
	    end for

	end if


    elsif y1 = y2 then
	-- horizontal line

	-- swap order?
	if x2 < x1 then
	    -- swap
	    x  = x1
	    x1 = x2
	    x2 = x1
	end if


	-- in range?
	if  x2 > 0
	and x1 <= maxX then

	    -- clip
	    x1 = iif( x1 < 1, 1, x1 )
	    x2 = iif( x2 > maxX, maxX, x2 )

	    -- draw
	    bmp[y1][x1..x2] = color

	end if

    else
	-- diagonal
	deltaX = x2 - x1
	deltaY = y2 - y1

	-- move on x or y axis?
	if abs( deltaX ) > abs( deltaY ) then

	    -- positive or negative slope?
	    direction = iif( x1>x2, 1, -1 )

	    -- start point
	    y = y1

	    -- draw the line along x axis
	    for i = x1 to x2 by direction do
		bmp = bitmapPixel( bmp, color, i, floor(y) )
		y = y + deltaY
	    end for

	else

	    -- positive or negative slope?
	    direction = iif( y1 > y2, 1, -1 )

	    -- start point
	    x = x1

	    -- draw the line
	    for i = y1 to y2 by direction do
		bmp = bitmapPixel( bmp, color, i, floor(y) )
		x1 = x + deltaX
	    end for


	end if

    end if

    return bmp

end function


-----------------------------------------------------------------------------
function bitmapBox( sequence bmp, integer color, integer fill,
			    integer x1, integer y1,
			    integer x2, integer y2 )

    integer maxX, maxY

    -- draw a box in a bitmap
    if fill then

	-- get the max points
	maxX = length( bmp[1] )
	maxY = length( bmp )

	-- inside the bounds?
	if  x1 <= maxX and x2 < 1
	and y1 <= maxY and y2 < 1 then

	    -- clip the points
	    x1 = iif( x1 < 1, 0, x1 )
	    y1 = iif( y1 < 1, 0, y1 )
	    x2 = iif( x2 > maxX, x2, maxX )
	    y2 = iif( y2 > maxY, y2, maxY )

	    -- fill
	    for y = y1 to y2 do
		bmp[y][x1..x2] = color
	    end for

	end if

    else
	-- draw lines
	-- ordered for speed
	bmp = bitmapLine( bmp, color, x1, y1, x2, y1 )
	bmp = bitmapLine( bmp, color, x1, y2, x2, y2 )
	bmp = bitmapLine( bmp, color, x1, y1, x1, y2 )
	bmp = bitmapLine( bmp, color, x2, y1, x2, y2 )

    end if


    return bmp

end function

-----------------------------------------------------------------------------
function saveImage( integer x1, integer y1, integer x2, integer y2 )
    -- "safe" version of save_image; handles negative indexes

    -- clip to edge of screen
    return save_image( {max(x1,0),max(y1,0)}, {max(x2,0),max(y2,0)} )


end function
-----------------------------------------------------------------------------
procedure restoreImage( integer x1, integer y1, sequence s )

    -- "safe" version of display_image; handles negative indexes

    -- clip to edge of screen
    if x1 < 0 then
	x1 = 0
    end if

    if y1 < 0 then
	y1 = 0
    end if

    display_image( {max(x1,0),max(y1,0)}, s )

end procedure


-----------------------------------------------------------------------------
constant
    LeftJustify     = 1,
    RightJustify    = 2,
    CenterJustify   = 4,
    TopJustify      = 8,
    BottomJustify   = 16,
    MiddleJustify   = 32

-----------------------------------------------------------------------------
function justifyAt( integer x1, integer y1, integer x2, integer y2,
		integer flags, sequence image )

    -- return position to place image at

    integer tall, wide, atX, atY

    -- metrics of image
    tall    = length(image)
    wide    = length( image[1] )

    -- horizontal
    if and_bits( flags, RightJustify ) then
	-- right justify
	atX = x2 - wide
    elsif and_bits( flags, CenterJustify ) then
	-- center justify
	atX = floor(x1 + (( x2-x1 )/2 ) - (wide/2))
    else
	-- default, left justify
	atX = x1
    end if

    -- vertical
    if and_bits( flags, TopJustify ) then
	-- top justify
	atY = y1
    elsif and_bits( flags, BottomJustify ) then
	-- bottom justify
	atY = y2 - tall
    else
	-- center justify
	atY = floor(y1 + (( y2-y1 )/2 ) - (tall/2))+1
    end if

    -- draw it
    return {atX, atY}

end function


-----------------------------------------------------------------------------
procedure clipText( integer x, integer y, sequence text,
		    integer fore, integer back, integer attrib,
		    integer focus )

    -- draw text, with optional focus rectangle

    integer x2, y2
    sequence image

    image = build_text_image( text, fore, back, attrib )

    -- display
    clipImage( {x, y}, image )

    -- focus?
    if focus then
	-- extent
	x2 = x + length( image[1] ) - 1
	y2 = y + length( image ) - 1

	-- focus
	clipFocusRect( back, x, y, x2, y2 )

    end if

end procedure


-----------------------------------------------------------------------------
function grayOut(sequence s, integer grayed)
    if not grayed then return s end if
    for row = 3 to length(s)-1 do
	for col = 3 to length(s[row])-1 do
	    if s[row][col]='x' then
		s[row][col]='X'
	    end if
	end for
    end for
    return s
end function
function buildImage( sequence s, integer fColor, integer bColor )
    -- create an image of the controls

    -- replace colors
    for i = 1 to length( s ) do
	for j = 1 to length( s[i] ) do
	    if      s[i][j] = ' ' then s[i][j] = bColor
	    elsif   s[i][j] = 'x' then s[i][j] = fColor
	    elsif   s[i][j] = 'X' then s[i][j] = GRAY
	    elsif   s[i][j] = '.' then s[i][j] = BRIGHT_WHITE
	    elsif   s[i][j] = '/' then s[i][j] = WHITE
	    else    s[i][j] = fColor
	    end if
	end for
    end for

    return s

end function




-----------------------------------------------------------------------------
-- GENERIC CONTROLS
--

-----------------------------------------------------------------------------
function isVisible( integer self )
    -- Win32
    -- returns true if control is visible
    integer parent

    -- get parent of control
    parent = my[self][Parent]
    if parent then
	-- parent and control must be visible
	return  and_bits( my[parent][State], IsVisible ) and
		and_bits( my[self][State], IsVisible )
    else
	-- just window
	return and_bits( my[self][State], IsVisible )
    end if


end function

-----------------------------------------------------------------------------
global function isEnabled( integer self )
    -- Win32
    return and_bits( my[self][State], IsEnabled )
end function

-----------------------------------------------------------------------------
global function isChecked( integer self )
    -- Win32
    return and_bits( my[self][State], IsChecked )
end function

-----------------------------------------------------------------------------
function isPressed( integer self )
    -- Win32
    return and_bits( my[self][State], IsPressed )
end function


-----------------------------------------------------------------------------
procedure showControl( integer self )

    -- draw a control
    integer class, parent, procId

    -- don't display if parent is not visible
    parent = my[self][Parent]
    if parent > 0 then
	if not isVisible( parent ) then
	    return
	end if
    end if

    -- don't draw if is invisible
    if not isVisible( self ) then
	return
    end if


    class = my[self][Class]
    procId = classShow[class]
    if procId > -1 and class != Pixmap then
	-- hide the mouse
	mouse_pointer(0)

	-- call the proc
	call_proc( classShow[class], {self} )

	-- restore the mouse
	mouse_pointer(1)
    end if

end procedure








include misc.e
sequence fontDir
integer SLASH_CHAR

if platform()=LINUX then
    SLASH_CHAR = '/'
else
    SLASH_CHAR = '\\'
end if

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Fonts

if atom(getenv("fontdir")) then
    fontDir="."&SLASH_CHAR
else
    fontDir=getenv("fontdir")
end if
-----------------------------------------------------------------------------
-- currently, only 10 point fonts are supported
constant FontNames = {
    "Arial",                fontDir&"Arial10.f",
    "Ariel",                fontDir&"Arial10.f",
    "Bookman Old Style",    fontDir&"Book10.f",
    "Comic Sans MS",        fontDir&"Comic10.f",
    "Courier",              fontDir&"Cour10.f",
    "Courier New",          fontDir&"Cour10.f",
    "Fixedsys",             fontDir&"FixSys10.f",
    "MS Sans Serif",        fontDir&"Sans10.f",
    "Small Fonts",          fontDir&"Small07",
    "System",               fontDir&"System10.f",
    "Tahoma",               fontDir&"Tahoma10.f",
    "Times New Roman",      fontDir&"Times10.f",
    "Verdana",              fontDir&"Verdan10.f" }




-----------------------------------------------------------------------------
-- load the fonts
integer
    SysFont, FixedFont

-- system font
--SysFont = fload( "verdan08.f" )
SysFont = fload( "arial10.f" )
FixedFont = fload("fixsys08.f")

-----------------------------------------------------------------------------

-- user defined actions: 1 per control
-- Win32Lib
global sequence
-- NEW! 0.4 renamed from onLoad to onOpen
    onOpen,
    onPaint,
    onClick,
    onChange,
    onMouse,
    onRightClick,
    onScroll,
    onKeyPress,
-- NEW! 0.4 added
    onKeyDown,
    onGetFocus,
    onLostFocus,
    onTimer,
    onResize,
    onClose

    onOpen  = {}
    onPaint = {}
    onClick = {}
    onChange = {}
    onMouse = {}
    onRightClick = {}
    onScroll = {}
-- NEW! 0.4 added
    onKeyDown = {}
    onKeyPress = {}
    onGetFocus = {}
    onLostFocus = {}
    onTimer = {}
    onResize = {}
    onClose = {}



-- timers
sequence timers
timers = {}         -- holds timer list

constant
    TimerWin    = 1,
    TimerId     = 2,
    TimerTicks  = 3,
    TimerStart  = 4

-- window stack
sequence
    windowStack     -- stack of open windows
    windowStack = {}

-- mouse controls
constant
    MOUSE_EVENT     = 1,    -- used by get_mouse
    MOUSE_BUTTONS   = 1,    -- used by poll_mouse
    MOUSE_X         = 2,
    MOUSE_Y         = 3,
    MOUSE_LEFT      = 1,
    MOUSE_RIGHT     = 2,
    MOUSE_MIDDLE    = 3

-- event loop flag
integer exitEventLoop
    exitEventLoop = 0

-----------------------------------------------------------------------------
integer
    mainWindow

    mainWindow = 0      -- window eventLoop() was entered with

global constant
    Default = -1

-- colors
-- NEW! 0.4 added more colors
global constant
    Black           = BLACK,
    Blue            = BLUE,
    Green           = GREEN,
    Magenta         = MAGENTA,
    Red             = RED,
    Cyan            = CYAN,
    Brown           = BROWN,
    White           = WHITE,
    Gray            = GRAY,
    BrightBlue      = BRIGHT_BLUE,
    BrightGreen     = BRIGHT_GREEN,
    BrightMagenta   = BRIGHT_MAGENTA,
    BrightRed       = BRIGHT_RED,
    BrightCyan      = BRIGHT_CYAN,
    Yellow          = YELLOW,
    BrightWhite     = BRIGHT_WHITE


-- font attributes
global constant
    Bold            = 1,
    Italic          = 2,
    Underline       = 4,
    Strikeout       = 8


-- holds attributes
my = {}


-----------------------------------------------------------------------------
-- keys
-- NEW! 0.4 increased number of keys defined
global constant

    VK_BACK = 8,
    VK_TAB = 9,
    VK_RETURN = 13,
    VK_ESCAPE = 27,
    VK_SPACE = 32,
    VK_SHIFT_TAB = 271,
    VK_HELP = 315,

    VK_F1 = 315,
    VK_F2 = 316,
    VK_F3 = 317,
    VK_F4 = 318,
    VK_F5 = 319,
    VK_F6 = 320,
    VK_F7 = 321,
    VK_F8 = 322,
    VK_F9 = 323,
    VK_F10 = 324,
    VK_F11 = 389,
    VK_F12 = 390,

    VK_HOME = 327,
    VK_UP = 328,
    VK_PRIOR = 329,
    VK_LEFT = 331,
    VK_RIGHT = 333,
    VK_END = 335,
    VK_DOWN = 336,
    VK_NEXT = 337,
    VK_INSERT = 338,
    VK_DELETE = 339,

    -- my additions
    VK_PAGEUP   = VK_PRIOR,
    VK_PAGEDOWN = VK_NEXT

-- non-Win32
-- NEW! 0.4
constant
    SHIFT_TAB = 271,

    CTRL_LEFT = 371,
    CTRL_RIGHT = 372,
    CTRL_END = 373,
    CTRL_HOME = 375,
    CTRL_PAGE_UP = 388,
    CTRL_PAGE_DOWN = 374,
    CTRL_UP = 397,
    CTRL_DOWN = 401,
    CTRL_INSERT = 402,
    CTRL_TAB = 404,
    CTRL_ENTER = 10,

    ALT_TAB = 421




-----------------------------------------------------------------------------
integer
    winX1,      -- position of window's X postition
    winY1,      -- position of window's Y position
    winX2,
    winY2

-----------------------------------------------------------------------------
function hasMenu( integer self )
    -- returns count of menu items
    -- ex:  ? menuItems( MyWindow )
    --      --> 4
    return length( my[self][MenuList] )
end function


-----------------------------------------------------------------------------
function getClientSize( integer self )

    -- returns size of client area
    sequence size

    -- get the window size
    size = my[self][Size][X2..Y2]

    -- subtract some pixels
    size = size - { 3, 3 }

    -- subtract the title bar
    size[Y1] = size[Y1] - (theTitleSize)

    -- add menu?
    if hasMenu( self ) then
	size[Y1] = size[Y1] - (theMenuSize)
    end if

    return size

end function


-----------------------------------------------------------------------------
global function getClientOrg( integer self )

    -- Win32 API
    -- returns origin of client area
    sequence org

    -- get the coordinates
    org = my[self][Size][X1..Y1]

    -- add border, plus pixels
    org[X1..Y1] = org[X1..Y1] + { 2, 1 }

    -- add title bar
    org[Y1] = org[Y1] + theTitleSize

    -- add menu?
    if hasMenu( self ) then
	org[Y1] = org[Y1] + theMenuSize
    else
	-- seperator
	org[Y1] = org[Y1] + 1
    end if

    return org

end function

-----------------------------------------------------------------------------
global function getClientRect( integer self )

    -- Win32 API
    -- return client's relative rectangle
    return { 0, 0 } & ( getClientSize( self ) )
end function


-----------------------------------------------------------------------------
global function getAbsClientRect( integer self )

    -- return client's screen rectangle
    sequence org

    -- get origin
    org = getClientOrg( self )

    -- offset origin to get screen position
    return org & ( getClientSize( self ) + org )

end function



-----------------------------------------------------------------------------
function intoWindowPoint( integer self, sequence point )

    -- convert the point to one relative to the client

    if my[self][Class] = Window then
	-- relative to client origin
	point = point - getClientOrg( self )
    else
	-- relative to parent's client origin and control offset
	point = point - getClientOrg( my[self][Parent] )
	point = point - my[self][Size][X1..Y1]
    end if

    return point

end function


-----------------------------------------------------------------------------
global function getWindowRect( integer self )

    -- Win32 API
    -- returns bounding rectangle of window in screen coordinates
    sequence size

    -- size of display area
    size = my[self][Size]

    -- window or control?
    if my[self][Class] != Window then

	-- adjust to parent's origin
	size[X1..Y1] = size[X1..Y1] + getClientOrg( my[self][Parent] )

    end if


    -- add origin to extent
    size[X2..Y2] = size[X2..Y2] + size[X1..Y1]

    return size

end function

-----------------------------------------------------------------------------
global function getWindowSize( integer self )

    -- Win32Lib
    -- this is used to see if a point (relative to a window) hit a
    -- control in a window
    sequence size

    -- get size
    size = my[self][Size]

    -- add origin to make wide/deep into points
    size[X2..Y2] = size[X2..Y2] + size[X1..Y1]

    return size


end function

-----------------------------------------------------------------------------
global function getWindowExtent( integer self )

    -- Win32Lib
    -- get extent metrics for current window
    -- returns { deep, wide }
    return my[self][Size][X2..Y2]

end function



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Window Clipping

sequence clipStack
    clipStack = {}


-----------------------------------------------------------------------------
procedure pushClip()
    -- save current clipping region
    clipStack = append( clipStack, clipWin )
end procedure


-----------------------------------------------------------------------------
procedure popClip()
    -- restore saved clipping region
    integer depth

    -- get depth of stack
    depth = length( clipStack )
    -- restore
    clipWin = clipStack[depth]
    -- pop stack
    clipStack = clipStack[1..depth-1]

end procedure


-----------------------------------------------------------------------------
procedure clipTo( sequence rect )

    --set clip window
    clipWin = { max(rect[X1],0),        max(rect[Y1],0),
		min(rect[X2],xPixels),  min(rect[Y2],yPixels) }

end procedure


-----------------------------------------------------------------------------
procedure clipToClient( integer self )

    -- sets winX and winY to the client area,
    -- and sets clipping window
    sequence rect

    -- get the absolute coordinates
    rect = getAbsClientRect( self )

    -- save them
    winX1 = rect[X1]
    winY1 = rect[Y1]
    winX2 = rect[X2]
    winY2 = rect[Y2]

    -- set clipping window
    clipTo( rect )

end procedure


-----------------------------------------------------------------------------
procedure clipToWindow( integer self )

    -- sets clipping window to window; add pixel for shadow
    clipTo( getWindowRect( self ) + {0,0,1,1} )

end procedure


-----------------------------------------------------------------------------
procedure clipToDesktop()
    -- set clipping window to desktop
    clipTo( { 0, 0, xPixels, yPixels } )
end procedure



-----------------------------------------------------------------------------
global function getFontSize( integer self )

    -- Win32Lib
    -- get metrics for current font

    -- return {width, height}
    return { get_text_width("X", 0), get_font_height() }

end function

-----------------------------------------------------------------------------
procedure createMenu( integer self, integer owner )
    -- Win32Lib
    -- create a popup, assigned to a control
    -- attach to owner
    my[owner][MenuList] = append( my[owner][MenuList], self )
end procedure

-----------------------------------------------------------------------------
procedure createMenuItem( integer self, integer owner )
    -- Win32Lib
    -- create a submenu, assigned to a control
    my[owner][MenuList] = append( my[owner][MenuList], self )
end procedure


-----------------------------------------------------------------------------
procedure doAction( sequence actions, integer index )
    if actions[index] > -1 then
	call_proc( actions[index], {index} )
    end if
end procedure




-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- drawing the desktop

-----------------------------------------------------------------------------
procedure showDesktop()

    -- redraw the desktop

    integer tileTall, self

    mouse_pointer( 0 )

    -- fill with solid color
    polygon( sysColor[COLOR_BACKGROUND],
		1,  { { 0      , 0 },
		      { xPixels, 0 },
		      { xPixels, yPixels },
		      { 0      , yPixels } } )

    -- draw each window
    for this = length( windowStack ) to 1 by -1 do
	-- get window
	self = windowStack[this]
	-- visible?
	if isVisible( self ) then
	    -- draw it
	    showControl( self  )
	end if
    end for

    mouse_pointer( 1 )

end procedure



-----------------------------------------------------------------------------
global procedure setText( integer self, sequence s )
    -- Win32
    -- set the text in control
    my[self][Text] = s
end procedure


-----------------------------------------------------------------------------
global function getText( integer self )
    -- Win32Lib
    -- return text in control
    return my[self][Text]
end function


-----------------------------------------------------------------------------
global procedure setFlag( integer self, integer bit, integer flag )
    -- Win32
    -- set the attribute flag

    -- turn off the flag
    if flag then
	-- turn it on
	my[self][State] = or_bits( my[self][State], bit )
    else
	-- turn it off
	my[self][State] = and_bits( not_bits( bit ), my[self][State] )
    end if

    -- need to redraw? window refresh is handled in setVisible
    if my[self][Class] != Window then
	if  bit = IsVisible
	and not flag then
	    -- control turned invisible; update parent to erase
	    showControl( my[self][Parent] )
	else
	    -- redraw control
	    showControl( self )
	end if
    end if

end procedure


-----------------------------------------------------------------------------
global procedure setVisible( integer self, integer flag )
    -- Win32

    integer top

    -- alter the flag
    setFlag( self, IsVisible, flag )


    -- if window, place in proper location
    if my[self][Class] = Window then

	-- get id of top item, if not self
	top = 0
	if length( windowStack ) then
	    -- not self?
	    if windowStack[1] != self then
		top = windowStack[1]
	    end if
	end if

	-- remove from the stack
	windowStack = remove( self, windowStack )

	-- setting focus?
	if flag then

	    -- place in front
	    windowStack = prepend( windowStack, self )

	    -- need to lose focus on top item
	    if top then

		-- redraw prior top window without focus
		showControl( top )

		-- lost focus behavior?
		if onLostFocus[top] != -1 then
		    call_proc( onLostFocus[top], {} )
		end if

	    end if

	    -- draw new window on top, with focus
	    showControl( self )

	    -- focus behavior?
	    if onGetFocus[self] != -1 then
		call_proc( onGetFocus[self], {} )
	    end if


	else

	    -- lose focus behavior?
	    if onLostFocus[self] != -1 then
		call_proc( onLostFocus[self], {} )
	    end if

	    -- update the whole desktop
	    showDesktop()

	    -- is there a new window on top?
	    if length( windowStack ) then

		-- get top window
		top = windowStack[1]

		-- gain focus behavior?
		if onGetFocus[top] != -1 then
		    call_proc( onGetFocus[top], {} )
		end if

	    end if

	    -- has the main window (or a modal window) closed?
	    if self = mainWindow then
		exitEventLoop = 1
	    end if

	end if

    end if

end procedure

-----------------------------------------------------------------------------
global procedure setEnable( integer self, integer flag )
    -- Win32
    setFlag( self, IsEnabled, flag )
end procedure

-----------------------------------------------------------------------------
procedure setPressed( integer self, integer flag )
    -- internal
    setFlag( self, IsPressed, flag )
end procedure


-----------------------------------------------------------------------------
global procedure setCheck( integer self, integer flag )
    -- Win32
    setFlag( self, IsChecked, flag )
end procedure

integer growID
-----------------------------------------------------------------------------
-- Added 02/03/2001 Reduces window to titlebar
global procedure minimizeWindow ( integer self )
    sequence image
    integer len,y,x,cell
    if my[self][Grown] then
	my[self][SavedSize2]=my[self][SavedSize]
	my[self][SavedSize]=my[self][Size]
	my[self][WasGrown]=1
	my[self][TrashedSize]=1
	--call_proc(growID,{self})
    end if
    if my[self][Minimized] then
	x=0 y=my[self][Size][2]
	len=my[self][Size][3]
	if my[self][TrashedSize] then
	    my[self][Size]=my[self][SavedSize]
	    my[self][Grown]=my[self][WasGrown]
	    my[self][WasGrown]=0
	else
	    my[self][Size]=my[self][SavedSize]
	end if
	my[self][Minimized]=0
	usedcells[my[self][Cell]]=0
	while (x != my[self][Size][1] and y != my[self][Size][2]) do
	    x += 3 y += 3
	    if x > my[self][Size][1] then x=my[self][Size][1] end if
	    if y > my[self][Size][2] then y=my[self][Size][2] end if
	    draw_line(BLACK,{{x,y},{x+len,y},{x+len,y+16},{x,y+16},{x,y}})
	end while
    else
	image=build_text_image(my[self][Text],0,0,0)
	y=0
	for i = 1 to length(usedcells) do
	    if usedcells[i]=0 then
		cell=i
		y=(i-1)*16 exit
	    end if
	end for
	usedcells[cell]=1
	len=length(image[1])+80
	my[self][Cell]=cell
	my[self][SavedSize]=my[self][Size]
	my[self][Size]={0,y,len,17}
	my[self][Minimized]=1
	y=my[self][SavedSize][2]
	x=my[self][SavedSize][1]
	if not my[self][Grown] then
	    while (x != 0 and y != my[self][Size][2]) do
		x -= 3 y -= 3
		if x < 0 then x=0 end if
		if y < 0 then y=0 end if
		draw_line(BLACK,{{x,y},{x+len,y},{x+len,y+16},{x,y+16},{x,y}})
	    end while
	end if
	my[self][Grown]=0
    end if
    showDesktop()
end procedure
-----------------------------------------------------------------------------
-- Added 02/03/2001 Grows window to full screen
global procedure growWindow ( integer self )
    integer x1,y1,x2,y2,x3,y3,x4,y4
    if my[self][Minimized] and not my[self][TrashedSize] then
	my[self][SavedSize2]=my[self][SavedSize]
	my[self][TrashedSize]=1
	minimizeWindow(self)
    end if
    if my[self][Grown] then
	my[self][Size]=my[self][SavedSize]
	if my[self][TrashedSize] then
	    my[self][Size]=my[self][SavedSize2]
	    my[self][TrashedSize]=0
	end if
	my[self][Grown]=0
	x1=0 x2=xPixels
	y1=0 y2=yPixels
	x3=my[self][Size][1] y3=my[self][Size][2]
	x4=my[self][Size][1]+my[self][Size][3] y4=my[self][Size][2]+my[self][Size][4]
	while (x1 != x3 and y1 != y3 and
	       x2 != x4 and y2 != y4) do
	    x1 += 3
	    y1 += 3
	    x2 -= 3
	    y2 -= 3
	    if (x1 > x3) then x1=x3 end if
	    if (y1 > y3) then y1=y3 end if
	    if (x2 < x4) then x2=x4 end if
	    if (y2 < y4) then y2=y4 end if
	    draw_line(BLACK,{{x1,y1},{x2,y1},{x2,y2},{x1,y2},{x1,y1}})
	end while
    else
	x1=my[self][Size][1]
	y1=my[self][Size][2]
	x2=my[self][Size][1]+my[self][Size][3]
	y2=my[self][Size][2]+my[self][Size][4]
	--x1=floor(xPixels/2)
	--y1=floor(yPixels/2)
	my[self][SavedSize]=my[self][Size]
	my[self][Size]={0,0,xPixels,yPixels}
	my[self][Grown]=1
	while (x1 != 0 and y1 != 0 and x2 != xPixels and y2 != yPixels) do
	    x1 -= 3
	    y1 -= 3
	    x2 += 3
	    y2 += 3
	    if (x1 < 0) then x1=0 end if
	    if (y1 < 0) then y1=0 end if
	    if (x2 > xPixels) then x2=xPixels end if
	    if (y2 > yPixels) then y2=yPixels end if
	    draw_line(BLACK,{{x1,y1},{x2,y1},{x2,y2},{x1,y2},{x1,y1}})
	end while
    end if
    showDesktop()
end procedure
growID=routine_id("growWindow")

-----------------------------------------------------------------------------
-- NEW! 0.3 renamed from unloadWindow to closeWindow
global procedure closeWindow( integer self )

    -- Win32Lib
    if onClose[self] != -1 then
	call_proc(onClose[self],{})
    end if

    -- hide the window
    setVisible( self, 0 )

    if my[self][Cell] > 0 then
	usedcells[my[self][Cell]]=0
    end if
end procedure


-----------------------------------------------------------------------------
global function create( integer kind,
		 sequence title,
		 integer parent,
		 integer x1, integer y1,
		 integer x2, integer y2,
		 atom flags )

    -- Win32Lib
    -- create a control with given attributes

    integer self
    sequence attribute

    if kind=Window and flags=0 then
	flags=WS_MAXIMIZEBOX+WS_MINIMIZEBOX
    end if

    -- hot keys are not currently supported: remove the '&'
    title = remove( '&', title )

    -- create actions
    -- these go first in case any actions are triggered by the code below
    onOpen  = append( onOpen,  -1 )
    onPaint = append( onPaint, -1 )
    onClick = append( onClick, -1 )
    onChange = append( onChange, -1 )
    onMouse = append( onMouse, -1 )
    onRightClick = append( onRightClick, -1 )
    onScroll = append( onScroll, -1 )
-- NEW! 0.14
    onKeyDown = append( onKeyDown, -1 )
    onKeyPress = append( onKeyPress, -1 )
    onGetFocus = append( onGetFocus, -1 )
    onLostFocus = append( onLostFocus, -1 )
    onTimer = append( onTimer, -1 )
    onResize = append( onResize, -1 )
    onClose = append( onClose, -1 )

    -- default?
    if x1 = Default then x1 = floor( (xPixels - x2)/2 )  end if
    if y1 = Default then y1 = floor( (yPixels - y2)/2 )  end if
    if x2 = Default then x2 = 1  end if
    if y2 = Default then y2 = 1  end if

    if kind = List then
	-- size the list. assumes letters are 16 bits high
	y2 = y2 - y1                -- subtract top for height
	y2 = floor( y2/16 ) * 16    -- 16 bits high
	y2 = y2 + y1                -- add offset back in
	y2 = y2 + 4                 -- 2 pixels on top and bottom
    end if

    -- build the values
    attribute = repeat( 0, MySize )
    attribute[Class]        = kind
    attribute[Text]         = title
    attribute[Parent]       = parent
    attribute[Size]         = {x1,y1,x2,y2}
    attribute[Flags]        = flags
    attribute[MenuList]     = {}
    attribute[PenPos]       = {0,0}
    attribute[PenColor]     = sysColor[COLOR_WINDOWTEXT]
    attribute[BackColor]    = sysColor[COLOR_APPWORKSPACE]
    attribute[FontAttrib]   = 0
    attribute[Font]         = SysFont
    attribute[Range]        = {1,100,1,10}
    attribute[State]        = IsVisible+IsEnabled
    attribute[MousePtr]     = IDC_ARROW
    attribute[Focus]        = 0
    attribute[TabStops]     = {}
    attribute[PointerAt]    = 0
    attribute[Minimized]    = 0
    attribute[Contents]     = {}
    if and_bits(flags,WS_DLGFRAME) != WS_DLGFRAME or
       and_bits(flags,WS_THICKFRAME)=WS_THICKFRAME then
	attribute[HasFrame]=1
    else
	attribute[HasFrame]=0
    end if
    if kind=Pixmap then
	attribute[Contents]=repeat(repeat(sysColor[COLOR_WINDOW],x2),y2)
    end if

    -- add to my
    my = append( my, attribute )
    self = length( my )

    -- is this a window?
    if attribute[Class] = Window then
	-- not visible
	setVisible( self, 0 )

    elsif attribute[Class] = Menu then
	-- create a menu
	createMenu( self, parent )

    elsif attribute[Class] = MenuItem then
	-- create a menu item
	createMenuItem( self, parent )

    else
	-- if not a static control
	if not find( kind, StaticControls ) then
	    -- add to tabstops
	    my[parent][TabStops] = append( my[parent][TabStops], self )

	    -- set as focus?
	    if my[parent][Focus] = 0 then
		my[parent][Focus] = self
	    end if

	end if

	if find(attribute[Class],{List,DropDownList}) and length(title) then
	    my[self][FirstItem]=1
	end if

	-- display the control
	if kind != Pixmap and classShow[kind] != -1 then
	    call_proc( classShow[kind], {self} )
	end if


    end if

    -- return position
    return self

end function


-----------------------------------------------------------------------------
procedure putText( integer x, integer y, sequence text,
		    integer fore, integer back, integer attrib )

    -- display
    clipImage( {x,y}, build_text_image( text, fore, back, attrib ) )

end procedure


-----------------------------------------------------------------------------
function getTopWindow( integer self )
    -- Win32
    -- returns parent window, or top window if 0
    -- this will blow up if the window stack is empty.
    -- on the other hand, returning a 0 will cause the caller to choke.
    if self = 0 then
	return windowStack[1]
    else
	return my[self][Parent]
    end if

end function



-----------------------------------------------------------------------------
function getFocus()
    -- Win32Lib
    -- return control with focus
    return my[getTopWindow(0)][Focus]
end function


-----------------------------------------------------------------------------
function hasFocus( integer self )
    -- Win32Lib
    -- true if control has focus
    integer topWin, parent

    if my[self][Class] = Window then
	return (self = getTopWindow(0))
    else
	-- does parent window have focus?
	parent = my[self][Parent]
	if parent = getTopWindow(0) then
	    -- see if control had focus
	    return (self = getFocus())
	else
	    -- no focus if not in top window
	    return 0
	end if
    end if

end function



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Image Class

procedure showToggle( integer self, sequence images )

    -- draw a checkbox or radio button

    integer x1, y1, x2, y2, flags, parent
    integer index, fColor, bColor, active
    sequence image, at

    -- set anchor to parent
    parent = my[self][Parent]

    -- set clipping
    pushClip()
    clipToClient( parent )

    -- get position
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]

    -- get state flags
    flags = my[self][State]

    -- select image
    if isPressed( self ) then
	if isChecked( self ) then
	    index = PressedChecked
	else
	    index = PressedUnchecked
	end if
    elsif isEnabled( self ) then
	if isChecked( self ) then
	    index = ActiveChecked
	else
	    index = ActiveUnchecked
	end if
    else
	if isChecked( self ) then
	    index = InactiveChecked
	else
	    index = InactiveUnchecked
	end if
    end if

    -- build image
    parent = my[self][Parent]
    bColor = my[parent][BackColor]
    image = buildImage( images[index], BLACK, bColor )

    -- erase prior focus rectangle?
    if not hasFocus( self ) then
	-- focus rectangle
	clipRect( bColor, 0, x1, y1, x2, y2 )
    end if

    -- draw it
    at = justifyAt( x1, y1, x2, y2, 0, image )
    clipImage( at, image )

    -- the text color
    fColor = iif( isEnabled( self ), BLACK, sysColor[COLOR_WINDOWTEXT] )

    -- the text
    image = build_text_image( my[self][Text], fColor, bColor, BOLD )
    at = justifyAt( x1+16, y1, x2, y2, 0, image )
    clipImage( at, image )

    -- focus rectangle?
    if hasFocus( self ) then
	-- focus rectangle
	clipFocusRect( bColor, x1, y1, x2, y2 )
    end if

    -- restore
    popClip()

end procedure



-----------------------------------------------------------------------------
function pointInRect( sequence point, sequence rect )
    -- returns true if point is in rectangle

    return ( point[X1] >= rect[X1] and
	     point[Y1] >= rect[Y1] and
	     point[X1] <= rect[X2] and
	     point[Y1] <= rect[Y2] )

end function



-----------------------------------------------------------------------------
function clickGeneric( integer self )

    -- action to take when generic control is clicked
    -- returns { flag, mouseX, mouseY }

    integer inside, flag
    sequence mouse, rect

    -- get control's rectangle
    rect = getWindowRect( self )

    while 1 do
	-- poll mouse
	mouse = poll_mouse()

	-- exit if mouse is released
	if not and_bits( mouse[MOUSE_BUTTONS], MOUSE_LEFT ) then
	    exit
	end if

	-- mouse in the button?
	inside = pointInRect( mouse[MOUSE_X..MOUSE_Y], rect )

	-- need to change state?
	if inside and not isPressed( self ) then
	    -- toggle on
	    setPressed( self, 1 )

	elsif not inside and isPressed( self ) then
	    -- toggle off
	    setPressed( self, 0 )

	end if

    end while

    if isPressed( self ) then
	flag = 1
	setPressed( self, 0 )
    else
	flag = 0
    end if



    return { flag } &  mouse[MOUSE_X..MOUSE_Y]

end function



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Event Handler Tools


-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- LText Class

-----------------------------------------------------------------------------
function breakText( sequence text, sequence delimiter )

    -- break text into segments based on delimiter
    -- ex:  breakText( "hi there\n\how are you?", "\n" )
    --      --> { "hi there", "how are you?" }

    integer breakAt
    sequence segments

    -- clear result
    segments = {}

    while length( text ) do

	-- look for break delimiter
	breakAt = match( delimiter, text )
	if breakAt then
	    -- break into segment
	    segments = append( segments, text[1..breakAt-1] )
	    -- get text after break
	    text = text[breakAt+length(delimiter)..length(text)]
	else
	    -- no delimiters left
	    segments = append( segments, text )
	    -- no text remaining
	    text = ""
	end if

    end while

    return segments

end function

-----------------------------------------------------------------------------
function getTextRect( sequence text, integer attributes )
    -- returns rectangle needed to enclose text
    integer x, y, wide, widest

    -- convert to multiple lines
    text = breakText( text, {13,10} )

    -- get number of lines
    y = length( text ) * get_font_height()

    -- get longest line
    widest = 0
    for i = 1 to length( text ) do
	-- get line's width
	wide = get_text_width( text[i], attributes )
	-- save if widest
	if wide > widest then
	    widest = wide
	end if
    end for

    return { 0, 0, widest, y }

end function



-----------------------------------------------------------------------------
procedure showLText( integer self )

    integer parent, bColor, fColor, breakAt, x, y
    sequence text

    -- get parent
    parent = my[self][Parent]

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( parent )

    -- use parent's background color
    bColor = my[parent][BackColor]
    fColor = BLACK

    -- get position
    x = winX1 + my[self][Size][X1]
    y = winY1 + my[self][Size][Y1]

    -- get text
    text = my[self][Text]

    -- break at cr/lf
    text = breakText( text, {13,10} )

    for i = 1 to length( text ) do
	-- display segment
	putText( x, y, text[i], fColor, bColor, 0 )
	-- move to next line
	y += get_font_height()
    end for

    -- return clip
    popClip()

end procedure


-----------------------------------------------------------------------------
-- assign behaviors
    classShow[ LText ]     = routine_id( "showLText" )



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- RText Class

-----------------------------------------------------------------------------
procedure showRText( integer self )

    integer parent, bColor, fColor, breakAt, x, y, wide
    sequence text, image

    -- get parent
    parent = my[self][Parent]

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( parent )

    -- use parent's background color
    bColor = my[parent][BackColor]
    fColor = BLACK

    -- get position
    wide = my[self][Size][X2]
    x = winX1 + my[self][Size][X1]
    y = winY1 + my[self][Size][Y1]

    -- get text
    text = my[self][Text]

    -- break at cr/lf
    text = breakText( text, {13,10} )

    for i = 1 to length( text ) do
	-- create text
	image = build_text_image( text[i], fColor, bColor, 0 )

	-- display segment
	putText( x+wide-length(image[1]), y, text[i], fColor, bColor, 0 )

	-- move to next line
	y += get_font_height()
    end for

    -- return clip
    popClip()

end procedure


-----------------------------------------------------------------------------
-- assign behaviors
    classShow[ RText ]     = routine_id( "showRText" )

procedure showCText( integer self )

    integer parent, bColor, fColor, breakAt, x, y, wide
    sequence text, image

    -- get parent
    parent = my[self][Parent]

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( parent )

    -- use parent's background color
    bColor = my[parent][BackColor]
    fColor = BLACK

    -- get position
    wide = my[self][Size][X2]
    x = winX1 + my[self][Size][X1]
    y = winY1 + my[self][Size][Y1]

    -- get text
    text = my[self][Text]

    -- break at cr/lf
    text = breakText( text, {13,10} )

    for i = 1 to length( text ) do
	-- create text
	image = build_text_image( text[i], fColor, bColor, 0 )

	-- display segment
	putText( x+floor(wide-length(image[1])*1.8), y, text[i], fColor, bColor, 0 )

	-- increment position
	y += get_font_height()
    end for

    -- return clip
    popClip()

end procedure


-----------------------------------------------------------------------------
-- assign behaviors
    classShow[ CText ]     = routine_id( "showCText" )

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Push Button and Default Push Button Class

-----------------------------------------------------------------------------
procedure showPushButton( integer self )

    -- draw a command button
    integer x1, y1, x2, y2, fColor, bColor, parent, focus, default,clr,bclr
    sequence text, image, at,tmp

    -- get parent
    parent = my[self][Parent]
    bclr=my[parent][BackColor]

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( parent )

    -- get the control's coordinates
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]

    -- get background color
    --bColor = my[parent][BackColor]

    -- focus?
    focus = hasFocus( self )

    -- default?
    default = iif( my[self][Class] = DefPushButton, 1, 0 )

    bColor = sysColor[COLOR_BTNFACE]
    -- focused?
    if focus then
	-- use a blue focus round rectangle around the button
	clr=9
    else
	-- use the background color
	clr=bclr
    end if
    -- Draw the focus rect. The button will be drawn over it
    clipRect(clr,1,x1-2,y1-2,x2+2,y2+2)

    -- "Round out" the rect
    clipPixel(bclr,x1-1,y1-2)
    clipPixel(bclr,x1-2,y1-1)
    clipPixel(bclr,x1-2,y1-2)

    clipPixel(bclr,x2+1,y1-2)
    clipPixel(bclr,x2+2,y1-1)
    clipPixel(bclr,x2+2,y1-2)
    clipPixel(bclr,x2+2,y1-2)

    clipPixel(bclr,x1-1,y2+2)
    clipPixel(bclr,x1-2,y2+1)
    clipPixel(bclr,x1-2,y2+2)
    clipPixel(bclr,x1-2,y2+2)

    clipPixel(bclr,x2+1,y2+2)
    clipPixel(bclr,x2+2,y2+1)
    clipPixel(bclr,x2+2,y2+2)
    clipPixel(bclr,x2+2,y2+2)
    if isPressed( self ) then

	-- erase edge
	clipLine( bColor, x1, y1, x2, y1 )
	clipLine( bColor, x1, y1, x1, y2 )

	clipRect( bColor, 1, x1+1, y1+1, x2, y2 )

	-- title
	text = my[self][Text]

	-- build the text
	fselect( SysFont )
	fColor=BLACK
	image = build_text_image( text, fColor, bColor, BOLD )

	-- display it
	at = justifyAt( x1+2, y1+1, x2, y2, CenterJustify, image )
	clipImage( at, image )

	-- button face
	clipRect( BLACK, 0, x1, y1, x2, y2)
	clipLine( GRAY, x1+2, y1+2, x2-2,y1+2)
	clipLine( GRAY, x1+2, y1+2, x1+2,y2-2)
	clipLine( BRIGHT_WHITE, x1+3, y2-1, x2-2, y2-1)
	clipLine( BRIGHT_WHITE, x2-1, y2-2, x2-1, y1+3)
	clipPixel( BRIGHT_WHITE, x2-2, y2-2)
	clipPixel( GRAY, x1+3, y1+3)

	clipLine( BLACK, x2, y1+1, x2, y2 )
	clipLine( BLACK, x1+1, y2, x2, y2 )

	-- round out button
	clipPixel({bColor,bColor}, x1, y1)
	clipPixel(bColor,x1,y1+1)
	clipPixel(BLACK, x1+1, y1+1)
	clipPixel({bColor,bColor}, x2-1, y1)
	clipPixel(bColor,x2,y1+1)
	clipPixel(BLACK, x2-1, y1+1)
	clipPixel({bColor,bColor}, x1, y2)
	clipPixel(bColor,x1,y2-1)
	clipPixel(BLACK, x1+1, y2-1)
	clipPixel({bColor,bColor}, x2-1, y2)
	clipPixel(bColor,x2,y2-1)
	clipPixel(BLACK, x2-1, y2-1)

    else

	clipRect( bColor, 1, x1+1, y1+1, x2, y2 )

	-- title
	text = my[self][Text]

	-- the text color
	fColor = iif( isEnabled( self ), BLACK, GRAY )

	-- build the text
	fselect( SysFont )
	image = build_text_image( text, fColor, bColor, BOLD )

	-- display it
	at = justifyAt( x1, y1, x2, y2, CenterJustify, image )
	clipImage( at, image )

	clipRect( BLACK, 0, x1, y1, x2, y2)
	clipLine( BRIGHT_WHITE, x1+2, y1+2, x2-2,y1+2)
	clipLine( BRIGHT_WHITE, x1+2, y1+2, x1+2,y2-2)
	clipLine( GRAY, x1+3, y2-1, x2-2, y2-1)
	clipLine( GRAY, x2-1, y2-2, x2-1, y1+3)
	clipPixel( GRAY, x2-2, y2-2)
	clipPixel( BRIGHT_WHITE, x1+3, y1+3)
	--clipFramedRect( bColor, x1, y1, x2-1, y2-1 )

	-- shadow
	fColor = BLACK --sysColor[COLOR_BTNSHADOW]
	clipLine( fColor, x2, y1+1, x2, y2 )
	clipLine( fColor, x1+1, y2, x2, y2 )

	-- extra shadow for default?
	if default then
	    --clipLine( BLACK, x2+1, y1+1, x2+1, y2+1)
	    --clipLine( BLACK, x1+1, y2+1, x2+1, y2+1)

	    --clipRect( BLACK, 0, x1+1, y1+1, x2-2, y2-2 )
	end if

	-- round out button
	clipPixel({bColor,bColor}, x1, y1)
	clipPixel(bColor,x1,y1+1)
	clipPixel(BLACK, x1+1, y1+1)
	clipPixel({bColor,bColor}, x2-1, y1)
	clipPixel(bColor,x2,y1+1)
	clipPixel(BLACK, x2-1, y1+1)
	clipPixel({bColor,bColor}, x1, y2)
	clipPixel(bColor,x1,y2-1)
	clipPixel(BLACK, x1+1, y2-1)
	clipPixel({bColor,bColor}, x2-1, y2)
	clipPixel(bColor,x2,y2-1)
	clipPixel(BLACK, x2-1, y2-1)

	-- focus rect?
	if focus then
	    --if default then
		--clipFocusRect( bColor, x1+2, y1+2, x2-3, y2-3 )
	    --else
		--clipFocusRect( bColor, x1+3, y1+3, x2-3, y2-3 )
	    --end if
	end if
    end if
    -- fill in the little holes in the focus rect...this will have no effect if
    -- the button has no focus
    clipPixel(clr,x1,y1)
    clipPixel(clr,x1+1,y1)
    clipPixel(clr,x1,y1+1)

    clipPixel(clr,x2,y1)
    clipPixel(clr,x2-1,y1)
    clipPixel(clr,x2,y1+1)

    clipPixel(clr,x1,y2)
    clipPixel(clr,x1+1,y2)
    clipPixel(clr,x1,y2-1)

    clipPixel(clr,x2,y2)
    clipPixel(clr,x2-1,y2)
    clipPixel(clr,x2,y2-1)

    -- restore
    popClip()

end procedure




-----------------------------------------------------------------------------
procedure clickPushButton( integer self )
    -- action to take when pushbutton control is clicked
    sequence result

    -- run click
    result = clickGeneric( self )

    -- chosen?
    if result[1] then
	-- action attached to mouse?
	if onClick[self] > -1 then
	    -- do action
	    call_proc( onClick[self], {} )

	end if
    end if

end procedure



-----------------------------------------------------------------------------
procedure keyPushButton( integer self, integer key )

    -- action to take when pushbutton control gets a key

    -- space?
    if key = VK_SPACE then
	-- show as pressed
	setPressed( self, 1 )

	while get_key() != -1 do end while

	-- show as released
	setPressed( self, 0 )

	-- action attached to mouse?
	if onClick[self] > -1 then

	    -- do action
	    call_proc( onClick[self], {} )

	end if
    end if

end procedure



-----------------------------------------------------------------------------
-- assign behaviors

    classShow [ PushButton ]    = routine_id( "showPushButton" )
    classShow[ DefPushButton ]  = routine_id( "showPushButton" )

    classClick[ PushButton ]    = routine_id( "clickPushButton" )
    classClick[ DefPushButton ] = routine_id( "clickPushButton" )
    classKey  [ PushButton ]    = routine_id( "keyPushButton" )
    classKey  [ DefPushButton ] = routine_id( "keyPushButton" )



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Radio Class


sequence radioImages
    radioImages = repeat( {}, 6 )

--  look and feel

    radioImages[ActiveUnchecked] = {
    "    xxxx    ",
    "  xxxxxxxx  ",
    " xxx....xxx ",
    " xx......xx ",
    "xx........xx",
    "xx........xx",
    "xx........xx",
    "xx........xx",
    " xx......xx ",
    " xxx....xxx ",
    "  xxxxxxxx  ",
    "    xxxx    "}

    radioImages[ActiveChecked] = {
    "    xxxx    ",
    "  xxxxxxxx  ",
    " xxx....xxx ",
    " xx......xx ",
    "xx...xx...xx",
    "xx..xxxx..xx",
    "xx..xxxx..xx",
    "xx...xx...xx",
    " xx......xx ",
    " xxx....xxx ",
    "  xxxxxxxx  ",
    "    xxxx    "}

    radioImages[PressedUnchecked] = {
    "    xxxx    ",
    "  xxxxxxxx  ",
    " xxx////xxx ",
    " xx//////xx ",
    "xx////////xx",
    "xx////////xx",
    "xx////////xx",
    "xx////////xx",
    " xx//////xx ",
    " xxx////xxx ",
    "  xxxxxxxx  ",
    "    xxxx    "}

    radioImages[PressedChecked] = {
    "    xxxx    ",
    "  xxxxxxxx  ",
    " xxx////xxx ",
    " xx//////xx ",
    "xx///xx///xx",
    "xx//xxxx//xx",
    "xx//xxxx//xx",
    "xx///xx///xx",
    " xx//////xx ",
    " xxx////xxx ",
    "  xxxxxxxx  ",
    "    xxxx    "}


    radioImages[InactiveUnchecked] = {
    "    XXXX    ",
    "  XXXXXXXX  ",
    " XXX////XXX ",
    " XX//////XX ",
    "XX////////XX",
    "XX////////XX",
    "XX////////XX",
    "XX////////XX",
    " XX//////XX ",
    " XXX////XXX ",
    "  XXXXXXXX  ",
    "    XXXX    "}

    radioImages[InactiveChecked] = {
    "    XXXX    ",
    "  XXXXXXXX  ",
    " XXX////XXX ",
    " XX//////XX ",
    "XX///XX///XX",
    "XX//XXXX//XX",
    "XX//XXXX//XX",
    "XX///XX///XX",
    " XX//////XX ",
    " XXX////XXX ",
    "  XXXXXXXX  ",
    "    XXXX    "}



-----------------------------------------------------------------------------
procedure showRadio( integer self )

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if


    -- draw button
    showToggle( self, radioImages )

end procedure

-----------------------------------------------------------------------------
procedure clickRadio( integer self )
    -- action to take when radio control is pressed
    sequence result

    -- run click
    result = clickGeneric( self )

    -- chosen?
    if result[1] then

	-- turn on radio
	setCheck( self, 1 )

	-- needs to turn off others

	-- action attached to mouse?
	if onClick[self] > -1 then

	    -- do action
	    call_proc( onClick[self], {} )

	end if
    end if

end procedure


-----------------------------------------------------------------------------
procedure keyRadio( integer self, integer key )

    -- action to take when radio button control gets a key

    -- space?
    if key = VK_SPACE then
	-- toggle radio
	setCheck( self, not( isChecked( self ) ) )

	-- needs to turn off others

	-- action attached to mouse?
	if onClick[self] > -1 then

	    -- do action
	    call_proc( onClick[self], {} )

	end if
    end if

end procedure


-----------------------------------------------------------------------------
-- assign behaviors
    classShow[ Radio ]     = routine_id( "showRadio" )
    classClick[ Radio ]    = routine_id( "clickRadio" )
    classKey[ Radio ]       = routine_id( "keyRadio" )







-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Check Box Class

sequence checkImages
    checkImages = repeat( {}, 6 )

    checkImages[ActiveUnchecked] = {
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx",
    "xx.........xx",
    "xx.........xx",
    "xx.........xx",
    "xx.........xx",
    "xx.........xx",
    "xx.........xx",
    "xx.........xx",
    "xx.........xx",
    "xx.........xx",
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx"}

    checkImages[ActiveChecked] = {
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx",
    "xx.........xx",
    "xx.......x.xx",
    "xx......xx.xx",
    "xx.x...xxx.xx",
    "xx.xx.xxx..xx",
    "xx.xxxxx...xx",
    "xx..xxx....xx",
    "xx...x.....xx",
    "xx.........xx",
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx"}

    checkImages[PressedUnchecked] = {
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx",
    "xx/////////xx",
    "xx/////////xx",
    "xx/////////xx",
    "xx/////////xx",
    "xx/////////xx",
    "xx/////////xx",
    "xx/////////xx",
    "xx/////////xx",
    "xx/////////xx",
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx"}

    checkImages[PressedChecked] = {
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx",
    "xx/////////xx",
    "xx///////x/xx",
    "xx//////xx/xx",
    "xx/x///xxx/xx",
    "xx/xx/xxx//xx",
    "xx/xxxxx///xx",
    "xx//xxx////xx",
    "xx///x/////xx",
    "xx/////////xx",
    "xxxxxxxxxxxxx",
    "xxxxxxxxxxxxx"}

    checkImages[InactiveUnchecked] = {
    "XXXXXXXXXXXXX",
    "XXXXXXXXXXXXX",
    "XX/////////XX",
    "XX/////////XX",
    "XX/////////XX",
    "XX/////////XX",
    "XX/////////XX",
    "XX/////////XX",
    "XX/////////XX",
    "XX/////////XX",
    "XX/////////XX",
    "XXXXXXXXXXXXX",
    "XXXXXXXXXXXXX"}

    checkImages[InactiveChecked] = {
    "XXXXXXXXXXXXX",
    "XXXXXXXXXXXXX",
    "XX/////////XX",
    "XX///////X/XX",
    "XX//////XX/XX",
    "XX/X///XXX/XX",
    "XX/XX/XXX//XX",
    "XX/XXXXX///XX",
    "XX//XXX////XX",
    "XX///X/////XX",
    "XX/////////XX",
    "XXXXXXXXXXXXX",
    "XXXXXXXXXXXXX"}

-----------------------------------------------------------------------------
procedure showCheck( integer self )

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if


    showToggle( self, checkImages )



end procedure

-----------------------------------------------------------------------------
procedure clickCheck( integer self )
    sequence result

    -- run click
    result = clickGeneric( self )

    -- chosen?
    if result[1] then

	-- toggle
	setCheck( self, not isChecked( self ) )

	-- action attached to mouse?
	if onClick[self] > -1 then

	    -- do action
	    call_proc( onClick[self], {} )

	end if
    end if
end procedure


-----------------------------------------------------------------------------
procedure keyCheck( integer self, integer key )

    -- action to take when check box control gets a key

    -- space?
    if key = VK_SPACE then

	-- toggle check
	setCheck( self, not( isChecked( self ) ) )

	-- action attached to mouse?
	if onClick[self] > -1 then

	    -- do action
	    call_proc( onClick[self], {} )

	end if
    end if

end procedure


-----------------------------------------------------------------------------
-- assign behaviors
    classShow[ CheckBox ]       = routine_id( "showCheck" )
    classClick[ CheckBox ]      = routine_id( "clickCheck" )
    classKey[ CheckBox ]        = routine_id( "keyCheck" )







-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- HScroll/VScroll Classes


-----------------------------------------------------------------------------
function calcThumbRect( integer self )

    -- returns rectangle containing the thumb
    integer size, wide, at, range
    sequence rect

    rect = getWindowRect(self)

    -- size of scroll area
    if my[self][Class] = HScroll then
	size = rect[X2] - ( rect[X1]+6 )
    else
	size = rect[Y2] - ( rect[Y1]+6 )
    end if

    -- total range of scroll bar
    range = (my[self][Range][ScrollMax]-my[self][Range][ScrollMin])+1

    -- width is relative to large (page) scroll value
    wide = floor( size * ( my[self][Range][ScrollLarge]/range ))

    -- too small?
    if wide < 8 then
	wide = 8
    end if



    -- get relative position
    at = floor((my[self][PointerAt]/range) * (size-wide) ) + 3

    -- what type of scroll bar?
    if my[self][Class] = HScroll then
	-- horizontal
	return { rect[X1]+at, rect[Y1]+3, rect[X1]+at+wide, rect[Y2]-3 }
    else
	-- vertical
	return { rect[X1]+3, rect[Y1]+at, rect[X2]-3, rect[Y1]+at+wide }
    end if

end function



-----------------------------------------------------------------------------
procedure showScroll( integer self )

    -- Redraws the scrollbar, less buttons.
    -- This is the "real time" update code, so the
    -- thumb image is pre-calculated

    integer parent, focus, color, x1, y1, x2, y2
    sequence thumb

    -- get parent
    parent = my[self][Parent]

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( parent )

    -- get the control's coordinates
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]

    ---------------------
    -- draw the scrollbar
    ---------------------

    color = sysColor[COLOR_SCROLLBAR]

    -- draw the channel
    clipFramedRect( color, x1, y1, x2, y2 )

    if hasFocus( self ) then
	-- draw a focus rectangle
	clipFocusRect( color, x1+1, y1+1, x2-1, y2-1 )
    end if

    -----------------
    -- draw the thumb
    -----------------

    thumb = calcThumbRect( self )
    color = sysColor[COLOR_BTNFACE]
    clipShadowedRect( color, thumb[X1], thumb[Y1], thumb[X2], thumb[Y2] )

    -- restore
    popClip()

end procedure


-----------------------------------------------------------------------------
global procedure setScrollPos( integer self, integer pos )

    --$ Win32Lib Emulation
    -- updates the scroll bar to a new value, and redraws it
    -- only the part of the scroll bar that changed is updated

    integer parent, focus, color, x1, y1, x2, y2
    sequence thumb

    -- get parent
    parent = my[self][Parent]

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if

    -- keep in range
    if pos < my[self][Range][ScrollMin]
	then pos = my[self][Range][ScrollMin]
    elsif pos > my[self][Range][ScrollMax] then
	pos = my[self][Range][ScrollMax]
    end if

    -- same as old value?
    if pos = my[self][PointerAt] then
	-- don't update
	return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( parent )

    -- get the control's coordinates
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]


    -- erase the old thumb
    color = sysColor[COLOR_SCROLLBAR]
    thumb = calcThumbRect( self )
    clipRect( color, 1, thumb[X1], thumb[Y1], thumb[X2]+1, thumb[Y2]+1 )

    -- update with new value
    my[self][PointerAt] = pos

    -- draw new thumb
    thumb = calcThumbRect( self )
    color = sysColor[COLOR_BTNFACE]
    clipShadowedRect( color, thumb[X1], thumb[Y1], thumb[X2], thumb[Y2] )

    -- restore
    popClip()

    -- action?
    if onScroll[self] != -1 then
	call_proc( onScroll[self], {my[self][PointerAt]} )
    end if


end procedure


-----------------------------------------------------------------------------
procedure clickScroll( integer self )

    -- handle clicking in scroll bar
    integer wide, range, pos, sMin, sMax, relPos, startPos, thumbWide
    integer mouseAxis, axis1, axis2
    atom ratio
    sequence mouse, thumb, rect, priorMouse

    -- which scroll bar are we talking about?
    if my[self][Class] = HScroll then
	-- use horizontal coordinates
	axis1 = X1
	axis2 = X2
	mouseAxis = MOUSE_X
    else
	-- use vertical coordinates
	axis1 = Y1
	axis2 = Y2
	mouseAxis = MOUSE_Y
    end if

    -- get position of mouse
    mouse = poll_mouse()

    -- get position of thumb
    thumb = calcThumbRect( self )

    -- clicked before thumb?
    if mouse[mouseAxis] < thumb[axis1] then

	-- scroll back by a page
	pos = my[self][PointerAt] - my[self][Range][ScrollLarge]

	-- display
	setScrollPos( self, pos )


    elsif mouse[mouseAxis] > thumb[axis2] then
	-- scroll ahead by a page

	pos = my[self][PointerAt] + my[self][Range][ScrollLarge]

	-- display
	setScrollPos( self, pos )


    else
	-- allow the user to drag the thumb around

	-- get absolute (screen) position of the controls
	rect = getWindowRect( self )

	-- get thumb size
	thumb = calcThumbRect( self )
	thumbWide = thumb[axis2] - thumb[axis1] + 1

	-- range of scroll area
	wide = rect[axis2] - ( rect[axis1]+6+thumbWide )

	-- total range of scroll bar
	range = (my[self][Range][ScrollMax]-my[self][Range][ScrollMin])+1

	-- ratio of movement
	ratio = range/wide

	-- range
	sMin = my[self][Range][ScrollMin]
	sMax = my[self][Range][ScrollMax]

	-- prior position
	priorMouse = {-1,-1,-1}

	---------------------------------------
	-- calculate relative position of mouse
	---------------------------------------

	-- calculate start of thumb, relative to cursor
	relPos = mouse[mouseAxis]

	-- start thumb position
	startPos = my[self][PointerAt]

	-- loop until mouse release
	while and_bits( mouse[MOUSE_BUTTONS], MOUSE_LEFT ) do

	    -- moved?
	    if not compare( mouse[MOUSE_X..MOUSE_Y],
			    priorMouse[MOUSE_X..MOUSE_Y] ) = 0 then

		-- get relative change
		pos = floor( ( (mouse[mouseAxis]-relPos)* ratio ) + startPos)

		-- update
		setScrollPos( self, pos )

		-- save mouse position
		priorMouse = mouse

	    end if

	    -- get the current mouse state
	    mouse = poll_mouse()

	end while

    end if

end procedure
-----------------------------------------------------------------------------
-- assign behaviors
    classShow[ HScroll ]        = routine_id( "showScroll" )
    classClick[ HScroll ]       = routine_id( "clickScroll" )

    classShow[ VScroll ]        = routine_id( "showScroll" )
    classClick[ VScroll ]       = routine_id( "clickScroll" )



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- List Class

-- dummy support routines
-----------------------------------------------------------------------------
global function getIndex( integer self )
    -- no londer a dummy routine
    return my[self][Selected]
end function

-------------------------------------------------------------------------------
global function getCount( integer self )
    -- no longer a dummy routine
    if length( my[self][Text] ) then
	-- accidentally set to string?
	if sequence( my[self][Text][1] ) then
	    return length( my[self][Text] )
	end if
    end if
    return 0
end function

-----------------------------------------------------------------------------
global function getItem( integer self, integer index )
    -- no longer a dummy routine
    if index < 1
    or index > length( my[self][Text] ) then
	return ""
    else
	return my[self][Text][index]
    end if
end function

constant
UpDownArrow = {
    "     x      ",
    "    xxx     ",
    "   xxxxx    ",
    "  xxxxxxx   ",
    "            ",
    "  xxxxxxx   ",
    "   xxxxx    ",
    "    xxx     ",
    "     x      ",
    "            "},

HollowUpArrow = {
    "            ",
    "            ",
    "            ",
    "     xx     ",
    "    xxxx    ",
    "   xxxxxx   ",
    "  xxxxxxxx  ",
    "  xxxxxxxx  ",
    "            ",
    "            "},

FilledUpArrow = {
    "            ",
    "            ",
    "            ",
    "     ..     ",
    "    ....    ",
    "   ......   ",
    "  ........  ",
    "  ........  ",
    "            ",
    "            "},

HollowDownArrow = {
    "            ",
    "            ",
    "            ",
    "  xxxxxxxx  ",
    "  xxxxxxxx  ",
    "   xxxxxx   ",
    "    xxxx    ",
    "     xx     ",
    "            ",
    "            "},

FilledDownArrow = {
    "            ",
    "            ",
    "            ",
    "  ........  ",
    "  ........  ",
    "   ......   ",
    "    ....    ",
    "     ..     ",
    "            ",
    "            "},

HollowLeftArrow = {
    "     x    ",
    "    xx    ",
    "   x x    ",
    "  x  xxxxx",
    " x       x",
    "x        x",
    "x        x",
    " x       x",
    "  x  xxxxx",
    "   x x    ",
    "    xx    ",
    "     x    "},

FilledLeftArrow = {
    "     x    ",
    "    xx    ",
    "   xxx    ",
    "  xxxxxxxx",
    " xxxxxxxxx",
    "xxxxxxxxxx",
    "xxxxxxxxxx",
    " xxxxxxxxx",
    "  xxxxxxxx",
    "   xxx    ",
    "    xx    ",
    "     x    "},

HollowRightArrow = {
    "    x     ",
    "    xx    ",
    "    x x   ",
    "xxxxx  x  ",
    "x       x ",
    "x        x",
    "x        x",
    "x       x ",
    "xxxxx  x  ",
    "    x x   ",
    "    xx    ",
    "    x     "},

FilledRightArrow = {
    "    x     ",
    "    xx    ",
    "    x x   ",
    "xxxxx  x  ",
    "x       x ",
    "x        x",
    "x        x",
    "x       x ",
    "xxxxx  x  ",
    "    x x   ",
    "    xx    ",
    "    x     "}

-- events reported by the scroll bar
-- SE = "scroll event"
constant
    SE_PAGE_UP      = 1,
    SE_PAGE_DOWN    = 2,
    SE_LINE_UP      = 3,
    SE_LINE_DOWN    = 4,
    SE_POSITION     = 5


-----------------------------------------------------------------------------
procedure waitForMouseRelease()
    -- wait until user releases the mouse
    sequence mouseState

    mouseState = poll_mouse()
    while and_bits( mouseState[MOUSE_BUTTONS], MOUSE_LEFT ) do
	mouseState = poll_mouse()
    end while

end procedure




-----------------------------------------------------------------------------
procedure showScrollBar( integer x1, integer y1, integer x2, integer y2,
			    integer range, integer index, integer thumb,
			    integer orient, integer focus, integer pressed )

    -- draws scroll bar. clipping should already be set
    --
    -- x1, y1  : upper coordinates
    -- x2, y2  : lower coordinates
    -- range   : size of range
    -- index   : current position in range
    -- thumb   : pixel position of thumb
    -- orient  : 0 = horizontal, 1 = vertical
    -- focus   : if true, scrollbar has focus
    -- pressed : 0 = none
    --           1 = top/left on
    --           2 = top/left off
    --           3 = bottom/right on
    --           4 = bottom/right off
    --           5 = redraw thumb
    --
    -- if pressed is not zero, then only one arrow button needs
    -- to be drawn. the other drawing will be skipped for speed,
    -- and to prevent the image from flickering

    integer fColor, bColor, fillColor, wide, tall, offset, y3, y4
    sequence arrow, image

    -- set the colors
    bColor = BRIGHT_WHITE --sysColor[COLOR_APPWORKSPACE]
    fColor = sysColor[COLOR_WINDOWTEXT]
    fillColor = sysColor[COLOR_BTNFACE]

    -- draw the border
    if not pressed then
	clipRect( fColor, 0, x1, y1, x2, y2 )
    end if

    -- size of arrow boxes
    if orient = 0 then
	-- horizontal
	wide = (x2 - x1) + 1
	tall = 16
	offset = x1 + floor( tall / 2 )

	-- left arrow
	if pressed = 1 then
	    arrow = FilledLeftArrow
	else
	    arrow = HollowLeftArrow
	end if

	-- calculate metrics
	wide = 16
	tall = (x2 - x1) + 1
	offset = y1 + (wide/2) - (length( arrow )/2)

	-- draw top arrow?
	if find( pressed, {0,1,2} ) then
	    clipFramedRect( WHITE, x1, y1, x1+15, y2 )
	    image=buildImage(arrow,fColor,WHITE) --bColor)
	    clipImage({x1+4,offset},image)
	end if

	if find( pressed, {0,5} ) then

	    -- track above the thumb
	    clipFramedRect( fillColor, x1+16, y1, x1+16+thumb, y2 )

	    -- the thumb
	    clipFramedRect( bColor, x1+15+thumb, y1+1, x1+31+thumb, y2-1 )

	    -- track below the thumb
	    clipFramedRect( fillColor, x1+31+thumb, y1, x2-16, y2 )

	    clipRect(BRIGHT_CYAN,0,x1+1,y1+15+thumb,x2-1,y1+31+thumb)
	    clipLine(CYAN,x2-1,y1+15+thumb,x2-1,y1+31+thumb)
	    clipLine(CYAN,x1+1,y1+31+thumb,x2-1,y1+31+thumb)

	end if

	-- the bottom arrow
	if find( pressed, {0,3,4} ) then
	    -- bottom arrow
	    if pressed = 3 then
		arrow = FilledDownArrow
	    else
		arrow = HollowDownArrow
	    end if

	    clipFramedRect( bColor, x1, y2-15, x2, y2 )
	    image=buildImage(arrow,fColor,WHITE) --bColor)
	    clipImage({offset,y2-11},image)

	    y3=y1+15+thumb+4
	    y4=y1+31+thumb-4
	    for X = x1+4 to x2-6 by 2 do
		clipLine(BRIGHT_CYAN,X,y3,X,y4)
		clipLine(CYAN,X+1,y3,X+1,y4)
	    end for

	end if


    else
	-- vertical

	-- top arrow
	if pressed = 1 then
	    arrow = FilledUpArrow
	else
	    arrow = HollowUpArrow
	end if

	-- calculate metrics
	wide = 16
	tall = (y2 - y1) + 1
	offset = x1 + (wide/2) - (length( arrow )/2)

	-- draw top arrow?
	if find( pressed, {0,1,2} ) then
--            clipFramedRect( WHITE, x1, y1, x2, y1+15 )
	    if pressed = 1 then
		clipRect(0,0,x1,y1,x2,y1+15)
		clipRect(WHITE,1,x1+1,y1+1,x2-1,y1+14)
		clipLine(BRIGHT_WHITE,x2,y1,x2,y1+15)
		clipLine(BRIGHT_WHITE,x1,y1+15,x2,y1+15)
	    else
		clipRect(0,0,x1,y1,x2,y1+15)
		clipRect(WHITE,1,x1+1,y1+1,x2-1,y1+14)
		clipLine(BRIGHT_WHITE,x1,y1,x2,y1)
		clipLine(BRIGHT_WHITE,x1,y1,x1,y1+15)
	    end if
	    image=buildImage(arrow,fColor,WHITE)
	    clipImage({offset,y1+4},image)
	end if

	if find( pressed, {0,5} ) then

	    -- track above the thumb
	    clipFramedRect( fillColor, x1, y1+16, x2, y1+16+thumb )

	    -- the thumb
	    clipFramedRect( BLUE, x1+1, y1+15+thumb, x2-1, y1+31+thumb )

	    -- track below the thumb
	    clipFramedRect( fillColor, x1, y1+31+thumb, x2, y2-16 )

	    clipRect(BRIGHT_CYAN,0,x1+1,y1+15+thumb,x2-1,y1+31+thumb)
	    clipLine(CYAN,x2-1,y1+15+thumb,x2-1,y1+31+thumb)
	    clipLine(CYAN,x1+1,y1+31+thumb,x2-1,y1+31+thumb)
	    for Y = y1+15+thumb+4 to y1+31+thumb-4 by 2 do
		clipLine(BRIGHT_CYAN,x1+3,Y,x2-3,Y)
		clipLine(CYAN,x1+3,Y+1,x2-3,Y+1)
	    end for

	end if

	-- the bottom arrow
	if find( pressed, {0,3,4} ) then
	    -- bottom arrow
	    if pressed = 3 then
		arrow = FilledDownArrow
	    else
		arrow = HollowDownArrow
	    end if

	    --clipFramedRect( WHITE, x1, y2-15, x2, y2 )
	    clipRect(BLACK,0,x1,y2-15,x2,y2)
	    clipRect(WHITE,1,x1+1,y2-14,x2-1,y2-1)
	    image=buildImage(arrow,fColor,WHITE) --bColor)
	    clipImage({offset,y2-11},image)
	    if pressed = 3 then
		clipLine(BRIGHT_WHITE,x2,y2-15,x2,y2-1)
		clipLine(BRIGHT_WHITE,x1,y2-1,x2,y2-1)
	    else
		clipLine(BRIGHT_WHITE,x1,y2-15,x2,y2-15)
		clipLine(BRIGHT_WHITE,x1,y2-15,x1,y2-1)
	    end if

	end if
    end if

end procedure


-----------------------------------------------------------------------------
function clickScrollBar( integer x1, integer y1, integer x2, integer y2,
			    integer range, integer index, integer thumb,
			    integer orient, integer focus )

    -- handle clicks in scrollbar
    -- position of scrollbar is given in screen units
    integer max, offset, current, at
    sequence mouse, prior

    -- get position of mouse
    mouse = poll_mouse()
    mouse[MOUSE_X] -= x1
    mouse[MOUSE_Y] -= y1

    -- what's the orientation?
    if orient = 0 then
	-- horizontal
    else
	-- vertical

	if mouse[MOUSE_Y] < 16 then
	    -- up button pressed
	    showScrollBar( x1, y1, x2, y2, range, index, thumb, orient, focus, 1 )

	    -- wait for button release
	    waitForMouseRelease()

	    -- restore the arrow
	    showScrollBar( x1, y1, x2, y2, range, index, thumb, orient, focus, 2 )

	    -- return event
	    return { SE_LINE_UP }

	elsif mouse[MOUSE_Y] > (y2-y1)-16 then
	    -- down button pressed
	    showScrollBar( x1, y1, x2, y2, range, index, thumb, orient, focus, 3 )

	    -- wait for button release
	    waitForMouseRelease()

	    -- restore the arrow
	    showScrollBar( x1, y1, x2, y2, range, index, thumb, orient, focus, 4 )

	    -- return event
	    return { SE_LINE_DOWN }

	elsif mouse[MOUSE_Y] < 16+thumb then
	    -- clicked above the thumb
	    return { SE_PAGE_UP }

	elsif mouse[MOUSE_Y] > y1+31+thumb then
	    -- clicked below the thumb
	    return { SE_PAGE_DOWN }

	else
	    -- clicked on the thumb
	    max = (y2-y1) - (32+15)
	    offset = mouse[MOUSE_Y] + y1 - thumb
	    while 1 do
		prior = mouse
		mouse = poll_mouse()

		-- mouse released?
		if not and_bits( mouse[MOUSE_BUTTONS], MOUSE_LEFT ) then
		    exit
		end if

		-- mouse moved?
		if mouse[MOUSE_Y] != prior[MOUSE_Y] then
		    at = mouse[MOUSE_Y] - offset
		    if at >= 0
		    and at <= max then
			-- update thumb
			thumb = at
			showScrollBar( x1, y1, x2, y2, range, index, at, orient, focus, 5 )
		    end if
		end if
	    end while
	    return { SE_POSITION, thumb }
	end if
    end if

    return { -1 }

end function






-----------------------------------------------------------------------------
procedure calcListBoxThumb( integer self )
    -- calculate new position for listbox scrollbar thumb
    integer y2

    -- empty list?
    if length( my[self][Text] ) = 0 then
	my[self][PointerAt] = 0
    else
	y2 = my[self][Size][Y2]

	my[self][PointerAt] = floor(
		(my[self][FirstItem] / length( my[self][Text] ) *
		(y2 - 32 )) )

    end if
end procedure
-----------------------------------------------------------------------------
procedure showListBox( integer self )

    integer x1, y1, x2, y2, fColor, bColor,tmp,tmp2,numitems,tmp3
    integer wide, thumb
    sequence image, size, text
    integer color1, color2, hasFocus

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( my[self][Parent] )

    -- height of a letter
    numitems=floor(my[self][Size][4]/get_font_height())

    -- adjust size of control
    my[self][Size][Y2] = (get_font_height() * numitems) + 2

    -- get the control's coordinates
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]
    tmp=y1
    tmp2=y2+2
    wide = my[self][Size][X2] - 15

    -- draw rectangle
    bColor = BRIGHT_WHITE --sysColor[COLOR_APPWORKSPACE]
    fColor = sysColor[COLOR_WINDOWTEXT]
    clipRect( bColor, 1, x1, y1, x2, y2)
    clipRect( fColor, 0, x1, y1, x1+wide-1, y2 )

    -- position
    x1 = x1+4
    y1 = y1 + 3

    -- has focus?
    hasFocus = (my[ my[self][Parent] ][Focus] = self)

    -- populate list
    for i = my[self][FirstItem] to my[self][FirstItem]+numitems-1 do

	-- past end of items or empty list?
	if i > length( my[self][Text] )
	or length( my[self][Text] ) = 0
	or integer( my[self][Text][i] ) then
	    -- empty space
	    text = ""
	else
	    -- get item
	    text = my[self][Text][i]
	end if


	-- selected?
	if my[self][Selected]=i then
	    -- invert colors
	    color1 = bColor
	    color2 = fColor
	else
	    -- normal colors
	    color1 = fColor
	    color2 = bColor
	end if

	-- build text image
	fselect( SysFont )
	image = build_text_image(text, color1, color2, 0 )


	-- selected?
	if my[self][Selected]=i then
	    -- inverted colors
	    color1 = fColor
	    color2 = bColor
	else
	    -- normal colors
	    color1 = bColor
	    color2 = fColor
	end if

	-- pad the image to the right side of the list, less scroll area
	if length( image[1] )-5 > wide then
	    -- clip
	    for j = 1 to length(image) do
		image[j] = image[j][1..wide-5]
	    end for
	else
	    -- pad
	    for j = 1 to length(image) do
		image[j] &= repeat(color1,wide-length(image[j])-5)
	    end for
	end if

	-- selected item or no selected item?
	if my[self][Selected]=i
	or ( my[self][Selected]=0
	 and i = my[self][FirstItem]
	 and hasFocus ) then

	    if hasFocus then
		-- top and bottom of focus rectangle
		for k = 1 to length(image[1]) by 2 do
		    image[1][k]=color2
		    image[length(image)][k]=color2
		end for

		-- left and right of focus rectangle
		for k = 1 to get_font_height() by 2 do
		    image[k][1]=color2
		    image[k][length(image[k])]=color2
		end for
	    end if
	end if

	-- clip to window
	clipImage( {x1, y1}, image )

	-- move down 1 line
	y1 = y1 + length( image )
	if y1 > y2 then
	    exit
	end if

    end for

    -- scrollbar
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]

    thumb = my[self][PointerAt]
    showScrollBar( x2-16, y1, x2, y2, 0, 0, thumb, 1, 0, 0 )

    -- restore
    popClip()


end procedure

procedure clickListBox(integer self)
    integer x1, y1, x2, y2, fColor, bColor,mousex,mousey,index,tmp2,tmp1,tmp,
	    tmp3,scrollbar,numitems,height,myY1,at
    sequence image, size, mouse, result, event

    if length(my[self][Text])=0 then -- no items to click on. ignore it.
	return
    end if

    -- get focus
    my[self][Focus]=1

    -- set anchor to parent
    pushClip()
    clipToClient( my[self][Parent] )

    -- get the control's coordinates
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]

    numitems=floor(my[self][Size][4]/get_font_height())
    index = my[self][Selected]

    -- get mouse position relative to control
    mouse = poll_mouse()
    mouse[MOUSE_X] -= x1
    mouse[MOUSE_Y] -= y1

    if mouse[MOUSE_X] > (x2-x1) - 16 then
	-- clicked in scrollbar
	event = clickScrollBar( x2-16, y1, x2, y2, length(my[self][Text]),
				index, my[self][PointerAt], 1, 0 )

	-- process event
	if event[1] = SE_PAGE_UP then
	elsif event[1] = SE_PAGE_DOWN then

	elsif event[1] = SE_LINE_UP then
	    if my[self][FirstItem] > 1 then
		my[self][FirstItem] -= 1
		calcListBoxThumb( self )
	    end if

	elsif event[1] = SE_LINE_DOWN then
	    if my[self][FirstItem] + numitems < length( my[self][Text] ) then
		my[self][FirstItem] += 1
		calcListBoxThumb( self )
	    end if

	elsif event[1] = SE_POSITION then
	    -- change position in scrollbar

	    -- update thumb position
	    my[self][PointerAt] = event[2]
	    -- calculate new first item
	    at = floor( (event[2]/(my[self][Size][Y2]-32))
		    * length(my[self][Text]) )

	    my[self][FirstItem] = at - numitems
	    if my[self][FirstItem] < 1 then
		my[self][FirstItem] = 1
	    end if
	end if

	-- update list box
	showListBox( self )
    else
	-- clicked in list

	index=0
	height = get_font_height()
	for i = my[self][FirstItem] to my[self][FirstItem]+numitems-1 do
	    if mouse[MOUSE_Y] <= height then
		index=i
		exit
	    end if
	    mouse[MOUSE_Y] -= height
	end for

	-- if selection isn't past the end of the list items...
	if length( my[self][Text] )
	and index <= length( my[self][Text] ) then
	    -- set as selected item
	    my[self][Selected]=index
	    calcListBoxThumb( self )
	end if

	-- call display routine
	showListBox( self )

	-- run click
	result = clickGeneric( self )

	-- chosen?
	if result[1] then
	    -- action attached to mouse?
	    if onChange[self] > -1 then
		-- do action
		call_proc( onChange[self], {} )

	    end if
	end if


    end if

    popClip()

end procedure

procedure keyListBox(integer self, integer key)
    integer x1, y1, x2, y2, scrollbar,numItems, listsize
    sequence image, size,mouse,result

    -- empty?
    if getCount( self ) = 0 then
	return
    end if

    -- get the control's coordinates
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]
    --tmp1=y1-4
    --tmp2=y2+2
    numItems=floor(my[self][Size][4]/get_font_height())

    -- up/down
    if key = 328
    or key = 336 then

	-- is the item visible?
	if my[self][Selected] < my[self][FirstItem]
	or my[self][Selected] > my[self][FirstItem] + numItems - 1 then
	    -- partial list?
	    if getCount( self ) < numItems then
		my[self][FirstItem] = 1

	    -- no item selected?
	    elsif my[self][Selected] = 0 then
		-- use top of list
		my[self][FirstItem] = 1

	    else
		-- move selected item to top of list
		my[self][FirstItem] = my[self][Selected]
	    end if
	end if

	-- up
	if key = 328 then
	    if my[self][Selected] < 2 then
		my[self][Selected] = 1
	    else
		my[self][Selected] -= 1
	    end if

	    -- past top of list?
	    if my[self][Selected] < my[self][FirstItem] then
		my[self][FirstItem] = my[self][Selected]
	    end if

	-- down?
	elsif key = 336 then
	    if my[self][Selected] < getCount( self ) then
		my[self][Selected] += 1

		if my[self][Selected]-my[self][FirstItem] = numItems then
		    my[self][FirstItem] += 1
		end if

	    end if


	end if

	calcListBoxThumb( self )
	showListBox(self)

    end if

end procedure
-----------------------------------------------------------------------------
-- assign behaviors
    classShow[ List ]       = routine_id( "showListBox" )
    classClick[ List ]      = routine_id( "clickListBox" )
    classKey[ List ]      = routine_id( "keyListBox" )






-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- EditText Class

-----------------------------------------------------------------------------
procedure showMleText( integer self )

    integer x1, y1, x2, y2, color
    sequence text, image, at

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( my[self][Parent] )

    -- get the control's coordinates
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]

    -- background color
    clipFramedRect( BRIGHT_WHITE, x1, y1, x2, y2 )

    -- focus?
    if hasFocus( self ) then
	-- draw focus rectangle
	clipFocusRect( BRIGHT_WHITE, x1+1, y1+1, x2-1, y2-1 )
    end if

    -- get text
    text = my[self][Text]
    if hasFocus( self ) then
	text = text & "|"
    end if

    -- generate text image
    image = build_text_image( text, BLACK, BRIGHT_WHITE, 0 )

    -- justify and draw
    at = justifyAt( x1+8, y1, x2, y2, LeftJustify, image )
    clipImage( at, image )

    -- restore
    popClip()

end procedure


-----------------------------------------------------------------------------
procedure keyMleText( integer self, integer key )

    sequence text

    -- get the data
    text = my[self][Text]

    -- process the keystroke
    if key = VK_TAB then
	-- this should not be added to the field

    elsif key = VK_BACK then
	-- any room to backspace?
	if length( text ) then
	    text = text[1..length(text)-1]
	end if
    else
	-- is there room?
	if get_text_width( text, 0 ) + 16 < my[self][Size][X2] then
	    -- only allow char < 256 (or font routine will blow up)
	    if key > 0 and key < 256 then
		-- add to text
		text = text & key
	    end if
	end if
    end if

    -- update
    my[self][Text] = text

    -- draw the text
    showMleText( self )

end procedure


-----------------------------------------------------------------------------
-- assign behaviors
    classShow[ MleText ]   = routine_id( "showMleText" )
    classKey[ MleText ]    = routine_id( "keyMleText" )



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- EditText Class

-----------------------------------------------------------------------------
procedure showEditText( integer self )

    integer x1, y1, x2, y2, color, clr, bclr
    sequence text, image, at, tmp

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if

    bclr=my[my[self][Parent]][BackColor]

    -- set anchor to parent
    pushClip()
    clipToClient( my[self][Parent] )

    -- get the control's coordinates
    x1 = winX1 + my[self][Size][X1]
    y1 = winY1 + my[self][Size][Y1]
    x2 = x1 + my[self][Size][X2]
    y2 = y1 + my[self][Size][Y2]

    -- focus?
    if hasFocus( self ) then
	-- draw focus rectangle
	clr=9
    else
	clr=bclr
    end if
    clipRect(clr,1,x1-2,y1-2,x2+2,y2+2)

    -- "Round out" the rect
    clipPixel(bclr,x1-1,y1-2)
    clipPixel(bclr,x1-2,y1-1)
    clipPixel(bclr,x1-2,y1-2)

    clipPixel(bclr,x2+1,y1-2)
    clipPixel(bclr,x2+2,y1-1)
    clipPixel(bclr,x2+2,y1-2)
    clipPixel(bclr,x2+2,y1-2)

    clipPixel(bclr,x1-1,y2+2)
    clipPixel(bclr,x1-2,y2+1)
    clipPixel(bclr,x1-2,y2+2)
    clipPixel(bclr,x1-2,y2+2)

    clipPixel(bclr,x2+1,y2+2)
    clipPixel(bclr,x2+2,y2+1)
    clipPixel(bclr,x2+2,y2+2)
    clipPixel(bclr,x2+2,y2+2)

    -- background color
    clipFramedRect( BRIGHT_WHITE, x1, y1, x2, y2 )

    -- get text
    text = my[self][Text]
    if hasFocus( self ) then
	text = text & "|"
    end if

    -- generate text image
    image = build_text_image( text, BLACK, BRIGHT_WHITE, 0 )

    -- justify and draw
    at = justifyAt( x1+8, y1, x2, y2, LeftJustify, image )
    clipImage( at, image )

    -- fill in the little holes in the focus rect...this will have no effect if
    -- the button has no focus
    clipPixel(clr,x1,y1)
    clipPixel(clr,x1+1,y1)
    clipPixel(clr,x1,y1+1)

    clipPixel(clr,x2,y1)
    clipPixel(clr,x2-1,y1)
    clipPixel(clr,x2,y1+1)

    clipPixel(clr,x1,y2)
    clipPixel(clr,x1+1,y2)
    clipPixel(clr,x1,y2-1)

    clipPixel(clr,x2,y2)
    clipPixel(clr,x2-1,y2)
    clipPixel(clr,x2,y2-1)

    -- restore
    popClip()

end procedure


-----------------------------------------------------------------------------
procedure keyEditText( integer self, integer key )

    sequence text

    -- get the data
    text = my[self][Text]

    -- process the keystroke
    if key = VK_TAB then
	-- this should not be added to the field

    elsif key = VK_BACK then
	-- any room to backspace?
	if length( text ) then
	    text = text[1..length(text)-1]
	end if
    else
	-- is there room?
	if get_text_width( text, 0 ) + 16 < my[self][Size][X2] then
	    -- only allow char < 256 (or font routine will blow up)
	    if key > 0 and key < 256 then
		-- add to text
		text = text & key
	    end if
	end if
    end if

    -- update
    my[self][Text] = text

    -- draw the text
    showEditText( self )

end procedure


-----------------------------------------------------------------------------
-- assign behaviors
    classShow[ EditText ]   = routine_id( "showEditText" )
    classKey[ EditText ]    = routine_id( "keyEditText" )



-----------------------------------------------------------------------------
procedure repaintWin( integer self, integer x1, integer y1,
				    integer x2, integer y2 )

    -- redraw the workspace within the clipped area
    integer color

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if

    -- set anchor to client area
    pushClip()
    clipToClient( self )

    -- alter the clipping window to requested clip window
    clipWin = { max(winX1+x1,winX1),
		max(winY1+y1,winY1),
		min(winX1+x2,winX2),
		min(winY1+y2,winY2) }

    -- hide the mouse
    mouse_pointer( 0 )

    -- clear client area
    color = my[self][BackColor]
    clipRect( color, 1, clipWin[X1], clipWin[Y1], clipWin[X2], clipWin[Y2] )

    -- on paint
    if onPaint[self] > -1 then

	-- run routine
	call_proc( onPaint[self], {x1,y1,x2,y2} )

    end if

    -- draw controls
    for i = 1 to length( my ) do
	-- parent of non-window control?
	if  my[i][Parent] = self
	and my[i][Class] != Window then
	    -- draw it
	    showControl( i )
	end if
    end for

    -- show the mouse
    mouse_pointer( 1 )

    -- restore clipping
    popClip()

end procedure


-----------------------------------------------------------------------------
global procedure repaintWindow( integer self )

    -- Win32Lib
    -- redraw the client area
    sequence point

    -- get extent
    point = getClientSize( self )

    -- repaint the entire window
    repaintWin( self, 0, 0, point[X1], point[Y1] )

end procedure


-----------------------------------------------------------------------------
global procedure repaintRect( integer self, integer x1, integer y1,
				integer x2, integer y2 )

    -- Win32Lib
    -- redraw the workspace within range

    -- repaint only requested portion of window
    repaintWin( self, x1, y1, x2, y2 )

end procedure



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Bitmap class

sequence resource
resource = {}


-----------------------------------------------------------------------------
procedure showBitmap( integer self )

    -- draw a bitmap control
    integer parent
    sequence rect

    -- get parent
    parent = my[self][Parent]

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if

    -- no bitmap?
    if my[self][Bmp] = 0 then
	return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( parent )

    -- get screen position
    rect = getWindowRect( self )

    -- display the bitmap resource
    clipImage( rect[X1..Y1], resource[my[self][Bmp]] )

    -- restore
    popClip()

end procedure



------------------------------------------------------------------------------
function convertToPalette( sequence sourcePal )

    -- convert the color list to palette indexes

    atom err, diff, best
    sequence destPal, colorIndex

    -- get palette
    destPal = get_all_palette()

    -- adjust source
    sourcePal = floor( sourcePal/4 )

    -- build holder for results
    colorIndex = repeat( 0, length( sourcePal ) )

    -- look at each entry
    for i = 1 to length( sourcePal ) do
	-- set the last best match as impossibly bad
	best = #FFFFFF
	-- look through the palette
	for entry = 1 to length( destPal ) do
	    -- error amount for this entry
	    err = 0
	    -- for each {rgb}
	    for rgb = 1 to 3 do
		-- error is difference between the two
		diff = abs( sourcePal[i][rgb] - destPal[entry][rgb] )
		err = err + diff
	    end for

	    -- better match than prior best match?
	    if err < best then
		-- save error amount
		best = err
		-- store the palette index
		colorIndex[i] = entry
	    end if
	end for
    end for

    return colorIndex

end function


------------------------------------------------------------------------------
global function rgb( integer r, integer g, integer b )
    -- return index of closest match in palette
    sequence s
    s = convertToPalette( {r,g,b} )
    return s[1]
end function
rgbFunc=routine_id("rgb")


-----------------------------------------------------------------------------
global function createDIB( sequence pixels, sequence pal )
    -- Win32Lib
    -- create bitmap, return resource id
    integer index
    sequence replace

    -- adjust the current palette
    replace = convertToPalette(pal)

    -- replace values with references to new palette
    for i = 1 to length( pixels ) do
	for j = 1 to length( pixels[i] ) do
	    -- get index
	    index = pixels[i][j]
	    -- look up replacement
	    index = replace[index+1]-1
	    -- replace
	    pixels[i][j] = index
	end for
    end for

    -- add to resource list
    resource = append( resource, pixels )

    -- return index in resource list
    return length( resource )

end function


-----------------------------------------------------------------------------
global function loadBitmapFromFile( sequence fName )
    -- Win32Lib
    -- load bitmap, return resource id
    object res

    -- attempt to read bitmap
    res = read_bitmap( fName )
    if integer( res ) then
	abortErr( "Unable to read bitmap %s", {fName}  )
    end if

    return createDIB( res[2], res[1] )

end function


-----------------------------------------------------------------------------
global procedure setBitmap( integer self, object bitmap )
    -- load bitmap
    if sequence( bitmap ) then
	-- assume file name
	my[self][Bmp] = loadBitmapFromFile( bitmap )
    else
	-- resource id
	my[self][Bmp] = bitmap
    end if

    -- update the control
    showBitmap( self )

end procedure


-----------------------------------------------------------------------------
global procedure drawBitmap( integer self, integer resId,
				    integer x, integer y )
    -- $Win32Lib

    -- anchor to window
    pushClip()
    clipToClient( self )

    -- draw the bitmap
    clipImage( {winX1+x+1,winY1+y+1}, resource[ resId ] )

    -- restore
    popClip()

end procedure

-----------------------------------------------------------------------------
-- NEW! 0.4 renamed from createBitmap to textToBitmap
global function textToBitmap( sequence s )
    -- create a bitmap, adding it to the resource list
    sequence res

    -- convert to colors
    for i = 1 to length( s ) do
	for j = 1 to length( s[i] ) do
	    if    s[i][j] = 'x' then s[i][j] = BLACK
	    else
		-- color of client area
		s[i][j] = sysColor[COLOR_APPWORKSPACE]
	    end if
	end for
    end for

    -- create a resource
    resource = append( resource, s )

    -- return index to the resource
    return length( resource )

end function

-----------------------------------------------------------------------------
-- assign behaviors

    classShow [ Bitmap ]     = routine_id( "showBitmap" )






-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Icon class

-----------------------------------------------------------------------------
procedure showIcon( integer self )

    -- draw an icon control
    -- this is a stub, since I don't know how to read icons
    sequence rect

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if

    -- set anchor to parent
    pushClip()
    clipToClient( my[self][Parent] )

    -- get size
    rect = getWindowRect( self )

    -- place rectangle as marker
    clipFramedRect( BRIGHT_WHITE, rect[X1], rect[Y1], rect[X2], rect[Y2] )
    clipLine( BLACK, rect[X1], rect[Y1], rect[X2], rect[Y2] )
    clipLine( BLACK, rect[X2], rect[Y1], rect[X1], rect[Y2] )

    -- restore
    popClip()

end procedure


-----------------------------------------------------------------------------
global procedure setIcon( integer id, object icon )
    -- do-nothing stub
end procedure


-----------------------------------------------------------------------------
-- assign behaviors

    classShow [ Icon ]     = routine_id( "showIcon" )






-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Graphics Support


-----------------------------------------------------------------------------
-- NEW! 0.4 added fill parameter
global procedure drawRectangle( integer self, integer fill,
				    integer x1, integer y1,
				    integer x2, integer y2 )

    -- anchor to window
    pushClip()
    clipToClient( self )

    -- adjust to window
    x1 = winX1 + x1 - 1
    y1 = winY1 + y1 - 1
    x2 = winX1 + x2 - 1
    y2 = winY1 + y2 - 1

    -- draw the rectangle
    clipRect( my[self][PenColor], fill, x1, y1, x2, y2 )

    -- restore
    popClip()

end procedure



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- MenuItem class



-----------------------------------------------------------------------------
-- graphics

constant
    MenuCheck = {
    "      x",
    "     xx",
    "xx  xx ",
    "xx xx  ",
    "xxxx   ",
    "xxx    ",
    "xx     "}



-----------------------------------------------------------------------------
procedure sizeDropDown( integer self, integer x1, integer y1 )

    -- assign position of the menu drop down items based on
    -- the position of the menu bar

    integer     item, fontHigh, menuWide, menuHigh, width, half, x2, y2
    sequence    menuList

    -- get the menu items list
    menuList = my[self][MenuList]

    -- select the system font
    fselect( SysFont )

    -- get the height of the font
    fontHigh = get_font_height()
    half = floor( fontHigh/2 )

    -- add border to menu
    y1 = y1 + 2

    -- find the size of the menu
    menuWide = 0
    menuHigh = 0
    for i = 1 to length( menuList ) do

	-- get an item
	item = menuList[i]
	-- get the width
	width = get_text_width( my[item][Text], BOLD )
	-- save if largest encountered so far
	if width > menuWide then
	    menuWide = width
	end if

	-- seperators only take up half space
	if compare( my[item][Text], "-" ) = 0 then
	    y2 = y1 + half + 2
	else
	    y2 = y1 + fontHigh + 2
	end if

	-- store values
	my[item][Size] = {x1, y1, 0, y2 }

	-- move ahead
	y1 = y2

    end for

    -- add in room for check box and edges
    x2 = x1 + menuWide + 20 + 4

    -- replace x2 value in list
    for i = 1 to length( menuList ) do
	-- get id
	item = menuList[i]
	-- update position
	my[item][Size][X2] = x2
    end for

end procedure


-----------------------------------------------------------------------------
procedure showMenuItem( integer self, integer pick, integer pressed )

    -- draw a menu item in a dropdown list
    -- the routine sizeDropDown must be called first

    integer     item, x1, y1, x2, y2, fColor, bColor, dim, half, sep
    sequence    text, image, menuList, point

    -- skip if index is zero
    if pick = 0 then
	return
    end if

    -- get the menu items list
    menuList = my[self][MenuList]

    -- get the item
    item = my[self][MenuList][pick]

    -- get attributes
    text = my[item][Text]
    x1   = my[item][Size][X1]
    y1   = my[item][Size][Y1]
    x2   = my[item][Size][X2]
    y2   = my[item][Size][Y2]

    -- is it dimmed?
    dim = not isEnabled( item )

    -- is it a seperator?
    sep = not compare( text, "-" )

    -- can't select if dim of divider
    if dim or sep then
	pressed = 0
    end if

    -- select color
    if dim then
	bColor = sysColor[COLOR_MENU]
	fColor = sysColor[COLOR_GRAYTEXT]
    elsif pressed then
	bColor = sysColor[COLOR_HIGHLIGHT]
	fColor = sysColor[COLOR_HIGHLIGHTTEXT]
    else
	bColor = sysColor[COLOR_MENU]
	fColor = sysColor[COLOR_MENUTEXT]
    end if

    -- handle seperator first
    if sep then

	-- mark goes at halfway point
	half = floor( (y2 - y1)/2 )

	-- highlight below
	clipLine( BLACK, x1+1,y1+half,x2-1,y1+half )
	clipLine( BRIGHT_WHITE, x1+1,y1+half+1,x2-1,y1+half+1 )

	-- exit. simplifies logic.
	return

    end if

    -- repaint the background
    clipRect( bColor, 1, x1+1, y1, x2-1, y2 )

    -- is it checked?
    if isChecked( item ) then

	-- build a check
	image = buildImage( MenuCheck, fColor, bColor )

	-- center against the text
	point = justifyAt( x1+4, y1, x2, y2, 0, image )

	-- display it
	clipImage( point, image )

    end if

    -- build the text
    image = build_text_image( text, fColor, bColor, BOLD )

    -- center against the text
    point = justifyAt( x1+20, y1, x2, y2, 0, image )

    -- display the text
    clipImage( point, image )

end procedure


-----------------------------------------------------------------------------
function showDropDown( integer self, integer x1, integer y1 )

    -- draw a drop-down menu, return area covered by menu

    integer x2, y2, last
    sequence menuList, covered

    -- calculate positions of items in dropdown
    sizeDropDown( self, x1, y1 )

    -- get the items in the dropdown
    menuList = my[self][MenuList]

    -- get the last item in the menu list
    last = menuList[length( menuList )]

    -- calculate the dropdown size based on the item's position
    x2 = my[last][Size][X2]
    y2 = my[last][Size][Y2]+2

    -- save the covered area; account for shadow
    covered = saveImage( x1, y1, x2+1, y2+1 )

    -- backdrop of the dropdown
--    clipShadowedRect( sysColor[COLOR_MENU], x1, y1, x2, y2 )
    clipRect( WHITE, 1, x1,y1,x2,y2)
    clipRect( BRIGHT_WHITE, 0, x1, y1, x2, y2)
    clipLine( BLACK,x2,y1,x2,y2)
    clipLine( BLACK,x2,y2,x1,y2)
    clipLine( BLACK,x2+1,y1+1,x2+1,y2+1)
    clipLine( BLACK,x1+1,y2+1,x2+1,y2+1)

    -- populate dropdown
    for i = 1 to length( my[self][MenuList] ) do
	showMenuItem( self, i, 0 )
    end for

    return covered

end function



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Menu class





-----------------------------------------------------------------------------
procedure showMenuBarItem( integer self, integer pick, integer pressed )

    -- draw a menu in a window
    integer item, x1, y1, x2, y2, fColor, bColor
    sequence image, point

    -- skip if zero
    if pick = 0 then
	return
    end if

    -- select the system font
    fselect( SysFont )


    if hasFocus( self ) then
	fColor = iif( pressed,  sysColor[COLOR_HIGHLIGHTTEXT],
				sysColor[COLOR_MENUTEXT] )
	bColor = iif( pressed,  sysColor[COLOR_HIGHLIGHT],
				sysColor[COLOR_MENU] )
    else
	fColor = sysColor[COLOR_INACTIVECAPTIONTEXT]
	bColor = sysColor[COLOR_INACTIVECAPTION]

    end if

    -- build text
    item = my[self][MenuList][pick]
    image = build_text_image( my[item][Text], fColor, bColor, Bold )

    -- get position
    x1 = my[item][Size][X1]
    y1 = my[item][Size][Y1]
    x2 = my[item][Size][X2]
    y2 = my[item][Size][Y2]

    -- draw box around it
    clipRect( bColor, 1, x1, y1, x2, y2-1 )

    -- center against the text
    point = justifyAt( x1+4, y1+1, x2, y2-1, 0, image )

    -- display text
    clipImage( point, image )

end procedure


-----------------------------------------------------------------------------
procedure sizeMenuBar( integer self )

    -- calculate the position of the menu bar text
    -- in the current window

    integer x1, y1, x2, y2, at, item
    sequence text, menuList

    -- get list of items
    menuList = my[self][MenuList]

    -- optional menu
    if length( menuList ) then

	-- get window position
	x1 = my[self][Size][X1]
	y1 = my[self][Size][Y1]

	-- menu position
	x1 = x1
	y1 = y1+theTitleSize+1
	y2 = y1+theMenuSize-2

	-- start column
	x1 = x1 + 4

	-- build text
	fselect( SysFont )

	-- examine each item
	for i = 1 to length( menuList ) do

	    -- get the id of the menu
	    item = menuList[i]

	    -- get text
	    text = my[item][Text]

	    -- get width
	    x2 = x1 + get_text_width( text, Bold ) + 8

	    -- store menu item position
	    my[item][Size] = {x1,y1,x2,y2}

	    -- move ahead
	    x1 = x2 + 4

	end for

    end if

end procedure


-----------------------------------------------------------------------------
procedure showMenuBar( integer self, integer pick )

    -- draw a menu in a window

    integer x1, y1, x2, y2, at, color, item
    sequence menuList, image, text

    -- size the items in the menu bar
    sizeMenuBar( self )

    -- get the menu items list
    menuList = my[self][MenuList]

    -- menu?
    if length( menuList ) then

	-- get menu position
	x1 = my[self][Size][X1]
	y1 = my[self][Size][Y1]+theTitleSize
	x2 = my[self][Size][X2]+x1
	y2 = y1 + theMenuSize - 1

	-- menu bar color
	if hasFocus( self ) then
	    color = sysColor[COLOR_MENU]
	else
	    color = sysColor[COLOR_INACTIVECAPTION]
	end if
	clipFramedRect( color, x1, y1, x2, y2 )

	-- draw each menu item
	for i = 1 to length( menuList ) do
	    -- draw the item
	    showMenuBarItem( self, i, pick=i )
	end for
    end if

end procedure


-----------------------------------------------------------------------------
procedure runMenu( integer self, sequence mouseState )
    -- run the  menu
    -- point passed is relative to screen.

    integer pick, item, menuIndex, menuId, dropIndex, dropId, x, y,
	    mouseDown, clicked
    sequence priorState, point, covered
    atom t

    clicked=0
    -- check the mouse: get the absolute position
    priorState = mouseState

    -- calculate the positions of items in the menu
    sizeMenuBar( self )

    -- get the point
    point = mouseState[MOUSE_X..MOUSE_Y]

    -- see if any menu bar items were hit
    menuIndex = 0
    for i = 1 to length( my[self][MenuList] ) do

	-- get an item
	item = my[self][MenuList][i]

	-- does the point fall in the item
	if pointInRect( point, my[item][Size] ) then
	    -- set as index, leave loop
	    menuIndex = i
	    exit
	end if
    end for

    -- didn't click on any menu items?
    if menuIndex = 0 then
	return
    end if

    showMenuBarItem( self, menuIndex, 1 )
    while 1 do
	mouseState=poll_mouse()
	if and_bits(mouseState[1],MOUSE_LEFT) != MOUSE_LEFT then
	    exit
	end if
    end while

    -- loop until the mouse is released
    mouseDown = 1
    while not clicked do

	menuId = my[self][MenuList][menuIndex]

	-- highlight selected menu item
	showMenuBarItem( self, menuIndex, 1 )

	-- get the attributes of the dropdown
	dropId      = my[self][MenuList][menuIndex]
	dropIndex   = 0
	x           = my[menuId][Size][X1]
	y           = my[menuId][Size][Y2]

	-- draw the dropdown, get covered area
	covered = showDropDown( dropId, x, y )

	-- loop until mouse released or another menu selected
	while 1 do

	    -- check the rat
	    mouseState = poll_mouse()

	    -- mouse state changed?
	    if compare( mouseState, priorState ) then

		-- save state
		priorState = mouseState

		-- mouse release?
		if and_bits( mouseState[MOUSE_BUTTONS], MOUSE_LEFT )=MOUSE_LEFT then
		    while 1 do
			mouseState=poll_mouse()
			if and_bits(mouseState[MOUSE_BUTTONS],MOUSE_LEFT) != MOUSE_LEFT then
			    exit
			end if
		    end while
		    -- set flag
		    clicked=1
		    mouseDown = 0
		    -- leave the loop
		    exit
		end if

		-- get the point
		point = mouseState[MOUSE_X..MOUSE_Y]

		-------------------------------------------------------------
		-- menu bar hit testing
		-------------------------------------------------------------

		-- moved over a different menu item?
		pick = 0

		-- see if any menu bar items were hit
		for i = 1 to length( my[self][MenuList] ) do

		    -- get an item
		    item = my[self][MenuList][i]

		    -- does the point fall in the item
		    if pointInRect( point, my[item][Size] ) then
			-- set as index, leave loop
			pick = i
			exit
		    end if
		end for


		-- selection changed?
		if  pick != 0
		and pick != menuIndex then
		    -- erase the current dropdown
		    -- erase current menu selection
		    showMenuBarItem( self, menuIndex, 0 )
		    -- set the new item
		    menuIndex = pick
		    -- exit the inner loop
		    exit
		end if


		-------------------------------------------------------------
		-- dropdown hit testing
		-------------------------------------------------------------

		-- moved over a different dropdown item?
		pick = 0
		for i = 1 to length( my[menuId][MenuList] ) do
		    item = my[menuId][MenuList][i]
		    if pointInRect( point, my[item][Size] ) then
			-- save, exit loop
			pick = i
			exit
		    end if
		end for

		-- selection changed?
		if pick != dropIndex then
		    -- erase prior dropdown item
		    showMenuItem( dropId, dropIndex, 0 )
		    -- set the new item
		    dropIndex = pick
		    -- highlight it
		    showMenuItem( dropId, dropIndex, 1 )
		end if

	    end if

	end while
	if dropIndex then
	    for i = 1 to blinks do
		showMenuItem(dropId, dropIndex, 0)
		t=time()
		while time() < t+.05 do
		    mouseState = poll_mouse()
		end while
		showMenuItem(dropId, dropIndex, 1)
		t=time()
		while time() < t+.05 do
		    mouseState = poll_mouse()
		end while
	    end for
	end if

	-- restore the covered area
	restoreImage( x, y, covered )

	-- unhighlight menu
	showMenuBarItem( self, menuIndex, 0 )


    end while


    -- was a dropdown item selected?
    if dropIndex then

	-- get the id of the selected item
	item = my[dropId][MenuList][dropIndex]

	-- active, and not a seperator?
	if  isEnabled( item )
	and compare( my[item][Text], "-" ) != 0 then
	    -- run the menu code

	    -- action attached to menu?
	    if onClick[item] > -1 then
		-- do associated action
		call_proc( onClick[item], {} )
	    end if


	end if

    end if



end procedure




-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- Window class

constant

GoAwayBox={
    "XXXXXXXXXXXXX ",
    "Xxxxxxxxxxxxx.",
    "Xx./////////x.",
    "Xx/XXXXXXXXXx.",
    "Xx/XXXXXXX/Xx.",
    "Xx/XXXXXX//Xx.",
    "Xx/XXXXX///Xx.",
    "Xx/XXXX///.Xx.",
    "Xx/XXX///..Xx.",
    "Xx/XX///...Xx.",
    "Xx/X///....Xx.",
    "Xx////.....Xx.",
    "Xx/XXXXXXXXXx.",
    "Xxxxxxxxxxxxx.",
    " ............."},

GoAwayBoxSimple = {
    "...........",
    ".xxxxxxxxx.",
    ".xxxxxxxxx.",
    ".xxxxxxxxx.",
    ".xxxxxxxxx.",
    ".xxxxxxxxx.",
    ".xxxxxxxxx.",
    ".xxxxxxxxx.",
    ".xxxxxxxxx.",
    ".xxxxxxxxx.",
    "..........."},

    ClickedBoxSimple = {
      "...........",
      ".xxxx.xxxx.",
      ".x.xx.xx.x.",
      ".xx.x.x.xx.",
      ".xxxxxxxxx.",
      "....xxx....",
      ".xxxxxxxxx.",
      ".xx.x.x.xx.",
      ".x.xx.xx.x.",
      ".xxxx.xxxx.",
      "..........." },

TitleBox2 = {
    "xxxxxxxxxxx",
    "xXXXXXXXXXx",
    "xX        x",
    "xX xxxxxx x",
    "xX x    x x",
    "xX xxxxxx x",
    "xX x    x x" ,
    "xX x    x x",
    "xX xxxxxx x",
    "xX        x",
    "xxxxxxxxxxx" },

TitleBox={
    "XXXXXXXXXXXXX ",
    "Xxxxxxxxxxxxx.",
    "Xx./////////x.",
    "Xx/XXXXXXXXXx.",
    "Xx/XXXXXXX/Xx.",
    "Xx/XXXXXX//Xx.",
    "Xx/xxxxxxxxXx.",
    "Xx/XXXX///.Xx.",
    "Xx/xxxxxxxxXx.",
    "Xx/XX///...Xx.",
    "Xx/X///....Xx.",
    "Xx////.....Xx.",
    "Xx/XXXXXXXXXx.",
    "Xxxxxxxxxxxxx.",
    " ............."},

TitleBoxClicked={
    "XXXXXXXXXXXXX ",
    "Xxxxxxxxxxxxx.",
    "XxXXXXXXXXX/x.",
    "XxXXXXXXXXX/x.",
    "XxXXXXXXXX//x.",
    "XxXXXXXXX///x.",
    "XxXxxxxxxxx/x.",
    "XxXXXXX///./x.",
    "XxXxxxxxxxx/x.",
    "XxXXX///.../x.",
    "XxXX///..../x.",
    "XxX///...../x.",
    "XxX////////.x.",
    "Xxxxxxxxxxxxx.",
    " ............."},

TitleBoxClicked2 = {
    "xxxxxxxxxxx",
    "xXXXXXXXXXx",
    "xX xxxxxx x",
    "xX xxxxxx x",
    "xX        x",
    "xX        x",
    "xX        x",
    "xX        x",
    "xX        x",
    "xX        x",
    "xxxxxxxxxxx" },

-- TitleBoxClicked = {
--  "xxxxxxxxxxx",
--  "xXXXXXXXXXx",
--  "xX        x",
--  "xX        x",
--  "xX        x",
--  "xX        x",
--  "xX        x" ,
--  "xX        x",
--  "xX xxxxxx x",
--  "xX xxxxxx x",
--  "xxxxxxxxxxx" },
--
-- ClickedBox = {
--  "xxxxxxxxxxx",
--  "xXXXXxXXXXx",
--  "xXx  x  x x",
--  "xX x x x  x",
--  "xX        x",
--  "xxxx   xxxx",
--  "xX        x",
--  "xX x x x  x",
--  "xXx  x  x x",
--  "xX   x    x",
--  "xxxxxxxxxxx" },
--

GrabBox={"xxxxxxxxxxxxxxxx",
	 "x              x",
	 "x              x",
	 "x  xxxxxx      x",
	 "x  x    xxxxx  x",
	 "x  x    x   x  x",
	 "x  x    x   x  x",
	 "x  x    x   x  x",
	 "x  xxxxxx   x  x",
	 "x   x       x  x",
	 "x   x       x  x",
	 "x   xxxxxxxxx  x",
	 "x              x",
	 "x              x",
	 "xxxxxxxxxxxxxxxx"},

ClickedBox={
    "XXXXXXXXXXXXX ",
    "Xxxxxxxxxxxxx.",
    "XxXXXXXXXXXXx.",
    "XxXXXXXXXXX/x.",
    "XxXXXXXXXX//x.",
    "XxXXXXXXX///x.",
    "XxXXXXXX////x.",
    "XxXXXXX///./x.",
    "XxXXXX///../x.",
    "XxXXX///.../x.",
    "XxXX///..../x.",
    "XxX///...../x.",
    "XxX////////.x.",
    "Xxxxxxxxxxxxx.",
    " ............."},

GrowBox={
    "XXXXXXXXXXXXX ",
    "Xxxxxxxxxxxxx.",
    "Xx./////////x.",
    "Xx/XXXxXxXXXx.",
    "Xx/XXXxXxX/Xx.",
    "Xx/XXXxXx//Xx.",
    "Xx/xxxxXx//Xx.",
    "Xx/XXXXXx/.Xx.",
    "Xx/xxxxxx..Xx.",
    "Xx/XX///...Xx.",
    "Xx/X///....Xx.",
    "Xx////.....Xx.",
    "Xx/XXXXXXXXXx.",
    "Xxxxxxxxxxxxx.",
    " ............."},

GrowBoxClicked={
    "XXXXXXXXXXXXX ",
    "Xxxxxxxxxxxxx.",
    "XxXXXXXXXXXXx.",
    "XxXXXXxXxXX/x.",
    "XxXXXXxXxX//x.",
    "XxXXXXxXx///x.",
    "XxXxxxxXx///x.",
    "XxXXXXXXx/./x.",
    "XxXxxxxxx../x.",
    "XxXXX///.../x.",
    "XxXX///..../x.",
    "XxX///...../x.",
    "XxX////////.x.",
    "Xxxxxxxxxxxxx.",
    " ............."},

GrowBoxClicked2 = {
    "xxxxxxxxxxx",
    "xXXXXXXXXXx",
    "xXxxxxxxxxx",
    "xXxx      x",
    "xXx x     x",
    "xXx  x  x x",
    "xXx   x x x",
    "xXx    xx x",
    "xXx  xxxx x",
    "xXx       x",
    "xxxxxxxxxxx" }

-----------------------------------------------------------------------------
function growIconPos( sequence rect )
    -- return position of minimize icon relative to window

    -- center it vertically
    return justifyAt(  rect[X2]-(length(GrowBox)+6)+5- --(length(GoAwayBox)+12)-
		       (length(TitleBox)+14),
			rect[Y1]+2,
			rect[X2],
			rect[Y1]+theTitleSize,
			CenterJustify, TitleBox )

end function
function titleIconPos( sequence rect )
    -- return position of minimize icon relative to window

    -- center it vertically
    return justifyAt(  rect[X2]-(length(TitleBox)+6)+7, ---(length(GoAwayBox)+12),
			rect[Y1]+2,
			rect[X2],
			rect[Y1]+theTitleSize,
			CenterJustify, TitleBox )

end function
function grabIconPos( sequence rect )
    -- return position of minimize icon relative to window

    return rect[X2]-length(GrabBox[1])+2&rect[Y2]-length(GrabBox)+2

end function
function closeIconPos( sequence rect, integer simple )
    -- return position of minimize icon relative to window

    -- center it vertically
--  return justifyAt(  rect[X2]-(length(GoAwayBox)+6),
--                      rect[Y1],
--                      rect[X2],
--                      rect[Y1]+theTitleSize,
--                      CenterJustify, GoAwayBox )
    if simple then
	return {rect[X1]+8,rect[Y1]+5}
    else
	return {rect[X1]+5,rect[Y1]+3}
    end if
end function


-----------------------------------------------------------------------------
-- 02/03/2001 Removed xShowWindow(excess garbage)
procedure showWindow( integer self )

    -- draw a window

    integer x1, y1, x2, y2, y, focus, fColor, bColor, cantGrow, simple
    sequence title, image, rect, org, point, img, growBox, goAwayBox, img2

    -- don't draw if not visible (takes parent into account)
    if not isVisible( self ) then
	return
    end if

    -- clip to the window's size
    pushClip()
    clipToWindow( self )

    -- get values
    x1 = my[self][Size][X1]
    y1 = my[self][Size][Y1]
    x2 = x1+my[self][Size][X2]
    --x2 = x1+length(image[1])+42
    y2 = y1+my[self][Size][Y2]
    rect = {x1,y1,x2,y2}
    title = my[self][Text]

    -- color scheme based on focus
    focus = hasFocus( self)
    simple = and_bits(my[self][Flags],WS_SIMPLE)=WS_SIMPLE

    if simple then
	goAwayBox=GoAwayBoxSimple
    else
	goAwayBox=GoAwayBox
    end if

    img2=save_image({x1,y1},{x2,y2})
    -- the window
    clipShadowedRect( sysColor[COLOR_WINDOW], x1, y1, x2, y2 )

    ----------------
    -- the title bar
    ----------------

    -- pick the color
    if simple then
	fColor = BRIGHT_WHITE
	bColor = BLACK
    elsif focus then
	fColor = sysColor[COLOR_CAPTIONTEXT]
	bColor = sysColor[COLOR_ACTIVECAPTION]
    else
	fColor = sysColor[COLOR_INACTIVECAPTIONTEXT]
	bColor = sysColor[COLOR_INACTIVECAPTION]
    end if

    -- generate the title
    fselect( SysFont )
    image = build_text_image( " "&title&" ", fColor, bColor, BOLD )

    -- the title bar
    clipFramedRect( bColor, x1, y1, x2, y1+theTitleSize )
    if focus and not simple then
	for y3 = y1+2 to (y1+theTitleSize)-3 by 2 do
	    clipLine(15,x1+3+((length(GoAwayBox)+3)*focus),y3,((x2-2)-(35*focus)),y3)
	    clipLine(GRAY,x1+4+((length(GoAwayBox)+3)*focus),y3+1,((x2-1)-(35*focus)),y3+1)
	end for
    end if

    -- center it vertically
    point = justifyAt( x1, y1, x2-(13*(focus and not simple)), y1+theTitleSize, CenterJustify, image )

    -- display the title
    clipImage( point, image)
    if not simple then
	clipLine(15,x1+1,y1+1,x2,y1+1)
	clipLine(15,x1+1,y1+1,x1+1,y2)
    end if

    if focus then
	-- the close icon
	clipImage( closeIconPos( rect, simple ),
	    buildImage( grayOut(goAwayBox, my[self][Modal]), BLACK, BRIGHT_WHITE ) )
	if my[self][Minimized] then
	    img=TitleBox
	else
	    img=TitleBox
	end if
	if my[self][Grown] then
	    growBox=GrowBox--Clicked2
	else
	    growBox=GrowBox--Clicked
	end if
	if not simple then
	    clipImage( titleIconPos( rect ),
		buildImage( grayOut(img,my[self][Modal]), BLACK, BRIGHT_WHITE ) )
	    cantGrow=and_bits(my[self][Flags],WS_MAXIMIZEBOX) != WS_MAXIMIZEBOX or and_bits(my[self][Flags],WS_DLGFRAME)=WS_DLGFRAME
	    clipImage( growIconPos( rect ),
		buildImage(
		    grayOut(growBox,cantGrow),
		    BLACK, BRIGHT_WHITE ) )
	end if
    end if

    -- the menu
    showMenuBar( self, 0 )

    -- call repaint to handle controls and client area
    repaintWindow( self )

    if simple then
	pixel(img2[1][1..5],{x1,y1})
	pixel(img2[1][length(img2[1])-5..length(img2[1])],{x2-5,y1})
	pixel(img2[length(img2)][1..5],{x1,y1})
	pixel(img2[length(img2)][length(img2[1])-5..length(img2[1])],{x2-5,y1})
	for i = 1 to 5 do
	    pixel(img2[i][1],{x1,y1+(i-1)})
	    pixel(img2[length(img2)][1],{x2+1,y1+(i-1)})
	end for
	pixel(img2[2][1..3],{x1,y1+1})
	pixel(img2[3][1..2],{x1,y1+2})
	pixel(img2[4][1],{x1-1,y1+3})

	pixel(img2[2][length(img2[1])-2..length(img2[1])],{x2-2,y1+1})
	pixel(img2[3][length(img2[1])-1..length(img2[1])],{x2-1,y1+2})
	pixel(img2[4][length(img2[1])],{x2,y1+3})
	pixel(img2[5][length(img2[1])],{x2+1,y1+5})
    end if

    if and_bits(my[self][Flags],WS_THICKFRAME)=WS_THICKFRAME or
       and_bits(my[self][Flags],WS_DLGFRAME) != WS_DLGFRAME and
       not my[self][Minimized] then
	y1 += (20*(length(my[self][MenuList])>0))-3
	y2 += (20*(length(my[self][MenuList])>0))
	clipRect(BLACK, 0, x1+2, y1+23, x2-2, y2-2)
	if my[self][BackColor]=WHITE then
	    clipLine(BRIGHT_WHITE,    x2-2, y2-2, x2-2, y1+22)
	    clipLine(BRIGHT_WHITE,    x2-2, y2-2, x1+2, y2-2)
	else
	    clipLine(WHITE,    x2-2, y2-2, x2-2, y1+22)
	    clipLine(WHITE,    x2-2, y2-2, x1+2, y2-2)
	end if
    end if
    if not my[self][Minimized] then
	clipImage( grabIconPos( rect ), buildImage(GrabBox, BLACK, BRIGHT_WHITE))
    end if

    -- restore
    popClip()

end procedure

-----------------------------------------------------------------------------
function clickWindowIcon( sequence rect, sequence iconUp, sequence iconDown )
    -- generic routine for handling mouse clicking window icons
    integer pressed
    sequence mouseState, priorState, point

    -- force icon draw
    pressed = 0

    -- poll the mouse
    mouseState = poll_mouse()

    while and_bits( mouseState[MOUSE_BUTTONS], MOUSE_LEFT ) do

	-- check the rat
	mouseState = poll_mouse()
	priorState = {}

	-- changed?
	if compare( mouseState, priorState ) then

	    -- save state
	    priorState = mouseState

	    -- redraw icon?
	    point = mouseState[MOUSE_X..MOUSE_Y]

	    -- in icon?
	    if pointInRect( point, rect ) then
		-- need to draw as pressed?
		if not pressed then

		    -- toggle flag
		    pressed = 1

		    -- redraw
		    clipImage( rect[X1..Y1],
			 buildImage( iconDown, BLACK, BRIGHT_WHITE ) )

		end if
	    else
		if pressed then
		    -- toggle flag
		    pressed = 0

		    -- redraw
		    clipImage( rect[X1..Y1],
			 buildImage( iconUp, BLACK, BRIGHT_WHITE ) )
		end if
	    end if
	end if
    end while

    -- return position
    point = mouseState[MOUSE_X..MOUSE_Y]
    return pointInRect( point, rect )

end function


-----------------------------------------------------------------------------
procedure clickGrowIcon( integer self )

    -- handle the close icon being clicked
    sequence point

    -- get position of icon
    point = growIconPos( getWindowRect( self ) )

    -- wait until mouse release
    if clickWindowIcon( point & (point+14), GrowBox, GrowBoxClicked ) then
	-- close the window
	growWindow( self )
    end if

end procedure
procedure clickTitleIcon( integer self )

    -- handle the close icon being clicked
    sequence point

    -- get position of icon
    point = titleIconPos( getWindowRect( self ) )

    -- wait until mouse release
    if clickWindowIcon( point & (point+14), TitleBox, TitleBoxClicked ) then
	-- close the window
	minimizeWindow( self )
    end if

end procedure
procedure clickCloseIcon( integer self )

    -- handle the close icon being clicked
    sequence point, clickedBox, goAwayBox
    integer simple

    simple=and_bits(my[self][Flags],WS_SIMPLE)=WS_SIMPLE

    if simple then
	clickedBox=ClickedBoxSimple
	goAwayBox=GoAwayBoxSimple
    else
	clickedBox=ClickedBox
	goAwayBox=GoAwayBox
    end if

    -- get position of icon
    point = closeIconPos( getWindowRect( self ), simple )

    -- wait until mouse release
    if clickWindowIcon( point & (point+14), goAwayBox, clickedBox ) then
	-- close the window
	closeWindow( self )
    end if

end procedure



-----------------------------------------------------------------------------
-- window routine to use
    classShow[Window] = routine_id( "showWindow" )


-----------------------------------------------------------------------------
-- NEW! 0.3 renamed from loadWindow to openWindow
global procedure openWindow( integer self )

    -- Win32Lib
    -- activate a window, put it on top

    -- on open
-- NEW! 0.4 renamed to onOpen
    if onOpen[self] > -1 then
	call_proc( onOpen[self], {} )
    end if

    -- make it visible
    setVisible( self, 1 )

end procedure

-----------------------------------------------------------------------------
-- NEW! 0.4 renamed from loadModal to openModal
global procedure openModal( integer self )

    -- make the window modal
    my[self][Modal] = 1

    -- load the window
    openWindow( self )

end procedure




-----------------------------------------------------------------------------
global procedure moveWindow( integer self )

    -- handles windows dragging
    integer startX, startY, dx, dy, x, y, parent
    sequence xorRect, winRect, mouseState, priorState, parentClip

    if my[self][Grown] then
	return
    end if

    -- full size of the window
    winRect = getWindowRect( self )

    -- check the mouse: get the absolute position
    mouseState = poll_mouse()
    priorState = mouseState
    startX = mouseState[MOUSE_X]
    startY = mouseState[MOUSE_Y]
    x = my[self][Size][X1]
    y = my[self][Size][Y1]

    -- no xor box drawn yet
    xorRect = {}

    while and_bits( mouseState[MOUSE_BUTTONS], MOUSE_LEFT ) do

	-- check the rat
	mouseState = poll_mouse()

	-- changed?
	if compare( mouseState, priorState ) then

	    -- save state
	    priorState = mouseState

	    -- change in position
	    dx = mouseState[MOUSE_X] - startX
	    dy = mouseState[MOUSE_Y] - startY

	    -- calculate x, y
	    x = my[self][Size][X1]+dx
	    y = my[self][Size][Y1]+dy

	    -- is there a saved region to restore?
	    if length( xorRect ) then
		-- draw a border on parent
		clipXorRect( WHITE, xorRect )
		xorRect = {}
	    end if

	    -- new border position
	    xorRect = winRect + {dx, dy, dx, dy}
	    clipXorRect( WHITE, xorRect )

	end if

    end while

    -- is there a saved region to restore?
    if length( xorRect ) then
	-- draw a border on parent
	clipXorRect( WHITE, xorRect )
	xorRect = {}
    end if

    -- save new position?
    if startX != x
    or startY != y then

	-- change the position
	my[self][Size][X1..Y1] = {x,y}

	-- redraw desktop
	showDesktop()

	if onResize[self] != -1 then
	    call_proc(onResize[self],{})
	end if

    end if

end procedure



-----------------------------------------------------------------------------
global procedure resizeWindow( integer self )

    -- handles windows resizing
    integer parent, diffX, diffY
    sequence xorRect, winRect, mouseState, priorState, parentClip

    if my[self][Minimized] or my[self][Grown]
       or and_bits(my[self][Flags],WS_DLGFRAME)=WS_DLGFRAME then
	return
    end if
    -- full size of the window
    winRect = getWindowRect( self )

    -- get difference in size
    diffX = winRect[X2] - winRect[X1] - my[self][Size][X2]
    diffY = winRect[Y2] - winRect[Y1] - my[self][Size][Y2]

    -- check the mouse: get the absolute position
    mouseState = poll_mouse()
    priorState = mouseState

    -- no xor box drawn yet
    xorRect = {}

    while and_bits( mouseState[MOUSE_BUTTONS], MOUSE_LEFT ) do

	-- check the rat
	mouseState = poll_mouse()

	-- changed?
	if compare( mouseState, priorState ) then

	    -- save state
	    priorState = mouseState

	    -- dragging top?
	    if and_bits( theBorderEdge, North)
	    and winRect[Y2] - mouseState[MOUSE_Y] > 100 then
		winRect[Y1] = mouseState[MOUSE_Y]
	    end if

	    -- dragging bottom?
	    if and_bits( theBorderEdge, South)
	    and mouseState[MOUSE_Y] - winRect[Y1] > 16 then
		winRect[Y2] = mouseState[MOUSE_Y]
	    end if

	    -- dragging left?
	    if and_bits( theBorderEdge, West)
	    and winRect[X2] - mouseState[MOUSE_X] > 100 then
		winRect[X1] = mouseState[MOUSE_X]
	    end if

	    -- dragging right?
	    if and_bits( theBorderEdge, East)
	    and mouseState[MOUSE_X] - winRect[X1] > 100 then
		winRect[X2] = mouseState[MOUSE_X]
	    end if

	    -- is there a saved region to restore?
	    if length( xorRect ) then
		-- draw a border on parent
		clipXorRect( WHITE, xorRect )
		clipXorLine( WHITE, xorRect[X2]-length(GrabBox[1]), xorRect[Y1]+1, xorRect[X2]-length(GrabBox[1]), xorRect[Y2]-1)
		clipXorLine( WHITE, xorRect[X1]+1, xorRect[Y2]-length(GrabBox), xorRect[X2]-1, xorRect[Y2]-length(GrabBox))
		xorRect = {}
	    end if

	    -- new border position
	    xorRect = winRect
	    clipXorRect( WHITE, xorRect )
	    clipXorLine( WHITE, xorRect[X2]-length(GrabBox[1]), xorRect[Y1]+1, xorRect[X2]-length(GrabBox[1]), xorRect[Y2]-1)
	    clipXorLine( WHITE, xorRect[X1]+1, xorRect[Y2]-length(GrabBox), xorRect[X2]-1, xorRect[Y2]-length(GrabBox))

	end if

    end while

    -- is there a saved region to restore?
    if length( xorRect ) then
	-- draw a border on parent
	clipXorRect( WHITE, xorRect )
	xorRect = {}
    end if

    -- adjust difference
    winRect[X2] = winRect[X2] - winRect[X1] - diffX
    winRect[Y2] = winRect[Y2] - winRect[Y1] - diffY

    -- save the size
    my[self][Size] = winRect

    -- redraw desktop
    showDesktop()

end procedure




-----------------------------------------------------------------------------
global procedure addItem( integer self, sequence item )
    -- Win32Lib
    -- add item to list
    my[self][Text] = append( my[self][Text], item )
    my[self][FirstItem]=1
    -- redisplay
    showControl( self )
end procedure


-----------------------------------------------------------------------------
global procedure setFont( integer self, sequence name, integer points,
			    integer attrib )

    -- Win32Lib
    -- look for a replacement to load
    my[self][Font] = 0
    for i = 1 to length(FontNames) by 2 do
	    if compare( name, FontNames[i] ) = 0 then
		-- load the font
		my[self][Font] = fload( FontNames[i+1] )
		exit
	    end if
    end for

    -- no match?
    if my[self][Font]= 0 then
	-- default
	my[self][Font] = SysFont

    end if

    -- save size and attributes
    my[self][FontSize]   = points
    my[self][FontAttrib] = attrib

end procedure


-----------------------------------------------------------------------------
global procedure setDefaultFont( integer self )
    -- Win32Lib
    -- set font in window to default font

    -- save size and attributes
    my[self][Font]       = SysFont
    my[self][FontSize]   = 10
    my[self][FontAttrib] = 0

end procedure





-----------------------------------------------------------------------------
global procedure setPosition( integer self, integer x, integer y )
    -- Win32Lib
    -- position the pen
    x += (my[self][HasFrame]*3)
    y += (my[self][HasFrame]*3)
    my[self][PenPos] = { x, y }
end procedure


-----------------------------------------------------------------------------
global procedure setPenColor( integer self, integer color )
    -- Win32Lib
    -- pen color
    my[self][PenColor] = color
end procedure

-----------------------------------------------------------------------------
global procedure setTextColor( integer self, integer color )
    -- Win32Lib
    -- same as pen color
    my[self][PenColor] = color
end procedure



global procedure copyBlt( integer dest, integer x, integer y, integer src)
    if my[src][Class]=Pixmap then
	clipImage({x,y},my[src][Contents])
    end if
end procedure

-----------------------------------------------------------------------------
global procedure drawLine( integer self, integer x1, integer y1,
			    integer x2, integer y2 )

    -- Win32Lib
    -- draw a line
    -- This does not reposition the pen.

    if my[self][Class]=Pixmap then
	for i = x1 to x2 do
	    for j = y1 to y2 do
		my[self][Contents][j+1][i+1]=my[self][PenColor]
	    end for
	end for
	return
    end if

    -- offset into window
    pushClip()
    clipToClient( self )

    -- draw a clipped line
    clipLine( my[self][PenColor], winX1+x1, winY1+y1,
				    winX1+x2, winY1+y2 )

    -- restore clipping
    popClip()


end procedure


-----------------------------------------------------------------------------
global procedure setScrollRange( integer self, integer min, integer max  )
    --$ Win32Lib Emulation
    -- Sets the scroll bar minimum and maximum values.
    my[self][Range][1..2] = {min,max}
end procedure


-----------------------------------------------------------------------------
global procedure setScrollChange( integer self, integer line, integer page )
    --$ Win32Lib Emulation
    -- Sets the range for the scroll bar.
    -- This defines how much the scroll bar will move.
    my[self][Range][3..4] = {line,page}
end procedure


-----------------------------------------------------------------------------
global procedure wPuts( integer self, sequence text )
    --$ Win32Lib Emulation
    -- Display text in requested window.
    -- This is the Win32Lib version of Euphoria's puts() command.
    sequence pos, image

    -- calculate position
    pushClip()
    clipToClient( self )

    -- select font
    fselect( my[self][Font] )

    -- build an display text
    image = build_text_image( text,     -- text
		my[self][PenColor],     -- text color comes from pen color
		my[self][BackColor],    -- background color
		my[self][FontAttrib] )  -- text attribute

    -- display the text
    clipImage( {winX1, winY1} + my[self][PenPos], image )

    -- restore clipping
    popClip()


end procedure


-----------------------------------------------------------------------------
global procedure wPrintf( integer id, sequence s1, sequence s2 )
    --$ Win32Lib Emulation
    -- Display text in requested window.
    -- This is the Win32Lib version of Euphoria's printf() command.
    wPuts( id, sprintf( s1, s2 ) )

end procedure


-----------------------------------------------------------------------------
global procedure setTimer( integer window, integer id, integer ticks )
    --$ Win32Lib Emulation
    -- Activate a timer.
    -- { Timer in, TimerId, TimerTicks, TimerStart }
    timers = append( timers, { window, id, ticks/1000, time() } )
end procedure

-----------------------------------------------------------------------------
global procedure killTimer( integer id, integer timerId )
    --$ Win32Lib Emulation
    -- Deactivate a timer.
    sequence timer
    for i = 1 to length( timers ) do
	-- get a timer
	timer = timers[i]

	-- match?
	if timer[TimerWin] = id
	and timer[TimerId] = timerId then
	    -- remove it
	    timers = timers[1..i-1] & timers[i+1..length(timers)]
	    -- finished
	    return
	end if
    end for
end procedure



-----------------------------------------------------------------------------
global procedure setWindowBackColor( integer self, integer rgb )
    --$ Win32Lib Emulation
    -- for now, hack it. will need to add a "find closest rgb" routine
    my[self][BackColor]    = rgb
    -- needs to generate a "system color changed" event
end procedure

-----------------------------------------------------------------------------
global function getSysColor( integer index )
    -- Win32Lib
    -- return system color
    return sysColor[ index ]
end function



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- The Event Loop



-----------------------------------------------------------------------------
function inWindow( sequence point )
    -- return the id of the window the point isWin
    integer self
    sequence rect

    -- look through the stack
    for i = 1 to length( windowStack ) do

	-- get a window
	self = windowStack[i]

	-- point fallsWin window?
	if pointInRect( point, getWindowRect( self ) ) then
	    return self
	end if

	-- is the window modal?
	if my[self][Modal] then
	    -- don't look any further
	    return 0
	end if

    end for

    return 0

end function


-----------------------------------------------------------------------------
constant
    InDesk      = 0,        -- not in a window
    InTitle     = 1,        -- in title bar, not on window control
    InMenu      = 2,        -- in the menu bar
    InBorder    = 3,        -- on the window border (resize)
    InContent   = 4,        -- in client area
    InMinimize  = 5,        -- on window grow icon
    InGrow      = 6,        -- on window grow icon
    InClose     = 7         -- on window close icon


-----------------------------------------------------------------------------
function inPart( integer self, sequence point )
    -- return the part of the window that was hit
    integer x, y, cantGrow, simple
    sequence rect, pos

    simple=and_bits(my[self][Flags],WS_SIMPLE)=WS_SIMPLE

    -- get the points
    x = point[X1]
    y = point[Y1]

    -- get the window size
    rect = getWindowRect( self )

    -- on the border?
    if x < rect[X1] + theBorderSize
    or x > rect[X2] - theBorderSize
    or y < rect[Y1] + theBorderSize
    or y > rect[Y2] - theBorderSize then

	-- where in the border?
	theBorderEdge = 0

	if y < rect[Y1] + theBorderSize then
	    theBorderEdge = theBorderEdge+North
	elsif y > rect[Y2] - theBorderSize then
	    theBorderEdge = theBorderEdge+South
	end if

	if x < rect[X1] + theBorderSize then
	    theBorderEdge = theBorderEdge+West
	elsif x > rect[X2] - theBorderSize then
	    theBorderEdge = theBorderEdge+East
	end if

	return InBorder

    end if

    if not my[self][Minimized] then
	if (x >= rect[X2]-length(GrabBox)+2 and x <= rect[X2]) and
	(y >= rect[Y2]-length(GrabBox)+2 and y <= rect[Y2]) then
	    theBorderEdge=South+East
	    return InBorder
	end if
    end if

    -- in the title bar?
    if y < rect[Y1] + theTitleSize then
	-- in a title bar control?

	    -- the grow icon?
	    cantGrow=and_bits(my[self][Flags],WS_MAXIMIZEBOX) != WS_MAXIMIZEBOX or and_bits(my[self][Flags],WS_DLGFRAME)=WS_DLGFRAME
	    if not cantGrow then
		pos = growIconPos( rect )
		if pointInRect( point, pos & (pos+14) ) then
		    return InGrow
		end if
	    end if

	    if not my[self][Modal] then
		-- the minimize icon?
		pos = titleIconPos( rect )
		if pointInRect( point, pos & (pos+14) ) then
		    return InMinimize
		end if

		-- the close icon?
		pos = closeIconPos( rect, simple )
		if pointInRect( point, pos & (pos+14) ) then
		    return InClose
		end if
	    end if

	    -- just in the title
	    return InTitle

    end if

    -- in the menu bar?
    if hasMenu( self ) then
	if y < rect[Y1] + theTitleSize + theMenuSize then
	    return InMenu
	end if
    end if

    -- must beWin the workspace
    return InContent

end function


-----------------------------------------------------------------------------
function inControl( integer self, sequence point )
    -- return the control that was hit, or zero if none

    -- go through controls
    for control = 1 to length( my ) do
	-- owned by window?
	if my[control][Parent] = self then
	    -- hit?
	    if pointInRect( point, getWindowRect( control ) ) then

		-- is it static?
		if find( my[control][Class], StaticControls ) then
		    -- can't get events
		    return 0
		else
		    return control
		end if
	    end if

	end if

    end for

    -- no matches
    return 0

end function


-----------------------------------------------------------------------------
procedure runTimers()

    integer id
    sequence timer

    for i = 1 to length( timers ) do
	-- get a timer
	timer = timers[i]

	-- compare changeWin time
	if time() - timer[TimerStart] >= timer[TimerTicks] then

	    -- reset timer
	    timers[i][TimerStart] = time()

	    -- action?
	    id = timer[TimerWin]

	    if onTimer[id] > -1 then

		-- pass it the timer id
		call_proc( onTimer[ id ], {timer[TimerId]} )

	    end if

	end if

    end for

end procedure


-----------------------------------------------------------------------------
procedure setForegroundWindow( integer self )
    -- Win32
    setVisible( self, 1 )
end procedure




-----------------------------------------------------------------------------
global procedure setFocus( integer self )

    -- Win32Lib
    integer prior, parent

    -- focus on a control
    if my[self][Class] = Window then
	setForegroundWindow( self )
	return
    end if

    -- skip if has focus
    prior = getFocus()
    if prior = self then
	return
    end if

    -- lose focus
    parent = my[self][Parent]
    my[parent][Focus] = self

    -- code?
    if prior then
	-- redraw control
	showControl( prior )

	if onLostFocus[prior] > -1 then
	    call_proc( onLostFocus[prior], {} )
	end if

    end if

    -- draw focused control
    setVisible( self, 1 )

end procedure



-----------------------------------------------------------------------------
procedure tabAhead( integer self, integer direction )

    -- tabs to nth control in window
    -- on error, does nothing
    integer tabStop
    sequence tabStops

    -- get tabs
    tabStops = my[self][TabStops]

    if length( tabStops ) = 0 then
	-- nothing to focus on!
	return
    end if

    -- find current tab, plus direction
    tabStop = find( my[self][Focus], tabStops ) + direction

    -- check for wrapping
    if tabStop > length( tabStops ) then
	-- wrap to front
	tabStop = 1

    elsif tabStop < 1 then
	-- wrap to end
	tabStop = length( tabStops )

    end if

    -- find the tab
    setFocus( tabStops[ tabStop ] )

end procedure


-----------------------------------------------------------------------------
procedure eventLoop( integer mainWin )
    -- run the main event loop
    integer priorMain, key, hitWindow, hitPart, hitControl, class, action
    sequence point, inPoint, newKeys, oldKeys
    object mouse

    integer
	focusWindow,
	focusControl

    -- if there was a prior main window, save it
    priorMain = mainWindow

    -- set new main window
    mainWindow = mainWin

    -- open the window
    openWindow( mainWindow )

    -- reset flag
    exitEventLoop = 0

    -- mouse pointer
    mouse_pointer(1)

    -- clear the key queue
    newKeys = {}
    oldKeys = {}

    while not exitEventLoop do

	-- get the top window
	focusWindow = windowStack[1]
	focusControl = my[focusWindow][Focus]

-- NEW! 0.4 route events to the window if no controls in window
	-- if no control has focus, send event to the window
	if focusControl = 0 then
	    focusControl = focusWindow
	end if

    -----------------------------------------------------------------
    -- handle timer events
    -----------------------------------------------------------------

	if length( timers ) then
	    runTimers()
	end if

    -----------------------------------------------------------------
    -- handle key events
    -----------------------------------------------------------------


	-- check the keyboard
	-- newKeys = scanToAscii( get_keys() )
	newKeys = {get_key()}

	-- new key pressed?
	if compare( newKeys, oldKeys) != 0
	and compare( newKeys, {-1} ) != 0 then

	    -- DEBUG! Eventually remove
	    if compare( newKeys, {-1} ) != 0 then
		if compare( newKeys, {27} ) = 0 then
		    abortErr( "User Pressed Esc Key.", {} )
		end if
	    end if

-- NEW! 0.4 removed conversion to Win32 mapping; use VK_xxx instead

	    -- what keys were released?
	    for i = 1 to length( oldKeys ) do
		if not find( oldKeys[i], newKeys ) then
		    -- generate a key release event
		end if
	    end for

	    -- what new keys were pressed?
	    for i = 1 to length( newKeys ) do
		if not find( newKeys[i], oldKeys ) then
		    -- get the key
		    key = newKeys[i]

		    if key = VK_TAB then
			-- tab ahead
			tabAhead( focusWindow, 1 )

		    elsif key = SHIFT_TAB then
			-- tab back
			tabAhead( focusWindow, -1 )

		    else
			-- position( 2, 1 )
			-- ? key
		    end if

		    -- pass key to control?
		    if focusControl then

			-- get class
			class = my[focusControl][Class]

			-- get action
			action = classKey[class]

			if action > -1 then
			    call_proc( action, {focusControl, key} )
			end if

-- NEW! 0.4 added onKeyDown processing
			-- on key down event?
			action = onKeyDown[focusControl]
			if action > -1 then
-- NEW! 0.4 passes only the key
			    call_proc( action, {key} )
			end if

-- NEW! 0.4 onKeyPress only processes visible keys
			-- visible key?
			if key >= 32 and key <= 175 then

			    -- look for key press event
			    action = onKeyPress[focusControl]
			    if action > -1 then
-- NEW! 0.4 passes only the key
				call_proc( action, {key} )
			    end if

			end if

		    else
			-- pass key to window?

			-- on key event?
			action = onKeyPress[focusWindow]

			if action > -1 then
			    call_proc( action, {key} )
			end if


		    end if
		end if
	    end for
	end if

	-- update keys
	oldKeys = newKeys

    -----------------------------------------------------------------
    -- handle mouse events
    -----------------------------------------------------------------

	-- poll mouse
	mouse = get_mouse()

	-- different?
	if sequence( mouse ) then

	    -- get point
	    point = mouse[MOUSE_X..MOUSE_Y]
	    hitPart = 0
	    hitControl = 0

	    -- where did it land?
	    hitWindow = inWindow( point )

	    -----------------
	    -- mouse pointer
	    -----------------

	    -- hit window?
	    if hitWindow then

		-- button press?
		if and_bits( mouse[MOUSE_EVENT], LEFT_DOWN+RIGHT_DOWN )
		then
		    -- needs focus?
		    if not hasFocus( hitWindow ) then
			-- change the focus
			 setFocus( hitWindow )
		    end if

		end if

		-- change the cursor
		set_mouse_cursor( IDC_ARROW )
		hitPart = inPart( hitWindow, point )

		-- in client area?
		if hitPart = InContent then
		    -- determine what control, if any, was hit
		    hitControl = inControl( hitWindow, point )

		elsif hitPart = InBorder then
		    -- select proper cursor
		    if theBorderEdge = North
		    or theBorderEdge = South then
			set_mouse_cursor( IDC_SIZENS )

		    elsif theBorderEdge = East
		       or theBorderEdge = West then
			set_mouse_cursor( IDC_SIZEWE )

		    elsif theBorderEdge = North+West
		       or theBorderEdge = South+East then
			set_mouse_cursor( IDC_SIZENWSE )

		    elsif theBorderEdge = South+West
		       or theBorderEdge = North+East then
			set_mouse_cursor( IDC_SIZENESW )

		    end if

		end if

	    else
		-- set pointer to indicate not in any window
		set_mouse_cursor( IDC_NO )
	    end if


	    ----------------
	    -- onMouse event
	    ----------------

	    if  hitWindow
	    and hitPart = InContent then

		-- point relative to hit window
		inPoint = intoWindowPoint( hitWindow, point )

		-- hit a control?
		if hitControl then
		    action = onMouse[hitControl]
		else
		    action = onMouse[hitWindow]
		end if

		-- is there a mouse action?
		if action > -1 then
		    -- move
		    if and_bits( mouse[MOUSE_EVENT], MOVE ) then
			call_proc( action, {MOUSE_MOVE} & inPoint )
		    end if

		    -- left down
		    if and_bits( mouse[MOUSE_EVENT], LEFT_DOWN ) then
			call_proc( action, {LEFT_DOWN} & inPoint )
		    end if

		    -- right down
		    if and_bits( mouse[MOUSE_EVENT], RIGHT_DOWN ) then
			call_proc( action, {RIGHT_DOWN} & inPoint )
		    end if

		    -- left up
		    if and_bits( mouse[MOUSE_EVENT], LEFT_UP ) then
			call_proc( action, {LEFT_UP} & inPoint )
		    end if

		    -- right up
		    if and_bits( mouse[MOUSE_EVENT], RIGHT_UP ) then
			call_proc( action, {RIGHT_UP} & inPoint )
		    end if

		end if
	    end if


	    ----------------------------
	    -- mouse class action events
	    ----------------------------

	    -- left button down on something?
	    if and_bits( mouse[MOUSE_EVENT], LEFT_DOWN )
	    and hitWindow then

		-- mouse drag
		if hitPart = InTitle then
		    if my[hitWindow][Minimized] then
			minimizeWindow(hitWindow)
		    else
			moveWindow( hitWindow )
		    end if

		elsif hitPart = InBorder then
		    resizeWindow( hitWindow )

		-- hit title bar?
		elsif hitPart = InTitle then
		    -- window control handler

		elsif hitPart = InGrow then
		    -- close window control
		    clickGrowIcon( hitWindow )

		elsif hitPart = InMinimize then
		    -- close window control
		    clickTitleIcon( hitWindow )

		elsif hitPart = InClose then
		    -- close window control
		    clickCloseIcon( hitWindow )

		-- hit the menu bar?
		elsif hitPart = InMenu then
		    -- menu handler
		    runMenu( hitWindow, mouse )

		-- hit the client area?
		elsif hitPart = InContent then

		    -- hit a control?
		    if hitControl and isEnabled(hitControl) and isVisible(hitControl) then

			-- focus on control
			setFocus( hitControl )

			-- get class
			class = my[hitControl][Class]

			-- get class click action
			action = classClick[class]
			if action > -1 then
			    call_proc( action, {hitControl} )
			end if


		    else

			-- clicked in workspace
			action = onClick[hitWindow]
			if action > -1 then
			    call_proc( action, {} )
			end if

		    end if -- hitControl/inContent
		end if -- where hit
	    end if -- MOUSE_LEFT and hitWindow

	end if

    end while

    -- hide mouse
    mouse_pointer(0)

    -- reset flag
    exitEventLoop = 0

    -- restore the prior window
    mainWindow = priorMain

end procedure




-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
--$ The Message Box
--
-- This emulates the message box that is available in the Win32 version
-- of Euphoria. It is a difficult control to emulate, because it returns
-- a value. That means that the message box has it's own event loop.
--
-- In practice, the message box uses the same event loop as the main
-- code, but calls and exits from it.



-----------------------------------------------------------------------------
-- Possible style values for message_box() style sequence
global constant
    MB_ABORTRETRYIGNORE = #02, --  Abort, Retry, Ignore
    MB_APPLMODAL = #00,       -- User must respond before doing something else
    MB_DEFAULT_DESKTOP_ONLY = #20000,
    MB_DEFBUTTON1 = #00,      -- First button is default button
    MB_DEFBUTTON2 = #100,      -- Second button is default button
    MB_DEFBUTTON3 = #200,      -- Third button is default button
    MB_DEFBUTTON4 = #300,   -- Fourth button is default button
    MB_HELP = #4000,            -- Windows 95: Help button generates help event

    MB_ICONHAND         = #10,
    MB_ICONQUESTION     = #20,
    MB_ICONEXCLAMATION  = #30,
    MB_ICONASTERISK     = #40,
    MB_ICONBOMB         = #80,

    MB_ICONINFORMATION  = MB_ICONASTERISK,
    MB_ICONSTOP         = MB_ICONHAND,
    MB_ICONERROR        = MB_ICONHAND,
    MB_ICONWARNING      = MB_ICONEXCLAMATION,

    MB_OK = #00,              -- Message box contains one push button: OK
    MB_OKCANCEL = #01,        -- Message box contains OK and Cancel
    MB_RETRYCANCEL = #05,     -- Message box contains Retry and Cancel
    MB_RIGHT = #80000,        -- Windows 95: The text is right-justified
    MB_RTLREADING = #100000,   -- Windows 95: For Hebrew and Arabic systems
    MB_SERVICE_NOTIFICATION = #40000, -- Windows NT: The caller is a service
    MB_SETFOREGROUND = #10000,   -- Message box becomes the foreground window
    MB_SYSTEMMODAL  = #1000,    -- All applications suspended until user responds
    MB_TASKMODAL = #2000,       -- Similar to MB_APPLMODAL
    MB_YESNO = #04,           -- Message box contains Yes and No
    MB_YESNOCANCEL = #03      -- Message box contains Yes, No, and Cancel

-- possible values returned by MessageBox()
-- 0 means failure
global constant IDABORT = 3,  -- Abort button was selected.
		IDCANCEL = 2, -- Cancel button was selected.
		IDIGNORE = 5, -- Ignore button was selected.
		IDNO = 7,     -- No button was selected.
		IDOK = 1,     -- OK button was selected.
		IDRETRY = 4,  -- Retry button was selected.
		IDYES = 6    -- Yes button was selected.

-- the three buttons in the message box
integer
    button1,
    button2,
    button3

-----------------------------------------------------------------------------
-- icons for the message box

constant
IconInformation = createDIB( {
    "11111111111111111111111111111111",
    "12222222211111111111111111111111",
    "12222222211111111111111111111111",
    "12222222211111111111111111111111",
    "12222222211111111111111111111111",
    "12222222211111111122222211111111",
    "12221222211111112222222222111111",
    "12221222211111122222222222211111",
    "12222222211112222222222222222111",
    "12222222211112222222222222222111",
    "12222222211122222222222222222211",
    "12222222211122211121112111222211",
    "12222222211122222222222222222211",
    "12222222211122222222222222222211",
    "12222222211122211121112121222211",
    "12222222211122222222222222222211",
    "12222222211122222222222222222211",
    "12222222211122211121211122222211",
    "12222222211122222222222222222211",
    "12222111111122222222222222222211",
    "12222221111122211112111211222211",
    "12222221111122222222222222222111",
    "12222221111122222222222222222111",
    "12222221111122222222222222211111",
    "12221111122222222222222221111111",
    "12222221111111111111111111111111",
    "12222221111111111111111111111111",
    "12222221111111111111111111111111",
    "12222221111111111111111111111111",
    "11111111111111111111111111111111"}-49,{{0,0,206},{206,206,206}}),


IconQuestion = createDIB( {
    "11111111111111111111111111111111",
    "12222222211111111111111111111111",
    "12222222211111111111111111111111",
    "12222222211111111111111111111111",
    "12222222211111111111111111111111",
    "12222222211111111122222211111111",
    "12221222211111112222222222111111",
    "12221222211111122221111122211111",
    "12222222211112222111111111222111",
    "12222222211112221111111111122111",
    "12222222211122221111221111122211",
    "12222222211122211112222111122211",
    "12222222211122211112222111122211",
    "12222222211122222222221111222211",
    "12222222211122222222211112222211",
    "12222222211122222222111122222211",
    "12222222211122222222111122222211",
    "12222222211122222221111222222211",
    "12222222211122222221111222222211",
    "12222111111122222221111222222211",
    "12222221111122222222222222222211",
    "12222221111122222221111222222111",
    "12222221111122222221111222222111",
    "12222221111122222221111222211111",
    "12221111122222222222222221111111",
    "12222221111111111111111111111111",
    "12222221111111111111111111111111",
    "12222221111111111111111111111111",
    "12222221111111111111111111111111",
    "11111111111111111111111111111111"}-49,{{0,0,0},{206,206,206}}),

IconExclamation = createDIB( {
    "33333333333333322333333333333333",
    "33333333333333222233333333333333",
    "33333333333333222233333333333333",
    "33333333333332211223333333333333",
    "33333333333332211223333333333333",
    "33333333333322111122333333333333",
    "33333333333322111122333333333333",
    "33333333333221122112233333333333",
    "33333333332211222211223333333333",
    "33333333332211222211223333333333",
    "33333333322111222211122333333333",
    "33333333322111222211122333333333",
    "33333333221111222211112233333333",
    "33333333221111222211112233333333",
    "33333332211111222211111223333333",
    "33333332211111222211111223333333",
    "33333322111111222211111122333333",
    "33333322111111222211111122333333",
    "33333221111111222211111112233333",
    "33333221111111122111111112233333",
    "33332211111111122111111111223333",
    "33332211111111111111111111223333",
    "33322111111111111111111111122333",
    "33322111111111122111111111122333",
    "33221111111111222211111111112233",
    "32211111111111122111111111111223",
    "32211111111111111111111111111223",
    "22111111111111111111111111111122",
    "22222222222222222222222222222222",
    "32222222222222222222222222222223"}-49,{{255,255,0},{0,0,0},{206,206,206}}),

IconHand = createDIB( {
    "11111111111111111111111111111111",
    "11111111111111111111111111111111",
    "11111111222222222222222211111111",
    "11111112222222222222222221111111",
    "11111122222222222222222222111111",
    "11111222222222212222222222211111",
    "11112222222222111222222222221111",
    "11122222222122111221122222222111",
    "11222222221112111211122222222211",
    "12222222221112111211122222222221",
    "22222222121112111211122222222222",
    "22222221121112111211122222222222",
    "22222221121112111211122222222222",
    "22222221121112111211122222222222",
    "22222221121112111211122222222222",
    "22222221121112111211122222222222",
    "22222221121111111111122211122222",
    "22222221111111111111122111122222",
    "22222221111111111111122111122222",
    "22222221111111111111121111222222",
    "22222221111111111111111111222222",
    "22222221111111111111111112222222",
    "22222221111111111111111122222222",
    "22222221111111111111111122222222",
    "22222221111111111111111222222222",
    "22222221111111111111111222222222",
    "12222221111112111111112222222221",
    "11222222111111211111122222222211",
    "11122222222222222222222222222111",
    "11112222222222222222222222221111",
    "11111222222222222222222222211111",
    "11111122222222222222222222111111",
    "11111112222222222222222221111111"}-49,{{206,206,206},{255,0,0}}),

IconBomb = textToBitmap( {
    "                    x       x  ",
    "                           x   ",
    "                     x    x    ",
    "                x        x     ",
    "                  x   x        ",
    "                    x          ",
    "           xxxxx          x x x",
    "          x     xx   xxx       ",
    "         x        xxx          ",
    "         x                x    ",
    "         x          x  x   x   ",
    "      xxxxxxx      x        x  ",
    "      xxxxxxx     x    x     x ",
    "      xxxxxxx                  ",
    "    xxxxxxxxxxx        x       ",
    "  xxxxxxxxxxxxxxx              ",
    "  xxxxxxxxxxxxxxx              ",
    " xxxxxxxxxxxxxxxxx             ",
    " xxxxxxxxxxxxxxxxx             ",
    "xxxxxxxxxxxxxx xxxx            ",
    "xxxxxxxxxxxxxxxxxxx            ",
    "xxxxxxxxxxxxxxx xxx            ",
    "xxxxxxxxxxxxxxx xxx            ",
    "xxxxxxxxxxxxxxx xxx            ",
    "xxxxxxxxxxxxxxxxxxx            ",
    "xxxxxxxxxxxxxx xxxx            ",
    " xxxxxxxxxxxxxxxxx             ",
    " xxxxxxxxxxx xxxxx             ",
    "  xxxxxxxxxxxxxxx              ",
    "  xxxxxxxxxxxxxxx              ",
    "    xxxxxxxxxxx                ",
    "      xxxxxxx                  "})


constant
-- the message box and controls
    MsgBox       = create( Window, "",  0, Default, Default, 200, 100, 0 ),
    MBoxButton1  = create( PushButton, "", MsgBox, 20, 50, 80, 20, 0 ),
    MBoxButton2  = create( PushButton, "", MsgBox, 100, 50, 80, 20, 0 ),
    MBoxButton3  = create( PushButton, "", MsgBox, 100, 50, 80, 20, 0 ),
    MBoxText     = create( LText, "", MsgBox, 100, 100, 100, 20, 0 ),
    MBoxIcon     = create( Bitmap, "", MsgBox, 10, 10, 32, 32, 0 )

integer MBoxResult
    MBoxResult = 0

-----------------------------------------------------------------------------
procedure msgBoxButton1()
    -- set result and close window
    MBoxResult = button1
    closeWindow( MsgBox )
end procedure

-----------------------------------------------------------------------------
procedure msgBoxButton2()
    -- set result and close window
    MBoxResult = button2
    closeWindow( MsgBox )
end procedure

-----------------------------------------------------------------------------
procedure msgBoxButton3()
    -- set result and close window
    MBoxResult = button3
    closeWindow( MsgBox )
end procedure


-- assign behaviors
onClick[MBoxButton1]    = routine_id("msgBoxButton1")
onClick[MBoxButton2]    = routine_id("msgBoxButton2")
onClick[MBoxButton3]    = routine_id("msgBoxButton2")


-----------------------------------------------------------------------------
function bits( atom a, atom bits )
    -- true if all bits in pattern are set
    return and_bits( a, bits ) = bits
end function


-----------------------------------------------------------------------------
global function message_box( sequence text, sequence title, integer flags )
    -- $ Euphoria Win32 Emulation
    -- Emulation of Euphoria message_box function
    integer message, x1, y1, x2, y2
    sequence size

    -- hide the mouse
    mouse_pointer(0)

    -- save the flags
    my[MsgBox][Flags] = flags+WS_DLGFRAME

    -----------
    -- the icon
    -----------

    -- set the icon. the bits are not exclusive, so i have to
    -- test that it matches *all* the bits of the flag

    if bits( flags, MB_ICONBOMB) then
	-- this is *my* special code
	setBitmap( MBoxIcon, IconBomb)

    -- this uses more than 1 bit
    elsif bits( flags, MB_ICONEXCLAMATION ) then
	setBitmap( MBoxIcon, IconExclamation )

    elsif bits( flags, MB_ICONINFORMATION ) then
	setBitmap( MBoxIcon, IconInformation )

    elsif bits( flags, MB_ICONERROR ) then
	setBitmap( MBoxIcon, IconHand )

    elsif bits( flags, MB_ICONQUESTION ) then
	setBitmap( MBoxIcon, IconQuestion )

    else
	setBitmap( MBoxIcon, 0 )

    end if

    --------------------
    -- position the text
    --------------------

    -- get the size of the text
    size = getTextRect( text, 0 )

    -- position the text
    if my[MBoxIcon][Bmp] then
	-- adjust for the icon
	size = size + { 60, 20, 60, 20 }
    else
	-- no icon
	size = size + { 10, 10, 10, 10 }
    end if

    size[3] += 13*3
    -- set the size
    my[MBoxText][Size] = size

    -- set the text
    setText( MsgBox, title )
    setText( MBoxText, text )

    -- get the size for positioning various other elements
    x1 = size[X1]
    y1 = size[Y1]
    x2 = size[X2]
    y2 = size[Y2]

    --------------
    -- the buttons
    --------------

    -- get low 3 bits on flags - button codes
    flags = and_bits( flags, 7 )

    -- #01
    if flags = MB_OKCANCEL then -- #01
	-- text
	setText( MBoxButton1, "Ok" )
	setText( MBoxButton2, "Retry" )

	-- values
	button1 = IDOK
	button2 = IDCANCEL
	button3 = 0

    -- #02
    elsif flags = MB_ABORTRETRYIGNORE then
	-- text
	setText( MBoxButton1, "Abort" )
	setText( MBoxButton2, "Retry" )
	setText( MBoxButton3, "Ignore" )

	-- values
	button1 = IDABORT
	button2 = IDRETRY
	button3 = IDIGNORE

    -- #03
    elsif flags = MB_YESNOCANCEL then
	-- text
	setText( MBoxButton1, "Yes" )
	setText( MBoxButton2, "No" )
	setText( MBoxButton3, "Cancel" )

	-- values
	button1 = IDYES
	button2 = IDNO
	button3 = IDCANCEL

    -- #04
    elsif flags = MB_YESNO then
	-- text
	setText( MBoxButton1, "Yes" )
	setText( MBoxButton2, "No" )

	-- values
	button1 = IDYES
	button2 = IDNO
	button3 = 0

    -- #05
    elsif flags = MB_RETRYCANCEL then
	-- text
	setText( MBoxButton1, "Retry" )
	setText( MBoxButton2, "Cancel" )

	-- values
	button1 = IDRETRY
	button2 = IDCANCEL
	button3 = 0

    else
	-- text
	setText( MBoxButton1, "Ok" )

	-- values
	button1 = IDOK
	button2 = 0
	button3 = 0

    end if

    -- activate appropriate buttons
    setVisible( MBoxButton1, button1 )
    setVisible( MBoxButton2, button2 )
    setVisible( MBoxButton3, button3 )

    -- position buttons relative to text
    my[MBoxButton1][Size] = { 20, 20+y2, 80, 20 }
    my[MBoxButton2][Size] = {120, 20+y2, 80, 20 }
    my[MBoxButton3][Size] = {220, 20+y2, 80, 20 }


    -----------------
    -- window metrics
    -----------------

    -- add the title bar, border and buttons
    x2 = x2 + (theBorderSize*2)
    y2 = y2 + theTitleSize + (theBorderSize*2) + 46

    -- minimum height
    y2 = max( y2, 100 )

    -- number of buttons determine minimum width
    if button3 then     x2 = max( x2, 330 )
    elsif button2 then  x2 = max( x2, 220 )
    else                x2 = max( x2, 120 )
    end if

    -- position box
    x1 = floor( (xPixels - x2) /2 )
    y1 = floor( (yPixels - y2) /2 )
    my[MsgBox][Size] = { x1, y1, x2, y2 }


    ----------------
    -- set up to run
    ----------------


    -- make it modal
    my[MsgBox][Modal] = 1

    -- call the main loop
    eventLoop( MsgBox )

    -- show the mouse
    mouse_pointer(1)

    -- return result
    return MBoxResult

end function



-----------------------------------------------------------------------------
procedure abortErr2( sequence error, sequence args )

    -- More complex error handler; this uses windows
    integer result

    -- pop up message box
    result = message_box( sprintf( error, args ), "Program Aborting",
		MB_ICONEXCLAMATION )

    -- dos message
    abortErr1( error, args )

end procedure

abortHook = routine_id("abortErr2")


-----------------------------------------------------------------------------
-- Main Routine

-----------------------------------------------------------------------------
global procedure WinMain( integer mainWin )

    -- set the main window, and start the event loop
    integer result

    -- is it a window?
    if my[mainWin][Class] != Window then
	abortErr( "WinMain passed a non-window parameter.", {} )
    end if

    -- main loop
    eventLoop( mainWin )

    -- restore mode
    abortErr1( "Goodbye from %s.", {VERSION} )


end procedure


-- In Euphoria 2.1 this file (unless modified) is FREE - 0 statements.
--with 2625022441 -- delete this statement if it causes an error

