-- example12.exw
--
-- Demonstrate colored text

include dos32lib.e
without warning


-----------------------------------------------------------------------------
-- Euphoria keywords and routines built in to the interpreter, ex.exe

global constant keywords = {
    "if", "end", "then", "procedure", "else", "for", "return",
    "do", "elsif", "while", "type", "constant", "to", "and", "or",
    "exit", "function", "global", "by", "not", "include",
    "with", "without"}

global constant builtins = {
    "length", "puts", "integer", "sequence", "position", "object",
    "append", "prepend", "print", "printf", 
    "clear_screen", "floor", "getc", "gets", "get_key",
    "rand", "repeat", "atom", "compare", "find", "match",
    "time", "command_line", "open", "close", "trace", "getenv",
    "sqrt", "sin", "cos", "tan", "log", "system", "date", "remainder",
    "power", "machine_func", "machine_proc", "abort", "peek", "poke", 
    "call", "sprintf", "arctan", "and_bits", "or_bits", "xor_bits",
    "not_bits", "pixel", "get_pixel", "mem_copy", "mem_set",
    "c_proc", "c_func", "routine_id", "call_proc", "call_func", 
    "poke4", "peek4s", "peek4u"}


global constant library = {
    "all_palette", "allocate", "allocate_low", "allocate_string",
    "allow_break", "atom_to_float32", "atom_to_float64", "bits_to_int",
    "bk_color", "bytes_to_int", "call_back", "check_break", "crash_message",
    "current_dir", "cursor", "custom_sort", "define_c_func", "define_c_proc",
    "dir", "display_image", "display_text_image", "dos_interrupt",
    "draw_line", "ellipse", "float32_to_atom", "float64_to_atom",
    "free", "free_low", "get", "get_active_page", "get_all_palette",
    "get_display_page", "get_mouse", "get_position", "get_vector",
    "graphics_mode", "int_to_bits", "int_to_bytes", "lock_memory",
    "lower", "message_box", "mouse_events", "mouse_pointer",
    "open_dll", "palette", "polygon", "read_bitmap", "save_bitmap",
    "save_image", "save_screen", "save_text_image", "scroll",
    "seek", "set_active_page", "set_display_page", "set_rand",
    "set_vector", "sort", "sound", "text_rows", "tick_rate",
    "upper", "use_vesa", "value", "video_config", "wait_key",
    "where", "wildcard_file", "wildcard_match", "wrap"}



-----------------------------------------------------------------------------
-- Colors for syntax elements
global constant NORMAL_COLOR = Black,
		COMMENT_COLOR = Red,
		KEYWORD_COLOR = Blue,
		BUILTIN_COLOR = Magenta,
		LIBRARY_COLOR = Green,
		STRING_COLOR = Green,
		BRACKET_COLOR = {NORMAL_COLOR, Yellow, BrightBlue, Cyan, Green }

-- character classes
constant    DIGIT = 1,
	    OTHER = 2,
	    LETTER  = 3,
	    BRACKET = 4,
	    QUOTE   = 5,
	    DASH = 6,
	    WHITE_SPACE = 7,
	    NEW_LINE = 8

global sequence char_class

-- set up character classes for easier line scanning
-- (assume no 0 char)

	char_class = repeat(OTHER, 255)

	char_class['a'..'z'] = LETTER
	char_class['A'..'Z'] = LETTER
	char_class['_'] = LETTER
	char_class['0'..'9'] = DIGIT
	char_class['['] = BRACKET
	char_class[']'] = BRACKET
	char_class['('] = BRACKET
	char_class[')'] = BRACKET
	char_class['{'] = BRACKET
	char_class['}'] = BRACKET
	char_class['\''] = QUOTE
	char_class['"'] = QUOTE
	char_class[' '] = WHITE_SPACE
	char_class['\t'] = WHITE_SPACE
	char_class['\n'] = NEW_LINE
	char_class['-'] = DASH


sequence segment, segments
sequence line               -- the line being processed
integer segStart, segEnd    -- start and end of current segment of line
integer color               -- the current color

-----------------------------------------------------------------------------
procedure segflush(integer newColor)

	-- if the color is changing, start a new segment

	-- color changed?
	if newColor != color then
		-- build segment, and append it
		segment = line[segStart..segEnd]
		segments = append( segments, { color, segment }  )
		segment = ""
		segStart = segEnd+1
		color = newColor
	end if

end procedure

-----------------------------------------------------------------------------
global function BuildColorLine(sequence pline)

	-- color code a line of text. the text is broken up into segments in the
	-- form { colorPair, text }

	integer class, last, i, c, bracketLevel
	sequence word

	line = pline
	color = Black -- default color
	bracketLevel = 0
	segStart = 1
	segEnd = 0
	segments = {}
	segment = ""

	while 1 do

	    -- past end?
	    if segEnd = length( line ) then
		    exit
	    end if
    
	    -- get the character            
	    c = line[segEnd+1]
    
	    -- get the class
	    if c then
		-- find value
		class = char_class[c]
	    else
		-- null
		class = OTHER
	    end if
    
	    -- act on class of character
    
	    if class = WHITE_SPACE then
		-- continue
		segEnd = segEnd + 1
    
	    elsif class = LETTER then
		last = length(line)
		for j = segEnd + 2 to last do
		    c = line[j]
		    class = char_class[c]
		    if class != LETTER then
			if class != DIGIT then
			    last = j - 1
			    exit
			end if
		    end if
		end for
	
		word = line[segEnd+1..last]
	
		if    find(word, keywords) then  segflush(KEYWORD_COLOR)            
		elsif find(word, builtins) then  segflush(BUILTIN_COLOR)                    
		elsif find(word, library)  then  segflush(LIBRARY_COLOR)                    
		else                              segflush(NORMAL_COLOR)
		end if
		
		segEnd = last
    
	    elsif class <= OTHER then -- DIGIT too
		segflush(NORMAL_COLOR)
		segEnd = segEnd + 1
    
	    elsif class = BRACKET then
    
		if find(c, "([{") then
		    bracketLevel = bracketLevel + 1
		end if
    
		if  bracketLevel >= 1
		and bracketLevel <= length(BRACKET_COLOR) then
		    segflush(BRACKET_COLOR[bracketLevel])
		else
		    segflush(NORMAL_COLOR)
		end if
    
		if find(c, ")]}") then
		    bracketLevel = bracketLevel - 1
		end if
    
		segEnd = segEnd + 1
    
	    elsif class = DASH then
    
		-- next char also a dash?   
		if length( line ) > segEnd+1 then
		    if line[segEnd+2] = '-' then
			segflush(COMMENT_COLOR)
			segEnd = length(line)-1
			exit
		    end if
		end if
	
		segflush(NORMAL_COLOR)
		segEnd = segEnd + 1
    
	    else  -- QUOTE
		i = segEnd + 2
		while i < length(line) do
    
		    -- end quote of same type?
		    if line[i] = c then
			-- move past & exit
			i = i + 1
			exit
    
		    elsif line[i] = '\\' then
			-- skip; it's an embedded code
			if i < length(line)-1 then
			    i = i + 1 -- ignore escaped char
			end if
		    end if
    
    
		    i = i + 1
    
		end while
    
		-- output the string
		segflush(STRING_COLOR)
		segEnd = i - 1
    
	    end if
    
	end while

	-- build final segment
	segment = line[segStart..segEnd]
	if length( segment ) != 0 then
		segments = append( segments, { color, segment }  )
	end if

	return segments

end function


-----------------------------------------------------------------------------
function readFile( sequence fName )

    -- return a sequence of lines from the file
    -- convert '\n' to ' '

    integer handle
    sequence out
    object data

    -- open the file
    handle = open( fName, "r" )

    -- initialize data
    out = {}
    
    while 1 do
	-- read a line of text
	data = gets( handle ) 
	
	-- end of file?
	if atom( data ) then
	    exit
	end if

	-- remove line feed        
	if data[length(data)] = '\n' then
	    data[length(data)] = ' '
	end if
	
	-- add to data
	out = append( out, data )
	
    end while

    -- close the file
    close( handle )

    -- return the text    
    return out
    
end function

-----------------------------------------------------------------------------
-- read the file
sequence fText
fText = readFile( "ex12.ex" )    

-----------------------------------------------------------------------------
constant TextWin =  
    create( Window, "Text Window", 0, Default, Default, 600, 400, 0 )


-----------------------------------------------------------------------------
-- for moving up and down
integer topLine
topLine = 1
-----------------------------------------------------------------------------
procedure drawColorText()

    integer h, w, col, row, maxY
    sequence s, segs
    
    -- use fixed width font
    setFont( TextWin, "Courier New", 10, Bold )


    -- get font size          
    s = getFontSize( TextWin )
    w = s[1]
    h = s[2]   

    -- get screen size
    s = getWindowSize( TextWin )
    maxY = s[4] - s[2]
     
    -- set row
    row = 0
     
    -- place the text in the window    
    for i = topLine to length( fText ) do
	-- start position
	col = 0             
	
	-- "colorize" the text
	segs = BuildColorLine( fText[i] )
	
	-- display the segments, in color
	for j = 1 to length( segs ) do
	
	    -- set the color for this segment
	    setTextColor( TextWin, segs[j][1] )

	    -- set the column position
	    setPosition( TextWin, col * w, row * h )
	    
	    -- display the text
	    wPuts( TextWin, segs[j][2] )
	    
	    -- advance the column
	    col = col + length( segs[j][2] )
	    
	end for
	
	-- advance the row
	row = row + 1
	
	-- past end of window?
	if row * h > maxY then
	    -- done
	    exit
	end if
	
    end for

end procedure


-----------------------------------------------------------------------------
procedure onKeyDown_TextWin( integer key )

    integer refresh

    -- clear refresh flag
    refresh = 0

    -- key: up
    if key = VK_UP then
	-- don't scroll past top
	if topLine > 1 then     
	    -- move
	    topLine = topLine - 1
	    -- set redraw flag
	    refresh = 1
	end if                   

    -- key: down        
    elsif key = VK_DOWN then
	-- don't move past end
	if topLine < length( fText ) then
	    -- move
	    topLine = topLine + 1
	    -- set redraw flag
	    refresh = 1
	end if

    -- key: page up
    elsif key = VK_PAGEUP then
	-- don't move if at top
	if topLine > 1 then   
	    -- move
	    topLine = topLine - 20
	    
	    -- too far?
	    if topLine < 1 then
		-- clip
		topLine = 1
	    end if            
	    
	    -- set redraw flag              
	    refresh = 1
	end if

    -- key: page down            
    elsif key = VK_PAGEDOWN then
	
	if topLine < length( fText ) then
	    topLine = topLine + 20
	    if topLine > length( fText ) then
		topLine = length( fText )
	    end if                   
	    refresh = 1
	end if

    -- key: home        
    elsif key = VK_HOME then
	-- don't move if already there
	if topLine > 1 then
	    -- move
	    topLine = 1    
	    -- set redraw flag
	    refresh = 1
	end if
	
    -- key: end                
    elsif key = VK_END then
	-- don't move if already there
	if topLine < length( fText ) then
	    -- move
	    topLine = length( fText )
	    -- set redraw flag
	    refresh = 1       
	end if

    end if        
    
    -- is redraw flag set?
    if refresh then        
	-- trigger an onPaint event
	repaintWindow( TextWin )     
    end if             
    
    
end procedure

onKeyDown[TextWin] = routine_id( "onKeyDown_TextWin" )


-----------------------------------------------------------------------------
procedure onPaint_TextWin( integer x1, integer y1, integer x2, integer y2 )

    -- redraw the window
    drawColorText()

end procedure

onPaint[TextWin] = routine_id( "onPaint_TextWin" )

-----------------------------------------------------------------------------

-- run the main 
WinMain( TextWin )


