-- ex18.exw
--
-- Game demonstrating Win32Lib features
-- version 2.1 - display of information uses static text controls
-- version 2.0 - game now has look ahead to "good" moves

include dos32lib.e
without warning

-----------------------------------------------------------------------------
constant 
    Win         =  create( Window, "Connect", 0, Default, Default, 250, 300, 0 ),
        Winner  = create( LText, "", Win, 0, 0, 200, 20, 0 ),
        Explain = create( CText, "", Win, 0, 220, 250, 30, 0 ),
        MFile   =  create( Menu, "&File",   Win, 0,0,0,0,0 ),
            NewGame     =  create( MenuItem, "&New Game", MFile, 0,0,0,0,0 ),
            Sep1        =  create( MenuItem, "-", MFile, 0,0,0,0,0 ),          
            ExitGame    =  create( MenuItem, "E&xit", MFile, 0,0,0,0,0 ),
        MOpts   =  create( Menu, "&Options",   Win, 0,0,0,0,0 ),
            Level1Game  =  create( MenuItem, "Level &1 (Easy)", MOpts, 0,0,0,0,0 ),
            Level2Game  =  create( MenuItem, "Level &2 (Moderate)", MOpts, 0,0,0,0,0 ),
            Level3Game  =  create( MenuItem, "Level &3 (Hard)", MOpts, 0,0,0,0,0 ),
            Sep2        =  create( MenuItem, "-", MOpts, 0,0,0,0,0 ),
            ExplainMove =  create( MenuItem, "Explain Move", MOpts, 0,0,0,0,0 ),
        MHelp   =  create( Menu, "&Help", Win, 0,0,0,0,0 ),
            AboutGame   =  create( MenuItem, "&About Connect...", MHelp, 0,0,0,0,0 )
            
setCheck( Level1Game, 1 )
setCheck( ExplainMove, 1 )            

-----------------------------------------------------------------------------
constant BlackBall = textToBitmap( {
    "       xxxxxxxx       ",
    "     xxxxxxxxxxxx     ",
    "    xxxxxxxxxxxxxx    ",
    "   xxxxxx x  xxxxxx   ",
    "  xxxxx   xxxxxxxxxx  ",
    " xxxx  xxxxxxxxxxxxxx ",
    " xxxx x xxxxxxxxxxxxx ",
    "xxxx  xxxxxxxxxxxxxxxx",
    "xxxx xxxxxxxxxxxxxxxxx",
    "xxx  xxxxxxxxxxxxxxxxx",
    "xxxx xxxxxxxxxxxxxxxxx",
    "xxx xxxxxxxxxxxxxxxxxx",
    "xxxxxxxxxxxxxxxxxxxxxx",
    "xxxxxxxxxxxxxxxxxxxxxx",
    "xxxx xxxxxxxxxxxxxxxxx",
    " xxxxxxxxxxxxxxxxxxxx ",
    " xxxxxxxxxxxxxxxxxxxx ",
    "  xxxxxxxxxxxxxxxxxx  ",
    "   xxxxxxxxxxxxxxxx   ",
    "    xxxxxxxxxxxxxx    ",
    "     xxxxxxxxxxxx     ",
    "       xxxxxxxx       "} )

constant WhiteBall = textToBitmap( {
    "       xxxxxxxx       ",
    "     xxxxxxxxxxxx     ",
    "    xxx       xxxx    ",
    "   xx        x  xxx   ",
    "  xx          xx xxx  ",
    " xx         xx xxxxxx ",
    " xx             x xxx ",
    "xx          x xxxxxxxx",
    "xx         x x xxxxxxx",
    "x           x x xxxxxx",
    "xx x    x x x xxxxxxxx",
    "x x    x   x xxxxxxxxx",
    "xx  xx x xx xx xxxxxxx",
    "xx x  x x xxxxxxxxxxxx",
    "xxxxxx xxx xxxxxxxxxxx",
    " xx x xx xxxxxxxxxxxx ",
    " xxxxxxxxxxxxxxxxxxxx ",
    "  xxxxxxxxxxxxxxxxxx  ",
    "   xxxxxxxxxxxxxxxx   ",
    "    xxxxxxxxxxxxxx    ",
    "     xxxxxxxxxxxx     ",
    "       xxxxxxxx       "} )

constant Arrow = textToBitmap( {
    "     xxxxxxxxxxxx     ",
    "     x         xx     ",
    "     x x x x x xx     ",
    "     x  x x x xxx     ",
    "     x x x x xxxx     ",
    "     x  x x x xxx     ",
    "     x x x x xxxx     ",
    "     x  x x x xxx     ",
    "     x x x x xxxx     ",
    "     x  x x x xxx     ",
    "     x x x x xxxx     ",
    "xxxxxx  x x x xxxxxxxx",
    " x   x x x x x x xxxx ",
    "  x   x x x x x xxxx  ",
    "   x   x x x x xxxx   ",
    "    x   x x x xxxx    ",
    "     x   x x xxxx     ",
    "      x   x xxxx      ",
    "       x   xxxx       ",
    "        x  xxx        ",
    "         x xx         ",
    "          xx          "} )

-----------------------------------------------------------------------------
constant
    Spacing = 24,
    Columns = 8,
    Rows    = 8   
    
integer level, explainFlag
    level = 1           -- can change from menu
    explainFlag = 1     -- explain move
    
-----------------------------------------------------------------------------
-- create an 8x8 grid
sequence 
    tile,           -- holds bitmap of tile
    legal,          -- next position in column
    moves           -- holds moves returned by evaluate routines
    
integer 
    gameOver        -- true if game is over

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- The Menu Actions

-----------------------------------------------------------------------------
procedure newGame()                  
    
    -- set up the board for a new game

    -- initialize the grid
    tile    = repeat( repeat( 0, Columns ), Rows )
    legal   = repeat( 0, Columns )
    
    -- set legal moves
    for i = 1 to Columns do
        tile[i][Rows] = Arrow
        legal[i] = Rows
    end for

    -- clear text and flags
    setText( Explain, "" )
    setText( Winner, "" )
    gameOver = 0
    
    
    -- refresh window
    repaintWindow( Win )

end procedure
onClick[NewGame] = routine_id("newGame")


-----------------------------------------------------------------------------
procedure exitGame()                  
    -- exit the game
    closeWindow( Win )    
end procedure
onClick[ExitGame] = routine_id("exitGame")


-----------------------------------------------------------------------------
procedure setLevel( integer i )
    -- set the difficulty level

    -- save level    
    level = i    
    
    -- change check on menu
    setCheck( Level1Game, i = 1 )
    setCheck( Level2Game, i = 2 )
    setCheck( Level3Game, i = 3 )
    
end procedure


-----------------------------------------------------------------------------
procedure level1Game()
    -- change level difficulty
    setLevel(1)
end procedure
onClick[Level1Game] = routine_id("level1Game")

-----------------------------------------------------------------------------
procedure level2Game()
    -- change level difficulty
    setLevel(2)
end procedure
onClick[Level2Game] = routine_id("level2Game")

-----------------------------------------------------------------------------
procedure level3Game()
    -- change level difficulty
    setLevel(3)
end procedure
onClick[Level3Game] = routine_id("level3Game")


-----------------------------------------------------------------------------
procedure setExplain( integer i )
    -- set explanation on or off

    -- save flag
    explainFlag = i    
    
    -- change check on menu
    setCheck( ExplainMove, i )
    
    -- hide/show control
    setVisible( Explain, explainFlag )
    
    -- refresh window
    repaintWindow( Win )
    
end procedure


-----------------------------------------------------------------------------
procedure explainMove()
    -- toggle option on/off
    setExplain( not isChecked( ExplainMove ) )
end procedure
onClick[ExplainMove] = routine_id("explainMove")


-----------------------------------------------------------------------------
procedure aboutGame()
    integer result

    result = message_box(   "Connect Game Demo" & 13 & 10 &
                            "Click on an arrow to move." & 13 & 10 &
                            "Objective:" & 13 & 10 &
                            "Connect four balls in any direction.",
                            "Connect Game Demo",
                            MB_ICONINFORMATION+MB_TASKMODAL )
end procedure
onClick[ AboutGame ] = routine_id( "aboutGame" )




-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- The Game Actions

-----------------------------------------------------------------------------
procedure setTile( integer i, integer j, atom hBitmap )

    -- set at tile on the board

    integer x, y               
    
    -- change the tile
    tile[i][j] = hBitmap

    -- calculate position
    x = i * Spacing
    y = j * Spacing

    -- repaint the tile
    repaintRect( Win, x, y, x+Spacing-1, y+Spacing-1 )

end procedure


-----------------------------------------------------------------------------
procedure placeMarker( integer i, integer j, atom hBitmap )

    -- place a marker on the board, move arrow
    
    -- place the marker
    setTile( i, j, hBitmap )
    
    -- top of column?
    if j = 1 then
        -- no legal moves left in column
        legal[i] = 0
    else              
        -- note legal move
        legal[i] = j-1    
        -- place marker    
        setTile( i,j-1,Arrow)
    end if
    
end procedure




-----------------------------------------------------------------------------
function equal( object o1, object o2 )

    -- return true if objects are equal
    -- this is more clear than  ... compare() = 0 ...
    -- ex:  equal( 10, 10 )
    --      --> 1               
    
    return not compare( o1, o2 )
    
end function



-----------------------------------------------------------------------------
constant         
    -- the directions for the moves
    Vertical        = { {0,0}, {0,1},  {0,2},  {0,3} },
    Horizontal      = { {0,0}, {1,0},  {2,0},  {3,0} },
    Diagonal        = { {0,0}, {1,1},  {2,2},  {3,3} },
    BackDiagonal    = { {0,0}, {-1,1}, {-2,2}, {-3,3} }


-----------------------------------------------------------------------------
procedure addMove( integer i, integer j )
    -- add move to moves list
    moves = append( moves, { i, j } )
end procedure


-----------------------------------------------------------------------------
function tileAt( integer index, integer i, integer j, sequence direction )
    -- return tile at position, offset into pattern by index
    integer u, v
    
    -- get the position
    u = i + direction[index][1]
    v = j + direction[index][2]

    -- indexes in range    
    if u < 1 or u > Columns
    or v < 1 or v > Rows then
        return -1
    end if
    
    return tile[u][v]
    
end function         

-----------------------------------------------------------------------------
function legalMove( integer i, integer j, sequence direction, sequence pattern )
    -- return true if pattern matches vertical slice
    return  tileAt( 1, i, j, direction ) = pattern[1]
    and     tileAt( 2, i, j, direction ) = pattern[2]
    and     tileAt( 3, i, j, direction ) = pattern[3]
    and     tileAt( 4, i, j, direction ) = pattern[4]    
end function

-----------------------------------------------------------------------------
procedure checkMove( integer i, integer j, sequence direction, sequence pattern )
    -- collects move if legal
    if legalMove( i, j, direction, pattern ) then
        -- look in each positon
        for pos = 1 to 4 do    
            -- if legal move (arrow)
            if tileAt( pos, i, j, direction ) = Arrow then
                -- add the move
                addMove( i+direction[pos][1], j+direction[pos][2] )                        
            end if
        end for
    end if
end procedure


-----------------------------------------------------------------------------
function hasLegalDirection( integer i, integer j, sequence pattern )
    -- check if legal in any direction
    return  legalMove( i, j, Horizontal,    pattern )
    or      legalMove( i, j, Vertical,      pattern )
    or      legalMove( i, j, Diagonal,      pattern )
    or      legalMove( i, j, BackDiagonal,  pattern )
end function

-----------------------------------------------------------------------------
procedure checkAllDirections( integer i, integer j, sequence pattern )
    -- check for pattern match in all directions
    checkMove( i, j, Horizontal,   pattern )
    checkMove( i, j, Vertical,     pattern )
    checkMove( i, j, Diagonal,     pattern )
    checkMove( i, j, BackDiagonal, pattern )
end procedure

-----------------------------------------------------------------------------
function hasAnyLegalMatch( sequence pattern )
    -- check all board positions for legal move
    for i = 1 to Columns do
        for j = 1 to Rows do
            if hasLegalDirection( i, j, pattern ) then
                return 1
            end if
        end for
    end for
    return 0
end function

-----------------------------------------------------------------------------
procedure checkAllPositions( sequence pattern )
    -- check entire board for all patterns using this move
    for i = 1 to Columns do
        for j = 1 to Rows do
            checkAllDirections( i, j, pattern )
        end for
    end for
end procedure


-----------------------------------------------------------------------------
function findWinningMove( atom hBmp )
    -- look for winning patterns. returns true if win found for player
    
    return hasAnyLegalMatch( { hBmp, hBmp, hBmp, hBmp } )
    
end function


-----------------------------------------------------------------------------
function findBlock( atom hBmp )
    -- look for blocking patterns, return true if found any.
    -- blocking moves are stored in move
    
    -- blocking patterns
    checkAllPositions( { Arrow, hBmp,  hBmp,  hBmp } )
    checkAllPositions( { hBmp,  Arrow, hBmp,  hBmp } )
    checkAllPositions( { hBmp,  hBmp,  Arrow, hBmp } )
    checkAllPositions( { hBmp,  hBmp,  hBmp,  Arrow  } )

    return length( moves )
    
end function


-----------------------------------------------------------------------------
function findGoodMove( integer hBmp )
    -- look for "good" move, return true if found any.
    -- moves are stored in move

    checkAllPositions( { Arrow, Arrow, hBmp,  hBmp } )
    checkAllPositions( { hBmp,  Arrow, Arrow, hBmp } )
    checkAllPositions( { hBmp,  hBmp,  Arrow, Arrow  } )
    checkAllPositions( { Arrow, hBmp,  hBmp,  Arrow  } )
    checkAllPositions( { Arrow, hBmp,  Arrow, hBmp } )
    checkAllPositions( { hBmp,  Arrow, hBmp,  Arrow  } )

    return length( moves )
    
end function

-----------------------------------------------------------------------------
function findLegalMove( integer hBmp )

    -- collect a list of legal moves, returns true if found any
    -- legal moves are store in moves
    
    moves = {}
    for column = 1 to length( legal ) do
        if legal[column] then
            addMove( column, legal[column] )
        end if
    end for

    return length( moves )

end function


-----------------------------------------------------------------------------
function pick( sequence s )
    -- pick a random item from a list
    return s[rand(length(s))]
end function


-----------------------------------------------------------------------------
procedure computerMove()

    -- make a move by the computer
    sequence move
    sequence explain

    -- clear the moves
    moves = {}
    
    -- clear explanation
    explain = ""

    -- game already over?
    if gameOver then
        -- no need to look for move, game over
        setText( Explain, "Game already over" )
        
    -- look for a winning move        
    elsif findBlock( WhiteBall ) then
        -- found a winning move
        setText( Winner, "Computer Wins" )
        setText( Explain, "Found Winning Move" )
        
    elsif level > 1
    and   findBlock( BlackBall ) then
        -- found a move that blocks opponent
        setText( Explain, "Blocked Your Move" )
        
        
    elsif level > 2
    and  findGoodMove( WhiteBall  ) then
        -- found a move that strengthens position
        setText( Explain, "Found Good Move" )
                
    elsif level > 2
    and   findGoodMove( BlackBall  ) then
        -- found a move that weakens opponent's position
        setText( Explain, "Move Weakens Your Position" )
        
    elsif findLegalMove( BlackBall ) then
        -- found a legal move
        setText( Explain, "Found Legal Move" )
        
    else
        -- no moves remain
        setText( Winner, "It's a draw" )
        setText( Explain, "Now Moves Remaining" )
        
    end if

    -- found a move?
    if length( moves ) then
    
        -- pick a move at random
        move = pick( moves )
    
        -- place marker
        placeMarker( move[1], move[2], WhiteBall )
        
    end if

end procedure



-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- The Window Actions

-----------------------------------------------------------------------------
procedure Paint( integer x1, integer y1, integer x2, integer y2 )
    -- tile the bitmaps randomly across the window

    integer x, y
    
    -- top of grid
    drawRectangle( Win, 0, Spacing-1, Spacing-2, 
                    (Rows+1)*Spacing, (Columns+1)*Spacing )
    
    -- vertical lines    
    for i = 2 to Rows do
        x = (i*Spacing)-1
        drawLine( Win, x, Spacing-2, x, (Columns+1)*Spacing )
    end for

    -- draw the tiles
    for i = 1 to Columns do
        for j = 1 to Rows do
        
            -- calculate tile position
            x = (i * Spacing)
            y = (j * Spacing)
            
            -- in repaint rectangle?
            if x > x2 or x+Spacing-1 < x1
            or y > y2 or y+Spacing-1 < y1 then
                -- not in rectangle
            else  
                -- is there something to display?
                if tile[i][j] then
                    -- draw it
                    drawBitmap( Win, tile[i][j], x, y )
                end if
            end if
        end for
    end for                  
    
end procedure
onPaint[Win] = routine_id("Paint")

-----------------------------------------------------------------------------
procedure Mouse( integer event, integer x, integer y )
    
    -- uncover the clicked tile
    integer i, j, endGame
    object move

    -- left click?
    if event = LEFT_DOWN then
    
        -- convert to a tile
        i = floor( x/Spacing )
        j = floor( y/Spacing ) 
        
        -- clicked a tile?
        if  not gameOver                    -- game still running
        and i > 0 and i < Columns+1         -- legal column
        and j > 0 and j < Rows+1 then       -- legal row
        
            -- is it an arrow (legal move)?
            if tile[i][j] = Arrow then
        
                -- change it to a marker
                placeMarker( i, j, BlackBall )
                
                -- did the player win?
                if findWinningMove( BlackBall ) then
                    setText( Winner, "You Win" )
                    gameOver = 1
                end if
    
                -- computer's turn
                computerMove()            
                        
            end if
        
        end if          
    
    end if

                          
end procedure
onMouse[Win] = routine_id("Mouse")



-----------------------------------------------------------------------------
procedure Open()
    
    -- font to use
    setFont( Win, "Arial", 10, 0 )
    
    -- initialize
    newGame()           -- new game
    setLevel( 2 )       -- moderate difficulty
    setExplain( 1 )     -- explains moves
    
end procedure
onOpen[Win] = routine_id("Open")



WinMain( Win )

