-- generic.exw
--
-- Demo of "generic" application
-- 
-- Features:
--      Menus
--      File | Open         -- getOpenFileName() common control
--      File | Save As      -- getSaveFileName() common control
--      File | Close        -- closeWindow()
--      Help                -- messageBox() routine
--
-- Resizable MLE          
-- cut/copy/paste/undo/clear


include dos32lib.e
without warning


-- version
constant
    Ver     = "Version 1.0"

-- name of the currently open file
sequence openFileName
    openFileName = ""

-- default file types the app recognizes
constant FileTypes = {
    "Text File", "*.TXT",
    "Euphoria Program", "*.EX;*.EXW;*.E;*.EW",
    "All Files", "*.*" }
    

constant Generic = create( Window, "Untitled - Generic Application", 
                            0, Default, Default, 400, 200, 0 )


constant GText = create( MleText, "", 
                            Generic, Default, Default, 400, 200, 0 )

-- need to change to handle multiple windows
-- also, i need to keep track of the menu id, so i can edit
constant
    FileMenu    = create( Menu, "&File", Generic, 0, 0, 0, 0, 0 ),
    MenuNew     = create( MenuItem, "&New",     FileMenu, 0, 0, 0, 0, 0 ),
    MenuOpen    = create( MenuItem, "&Open",    FileMenu, 0, 0, 0, 0, 0 ),
    MenuSave    = create( MenuItem, "&Save",    FileMenu, 0, 0, 0, 0, 0 ),
    MenuSaveAs  = create( MenuItem, "Save &As", FileMenu, 0, 0, 0, 0, 0 ),
    MenuSep1    = create( MenuItem, "-",        FileMenu, 0, 0, 0, 0, 0 ),
    MenuClose   = create( MenuItem, "&Close",   FileMenu, 0, 0, 0, 0, 0 ),

    EditMenu    = create( Menu, "&Edit", Generic, 0, 0, 0, 0, 0  ),
    MenuUndo    = create( MenuItem, "&Undo",    EditMenu, 0, 0, 0, 0, 0 ),
    MenuSep2    = create( MenuItem, "-",        EditMenu, 0, 0, 0, 0, 0 ),
    MenuCut     = create( MenuItem, "Cu&t",     EditMenu, 0, 0, 0, 0, 0 ),
    MenuCopy    = create( MenuItem, "&Copy",    EditMenu, 0, 0, 0, 0, 0 ),
    MenuPaste   = create( MenuItem, "&Paste",   EditMenu, 0, 0, 0, 0, 0 ),
    MenuDelete  = create( MenuItem, "De&lete",  EditMenu, 0, 0, 0, 0, 0 ),
                                                  
    HelpMenu    = create( Menu, "&Help", Generic, 0, 0, 0, 0, 0  ),
    MenuAbout   = create( MenuItem, "&About...\tAlt+A", HelpMenu, 0, 0, 0, 0, 0 )


-----------------------------------------------------------------------------
-- MENU ACTIONS
-----------------------------------------------------------------------------
procedure onClick_MenuNew()
    -- clear file, and blank name

    -- clear window title    
    setText( Generic, "Untitled - Generic Application" )
    
    -- clear file name
    openFileName = ""
    
    -- clear the MLE
    setText( GText, "" )
    
end procedure           

onClick[ MenuNew ] = routine_id( "onClick_MenuNew" )


-----------------------------------------------------------------------------
procedure setOpenFileName( sequence fName )

    -- save the name of the open file
    -- update the title bar of the window

    sequence shortName

    -- save name
    openFileName = fName

    -- truncate path from name
    shortName = fName
    for i = length( fName ) to 1 by -1 do
        if fName[i] = '\\' then
            shortName = fName[i+1..length(fName)]
            exit
        end if
    end for
    
    -- update title bar
    setText( Generic, shortName & " - Generic Application" )
    
end procedure


-----------------------------------------------------------------------------
procedure onClick_MenuOpen()

    -- read a file into the MLE
    
    integer handle
    sequence fName, buffer
    object data
    
    -- get the file name    
    fName = getOpenFileName( Generic, "", FileTypes )
    
    -- entered a file name?
    if length( fName ) = 0 then
        return
    end if

    -- open the file
    handle = open( fName, "r" )
    
    -- read the file
    buffer = {}
    while 1 do
        -- get data
        data = gets(handle)
        
        -- end of file?
        if atom(data) then
            exit
        end if

        -- replace { ... 10 } with { ... 13, 10 }
        data = data[1..length(data)-1] & 13 & 10
        
        -- append
        buffer = buffer & data

    end while

    -- close the file
    close( handle )

    -- place the data in the mle
    setText( GText, buffer )

    -- set file name
    setOpenFileName( fName )    
    
end procedure

onClick[ MenuOpen ] = routine_id( "onClick_MenuOpen" )



-----------------------------------------------------------------------------
procedure saveFile( sequence fName )

    -- write the current data out as a file

    integer handle, ignore

    -- open the file
    handle = open( fName, "w" )
    if handle = -1 then
        -- give message and exit
        ignore = message_box(   "Save File Error", 
                                "Unable to create the file " & fName,
                                MB_ICONHAND+MB_TASKMODAL )
        return    
    end if
    
    -- write the file
    puts( handle, getText( GText ) )
    
    -- close the file
    close( handle )
    
    -- set file name
    setOpenFileName( fName )    

end procedure 


-----------------------------------------------------------------------------
procedure onClick_MenuSaveAs()
    
    -- save as
    sequence fName
    
    fName = getSaveFileName( Generic, openFileName, FileTypes )

    -- cancelled?    
    if length(fName) = 0 then
        return
    end if    
    
    -- save the file
    saveFile( fName )
    
end procedure

onClick[ MenuSaveAs ] = routine_id( "onClick_MenuSaveAs" )


-----------------------------------------------------------------------------
procedure onClick_MenuSave()
    
    -- save

    integer ignore    
    
    -- named file?
    if length( openFileName ) = 0 then
        -- give message and exit
        ignore = message_box(   "Save Error", 
                                "Please select a file name.",
                                MB_ICONHAND+MB_TASKMODAL )
        return
    end if
    
    -- write the file out
    saveFile( openFileName )
    
end procedure

onClick[ MenuSave ] = routine_id( "onClick_MenuSaveAs" )

-----------------------------------------------------------------------------
procedure onClick_MenuClose()
    -- close window
    closeWindow( Generic )
end procedure      

onClick[ MenuClose ] = routine_id( "onClick_MenuClose" )


-----------------------------------------------------------------------------
procedure onClick_MenuUndo()
    -- undo
    undo( GText )
end procedure      

onClick[ MenuUndo ] = routine_id( "onClick_MenuUndo" )

-----------------------------------------------------------------------------
procedure onClick_MenuCut()
    -- cut
    cut( GText )
end procedure      

onClick[ MenuCut ] = routine_id( "onClick_MenuCut" )


-----------------------------------------------------------------------------
procedure onClick_MenuCopy()
    -- copy
    copy( GText )
end procedure      

onClick[ MenuCopy ] = routine_id( "onClick_MenuCopy" )


-----------------------------------------------------------------------------
procedure onClick_MenuPaste()
    -- paste
    paste( GText )
end procedure      

onClick[ MenuPaste ] = routine_id( "onClick_MenuPaste" )


-----------------------------------------------------------------------------
procedure onClick_MenuDelete()
    -- clear
    clear( GText )
end procedure      

onClick[ MenuDelete ] = routine_id( "onClick_MenuDelete" )


-----------------------------------------------------------------------------
-- about... behavior
procedure onClick_MenuAbout()
    integer result

    result = message_box(   "Win32Lib Generic Demo"             -- text
                            & 13 & 10 & Ver,
                            "About This Demo",                  -- title
                            MB_ICONINFORMATION+MB_TASKMODAL )   -- icon
end procedure

onClick[ MenuAbout ] = routine_id( "onClick_MenuAbout" )


-----------------------------------------------------------------------------
procedure onResize_Generic( integer style, integer sizeX, integer sizeY )

    -- size the MLE control
    moveWindow( GText, 0, 0, sizeX, sizeY, 1 )

end procedure      

onResize[ Generic ] = routine_id( "onResize_Generic" )



WinMain( Generic )


